/**
 * @(#) Class to express data packs.
 */

package ohymosj;

import java.util.*;
import java.io.*;

/**
 * OhDataPack is an abstract class for treating a pair of data
 * time and data value which we call a datapack. <p>
 *
 * In hydrologic simulations in which the states are updated
 * dynamically, the data which are sent and received by
 * elements, subsytems, and total sytem of the hydrologic system
 * are pairs of data value(s) and a "data time". Here, "data
 * time" means the time to which the data are related. <p>
 *
 * For example, suppose we have the value of discharge Q at time
 * t at some stream point. Then, we call the value of discharge
 * Q "data value" and time t "data time". <p>
 *
 * We call a pair of data time and data value a "datapack," and
 * we prepare class OhDataPack, which is an abstract class used
 * as a base class from which various datapack classes are
 * derived.
 *
 * <pre>
 * .....
 *   "datapack" = "data time" + "data value"
 * ....
 * </pre>
 *
 * Because there are various data types for data value, there
 * could be various datapack classes. <p>
 *
 * Class OhDataPack is a base class from which various datapack
 * classes are derived.  Note that the classed derived from the
 * OhDataPack class should be immutable.
 *
 * <japanese>
 * <p>
 * f[^ƃf[^l̃yA琬f[^pbN^쐬邽
 * ߂̒ۃf[^^f <p>
 *
 * ̃NX OhDataPack ł́A
 *
 * <dl>
 *  <li> f[^ : _dataTime
 * </dl>
 *
 * LA
 *
 * <dl>
 *  <li> f[^̐ݒ@\
 *  <li> f[^̎擾@\
 * </dl>
 *
 * `܂B<p>
 *
 * ef󂷂f[^̃f́ÃNX OhDataPack 
 * pč쐬ĉB<p>
 *
 * NX OhDataPack păf[^pbN^쐬Tv
 * ܂BNX OhDataPack pč쐬f[^pbN
 * ^́Aimmutable ł悤ɂĂĂB
 *
 * <pre>
 * package ohymosj;
 * import java.util.*;
 * import java.io.*;
 * public class OhDataPackFoo extends OhDataPack
 *  implements Serializable, Cloneable
 * {
 *   private final Foo _dataValue;
 *   // constructors
 *   public OhDataPackFoo() { ... }
 *   public OhDataPackFoo(ohTimePointImmutable aDataTime,
 *     Foo aDataValue) { ... }
 *   public OhDataPackFoo(String aDataStr) { ... }
 *   // abstract methods in OhDataPack
 *   public String dataType() { ... }
 *   public void dpack2Ofile(PrintWriter os) { ... }
 *   public void dpack2File(PrintWriter os) { ... }
 *   // nonabstract methods
 *   public Object clone() { ... }
 *   public Foo getDataValue() { ... }
 * }
 * </pre>
 * </japanese>
 *
 * @author satoh, shiiba
 * @version $Revision: 1.13 $ $Date: 2004/01/18 14:05:10 $
 */
public abstract class OhDataPack    // abstract class "OhDataPack"
  implements OhConst, Cloneable
{
  // @begin{field} -------------------------------------------------------
  
  /**
   * The time when the data were recorded.
   *
   * <japanese>
   * <p>
   * f[^Bf[^NB
   * </japanese>
   *
   */
  protected OhTimePoint _dataTime;

  // @end{field} ---------------------------------------------------------

  
  // @begin{abstract method} ---------------------------------------------
  
  /**
   * Returns the data type.
   *
   * @return a <code>String</code> value
   */
  public abstract String dataType();


  // data transformation.
  
  /**
   * A data pack is written out to a file by the format for
   * internal processing.<p>
   *
   *
   * <japanese>
   * <p>
   * f[^pbNptH[}bgŃt@Cɏo
   * \bhłBۃ\bhłBptH[}bgƂ́A
   * f[^pbN\f[^sɈA( ) ň͂
   * otH[}bĝƂ܂B
   * </japanese>
   *
   * @param os a <code>PrintWriter</code> value
   */
  public abstract void dpack2File(PrintWriter os);
  
  /**
   * A data pack is written out to a file by simple format.
   *
   * <japanese>
   * <p>
   * f[^pbNȈՃtH[}bgŃt@Cɏo܂B<p>
   *
   * ȈՃtH[}bgƂ́A_year, _tsec, _dataValue 󔒂ŋ؂
   * Ĉsɏo܂B
   * </japanese>
   *
   * @param os a <code>PrintWriter</code> value
   */
  public abstract void dpack2Ofile(PrintWriter os);

  // @end{abstract method} -----------------------------------------------


  // @begin{constructor} -------------------------------------------------
  
  /**
   * Data_pack() initializes _dataTime to OhTimePoint().
   *
   * <japanese>
   * <p>
   * ftHgRXgN^ OhDataPack() ́Aoϐ 
   * _dataTime  new OhTimePoint() ɏ܂B
   * </japanese>
   *
   */
  protected OhDataPack() {
    _dataTime = new OhTimePoint();
  }

  // @end{constructor} ---------------------------------------------------

  
  // @begin{methods} -----------------------------------------------------
  
  // setting and getting data recording times
  
  /**
   * Returns the data time.
   *
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePointImmutable getDataTime() {
    return _dataTime;
  }
  
  /**
   * Sets the data time. 
   *
   * @param aDataTime an <code>OhTimePoint</code> value
   */
  protected void setDataTime(OhTimePointImmutable aDataTime) {
    _dataTime.copy(aDataTime);
  }

  // @end{methods} -------------------------------------------------------

} // OhDataPack
