/**
 * @(#) OhError; ohymosj error exception class.
 */

package ohymosj;

import java.util.*;
import java.util.regex.*;
import java.io.*;

/**
 * The OhError class is the Ohymosj error exception class which
 * is derived from the <b>Exception</b> class. <p>
 *
 * The <b>OhError</b> class object constructor is used as
 *
 * <pre>
 * .....
 *   try {
 *     ...
 *     OhError e = new OhError(errorMessage1, errorMessage2);
 *     throw e;
 *   }
 *   catch (OhError e) {
 *     System.out.print(e.what());
 *     e.printStackTrace();
 *     System.exit(1);
 *   }
 * .....
 * </pre>
 *
 * The example of the creation method of an error message is as
 * follows.
 *
 * <pre>
 * .....
 *   if(some error status was detected) {
 *     StringWriter buffer = new StringWriter();
 *     PrintWriter  out    = new PrintWriter(buffer);
 *     oPrint(out, " > ");
 *     String str = buffer.toString();
 *     out.close();
 *     throw new OhError("Already exits \"_dataPackArray\"." + str);
 *   }
 * .....
 * </pre>
 *
 * <japanese>
 * OhError NX́A<b>Exception</b> gONXłB
 * String ^̃tB[hϐ errorMessage ɁAO𑗏oƂ
 * ̏󋵂ݒ肵܂B\bh what() ́AerrorMessage Ԃ
 * B
 * </japanese>
 *
 * @author satoh, shiiba
 * @version $Revision: 1.14 $ $Date: 2004/01/16 00:02:31 $
 */
public class OhError extends Exception
  implements OhConst
{
  // @begin{field} ---------------------------------------------

  /**
   * Error message.
   *
   */
  String _errorMessage;

  // @end{field} -----------------------------------------------

  
  // @begin{constructor} ---------------------------------------

  /**
   * Creates a new <code>OhError</code> instance.
   *
   * @param aStr a <code>String</code> value
   */
  public OhError(String aStr) {
    _errorMessage = aStr;
  }

  /**
   * This method is not necessary, but included for
   * compatibility to the previous revision 1.0.
   *
   * @param firstLine a <code>String</code> value
   * @param secondLine a <code>String</code> value
   */
  public OhError(String firstLine, String secondLine) {
    _errorMessage = firstLine + NEWLINE + secondLine;
  }
  
  // @end{constructor} -----------------------------------------

  
  // @begin{methods} -------------------------------------------

  /**
   * Returns the message that was given when the exception
   * object was thrown.
   *
   * <japanese>
   * <p>
   * OhError O throw ꂽƂɗ^ꂽbZ[WԂ
   * ܂B
   * </japanese>
   *
   * @return a <code>String</code> value
   */
  public String what() {
    return _errorMessage;
  }
  
  /**
   * Returns the message which is made by adding 'aPrefix'
   * before each line of the message that was given when the
   * exception object was thrown.
   *
   * <japanese>
   * <p>
   * OhError OoꂽƂɗ^ꂽbZ[W̊es
   * s aPrefix }ĕԂ܂B
   * </japanese>
   *
   * @param aPrefix a <code>String</code> value
   * @return a <code>String</code> value
   */
  public String what(String aPrefix) {
    return OhCommon.addPrefix(_errorMessage, aPrefix);
  }

  public static String getStackTraceString(Exception e) {
    StackTraceElement[] st = e.getStackTrace();
    StringBuffer sb = new StringBuffer(e.toString() + NEWLINE);
    Pattern p = Pattern.compile(".*:\\d+\\)");
    for (int i = 0; i < st.length; i++) {
      String str = st[i].toString();
      Matcher matcher = p.matcher(str);
      if (matcher.matches()) {
        sb.append("    at " + st[i].toString() + NEWLINE);
      }
      else {
        sb.append("    " + st[i].toString() + NEWLINE);
      }
    }
    return sb.toString();
  }
  
  // @end{methods} ---------------------------------------------

} // OhError
