/**
 * @(#) Receive port model class "OhOutputFile" for output files.
 *
 * <a href="OhObject.html">OhObject</a>
 * <-- <a href="OhPort.html">OhPort</a>
 * <-- <a href="OhReceivePort.html">OhReceivePort</a>
 * <-- <a href="OhOutputPort.html">OhOutputPort</a>
 * <-- .
 */

package ohymosj;

import java.util.*;
import java.io.*;


/**
 * Model for output ports which receive datapacks and print them
 * in a file. <p>
 *
 * Class OhOutputFile stores "output file name" in
 * <a href="OhObjectl.html#field_summary">OhObjectl#_objectName</a>.
 *
 * Methods to output received datapacks to the output file are
 * defined as its methods.
 *
 * <japanese>
 * o̓t@CpM[qf\NXłB<p>
 *
 * ŠnŎMf[^pbNt@CɋL^܂B<p>
 *
 * ۃNXłB
 * <a href="OhOutputPort.html">OhOutputPort</a> NXhA
 * <a href="OhOutputFileDouble.html">OhOutputFileDouble</a>,
 * <a href="OhOutputFileInt.html">OhOutputFileInt</a>,
 * <a href="OhOutputFilelong.html">OhOutputFilelong</a>,
 * <a href="OhOutputFileFloat.html">OhOutputFileFloat</a>
 * NXh܂B
 * </japanese>
 *
 * @author satoh, shiiba
 * @version $Revision: 1.27 $ $Date: 2004/02/14 08:32:57 $
 */
public abstract class OhOutputFile extends OhOutputPort
{
  // @begin{fields} --------------------------------------------

  /**
   * The maximum length of an output data string.<p>
   *
   * <japanese>
   * o̓f[^̍ő咷łB<p>
   *
   * ̒́ÃIuWFNg̃RXgN^Ŏw肵
   * łB
   * </japanese>
   * <p>
   * The maximum length of an output data string.
   *
   */
  protected int _lineLength;
  
  /**
   * <japanese>
   * ŏIo̓f[^̃f[^L^B
   * </japanese>
   * <p>
   * Data record time of the last output data.
   *
   */
  protected OhTimePointImmutable _lastDataTime;

  /**
   * <japanese>
   * t@CɏoRgB
   * </japanese>
   * <p>
   * The comment string written out to a file.
   *
   */
  protected String _commentString;

  // @end{fields} ----------------------------------------------


  // @begin {abstract methods} ---------------------------------

  // 1. Connection function.

  /**
   * This is an abstract method.  Returns a string which
   * expresses the type of datapacks received by this object.<p>
   *
   * <japanese>
   * o̓f[^pbÑf[^ľ^𕶎ƂĕԂ܂B<p>
   *
   * ۃ\bhłB<p>
   * </japanese>
   * <p>
   *
   * @return a <code>String</code> value
   */
  public abstract String dataType();


  // 2. The write-in function of output files.

  /**
   * Writes the content of datapack aDataPack into the area to
   * which aStr points. The space which aStr points to must be
   * large enough to store the content of the datapack. <p>
   *
   * Since this is an abstract method, implementation of this
   * function is compulsory in derived classes.
   *
   * <japanese>
   * o̓f[^Ps̍쐬s܂B<p>
   *
   * ۃ\bhłB<p>
   *
   * o̓t@CPsɂ́AP̃f[^pbÑf[^eLq
   * ܂B<p>
   * </japanese>
   * <p>
   *
   * @param aStr a <code>String</code> value
   * @param aDataPack an <code>Object</code> value
   */
  public abstract void makeStringData(String aStr, Object aDataPack);

  // @end{abstract methods} ------------------------------------


  // @begin{initialization} ------------------------------------

  /**
   * Sets <a href="OhObject.html#field_summary">OhObject#_objectName</a> to aFileName and 
   * <a href="OhOutputFile.html#field_summary">_lineLength</a> to aLineLength. _objectName is
   * used also for specifying the output file name into which
   * the receiving object write data.<p>
   *
   * <japanese>
   * IuWFNg(t@C)Ao̓f[^̍ő咷
   * (_lineLength)AIuWFNgԍ 0 ɐݒ肵܂B
   * <p>
   *
   * <a href="OhOutputFile.html#field_summary">_lineLength</a>  0 傫ĂȂA
   * dݒƔf OhError O𑗏o܂B
   * <a href="OhOutputFile.html#field_summary">_lineLength</a>  
   * 0 ȉȂ΁A<a href="OhOutputFile.html#field_summary">_lineLength</a>  aLineLength ɐݒ肵
   * B
   * </japanese>
   * <p>
   *
   * @param aFileName a <code>String</code> value
   * @param aLineLength an <code>int</code> value
   */
  public void init(String aFileName, int aLineLength)
    throws OhError {
    if (aLineLength < 0) {
      throw new IllegalArgumentException(
        "Invalid value of aLineLength " + aLineLength +
        NEWLINE + oPrintString(" > ")
      );
    }

    if (aLineLength == 0)
      aLineLength = 128;

    super.init(aFileName, 0);
    if(_lineLength > 0) {
      throw new OhError(oPrintString(" > "));
    }
    _lineLength = aLineLength;
  }


  /**
   * Sets the argument string to the object name.<p>
   *
   * <japanese>
   * w肳ꂽt@CIuWFNgɋLAo̓f[^
   * ̍ő咷(_lineLength) 128 ɁAIuWFNgԍ 0 
   * ɐݒ肵܂B<p>
   * </japanese>
   *
   * @param aFileName a <code>String</code> value
   * @return a <code>boolean</code> value
   * @exception OhError if an error occurs
   */
  public void init(String aFileName) throws OhError {
    super.init(aFileName, 0);
    if (_lineLength > 0) {
      //; _lineLength  0 傫ꍇAdݒƔf 
      //; OhError IuWFNgX[܂B
      throw new OhError(oPrintString(" > "));
    }
    _lineLength = 128;
  }

  // @end{initialization} --------------------------------------


  // @begin{constructors} --------------------------------------

  /**
   * <japanese>
   * ftHgRXgN^łB<p>
   *
   * ftHgRXgN^
   * <a href="OhOutputFile.html#method_summary">init(java.lang.String, int)</a>
   * s܂B̌ʁAȉ̐ݒƂȂ܂B<p>
   *
   * <dl>
   *  <li> _objectName        : null
   *  <li> _objectNumber      : -1
   *  <li> _lineLength        : -1
   * </dl>
   *
   * <p>
   * ̑̃oϐɑ΂鏈͈RXgN^Ɠ
   * łB
   * ̏ꍇA<a href="OhOutputFile.html#method_summary">init(java.lang.String, int)</a>
   * păoϐĉB
   * </jananese>
   *
   */
  public OhOutputFile() {
    super();
    _lineLength = -1;

    _className     = "OhOutputFile";
    _commentString = null;
    _lastDataTime  = OhTimePoint.OhTimePointMin;
  }


  /**
   * <japanese>
   * NX <a href="OhOutputPort.html">OhOutputPort</a> ̃ftHgRXgN^
   * pA
   * <a href="OhOutputFile.html#method_summary">init(java.lang.String, int)</a>
   * pāA
   * oϐȉ̒ʂ菉܂B<p>
   *
   * <dl>
   *  <li> _className         : "OhOutputFile"
   *  <li> _objectName        : aFileName
   *  <li> _objectNumber      : 0
   *  <li> _lineLength        : aLineLength
   *  <li> _lastDataTime      : <a href="OhTimePoint.html#field_summary">OhTimePoint#OhTimePointMin</a>
   *  <li> _commentString     : null
   * </dl>
   * </japanese>
   *
   * @param aFileName a <code>String</code> value
   * @param aLineLength an <code>int</code> value
   */
  public OhOutputFile(String aFileName, int aLineLength) {
    super();
    _lineLength = -1;

    _className     = "OhOutputFile";
    _commentString = null;
    _lastDataTime  = OhTimePoint.OhTimePointMin;

    try {
      init(aFileName, aLineLength);
    }
    catch (OhError e) {
    }
  }


  /**
   * <japanese>
   * NX <a href="OhOutputPort.html">OhOutputPort</a> ̃ftHgRXgN^pA
   * <a href="OhOutputFile.html#method_summary">init(java.lang.String)</a> pāA
   * oϐȉ̒ʂ菉܂B<p>
   *
   * <dl>
   *  <li> _className         : "OhOutputFile"
   *  <li> _objectName        : aFileName
   *  <li> _objectNumber      : 0
   *  <li> _lineLength        : -1
   *  <li> _lastDataTime      : <a href="OhTimePoint.html#field_summary">OhTimePoint#OhTimePointMin</a>
   *  <li> _commentString     : null
   * </dl>
   * </japanese>
   *
   * @param aFileName a <code>String</code> value
   * @exception OhError if an error occurs
   */
  public OhOutputFile(String aFileName) {
    super();
    _lineLength = -1;

    _className     = "OhOutputFile";
    _commentString = null;
    _lastDataTime  = OhTimePoint.OhTimePointMin;
    try {
      init(aFileName);
    }
    catch (OhError e) {
    }
  }

  // @end{constructors} ----------------------------------------

  
  // @begin{methods} -------------------------------------------

  /**
   * <japanese>
   * o̓f[^̍ő咷 _lineLength Ԃ܂B<p>
   *
   * ̒́ÃIuWFNg̃RXgN^Ŏw肵̂łB
   * </japanese>
   * <p>
   * Returns "_lineLength".
   *
   * @return an <code>int</code> value
   */
  public int getLineLength() {
    return _lineLength;
  }


  /**
   * <japanese>
   * ŏIo̓f[^̃f[^L^ _lastDataTime Ԃ܂B
   * </japanese>
   * <p>
   * Returns "_lastDataTime".
   *
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePointImmutable getLastDataTime() {
    return _lastDataTime;
  }


  // 1. Connection function.

  // 2. The write-in function of output files.

  /**
   * Copys the string <b>aCmt</b> to the field
   * <b>_commentString</b>. The string copied to
   * <b>_commentString</b>
   * will be printed out into the output file, while the method
   * <a href="OhOutputFile.html#method_summary">writeHeader()</a> is executed. See the source code of
   * <a href="OhOutputFile.html#method_summary">writeHeader()</a>.
   * <p>
   *
   * <japanese>
   * t@CɏoRg _commentString 
   * <b>aCmt</b> 
   * ݒ肵܂B_commentString ɐݒ肵́A
   * <a href="OhOutputFile.html#method_summary">writeHeader()</a> sƂɏo̓t@Cɏo
   * ܂B
   * </japanese>
   * <p>
   *
   * @param aCmt a <code>String</code> value
   */
  public void setCommentString(String aCmt) {
    _commentString = aCmt;
  }


  /**
   * <japanese>
   * o̓f[^݂܂B<p>
   *
   * o̓t@C̃wb_ƂāA^f[^pbN𑗐MĂ
   * M[qɊւL܂B<p>
   *
   * setCommentString(String) Őݒ肵񂪏o܂B
   * o̓f[^ӖĂ邩APʂ͉Ȃǂo
   * B<p>
   *
   * ܂Af[^Ԍ^z̒l\܂B<p>
   *
   * yz
   * <pre>
   *   # tplkin.oa 
   *   # -------------------------------------------------------
   *   #   OhSendDouble sp 0 
   *   #       _owner > PowerLawKinematicWave stream 0
   *   #       _owner >> OhScfTotalSystem totalSystem 0
   *   # -------------------------------------------------------
   *   # Year Sec Frac Discharge(m**3/sec)
   *   # DataConnectionType = 0:BROKENLINE 1:STEP 2:PULSE 3:NOINTP  
   *   #& dctVec 1  0
   * </pre>
   *
   * wb_ɂ͍s # ݁ARgAEĝŁA
   * ̏o̓t@CĂѓ̓t@CƂėp邱Ƃł܂B<p>
   * 
   * o̓t@CI[vłȂƗO𑗏o܂B<p>
   * 
   * wb_ύXꍇ́AhNXōĒ`ĉB<p>
   * </japanese>
   *
   * <p>
   *
   * Writes a header part to output file. The header contains <p>
   * <ul>
   *  <li> file name
   *  <li> information about the send port which sends data to
   *       this port
   *  <li> "_commentString if the "commentString" of the send
   *       port is not null
   * </ul>
   * <p>
   * If the output file cannot be opened, this method throws an
   * exception of type <a href="OhError.html">OhError</a>.
   *
   * @return a <code>boolean</code> value
   * @exception IOException if an error occurs
   */
  public boolean writeHeader() throws IOException {  //; (*2)
    
    BufferedWriter bw = null;
    
    try {
      bw = new BufferedWriter(new FileWriter(_objectName));
    }
    catch(IOException e) {
      throw e;
    }

    PrintWriter pw = new PrintWriter(bw, true);    //; "true" is autoFlush.

    pw.print("# " + _objectName + " " + NEWLINE + "# ");

    for (int k = 0; k < 70; k++)
      pw.print("-");

    pw.print(NEWLINE);

    ((OhObject)_sender).oPrint(pw, "#   ");

    pw.print("# ");

    for (int k = 0; k < 70; k++)
      pw.print("-");

    pw.print(NEWLINE);

    if (_commentString != null)
      pw.print("# " + _commentString + NEWLINE);
    
    String senderCom = ((OhSenderPort)_sender).getCommentString();

    if (senderCom != null)
      pw.print("# " + senderCom + NEWLINE);  //; aho
    
    Vector dctVec = ((OhSenderPort)_sender).getDctVec();

    int n = dctVec.size();

    pw.print("# DataConnectionType = ");

    for (int i = 0; i < DataConnectionType.getNumOfDCT(); i++) {
      pw.print(i + ":" + DataConnectionType.getDctTypeStr(i) + " ");
    }

    pw.print(NEWLINE + "#& dctVec " + n + " ");

    for (int i = 0; i < n; i++) {
      pw.print(" " +
        ((DataConnectionType) dctVec.elementAt(i)).toInt());
    }  

    pw.print(NEWLINE);
    pw.print(NEWLINE);

    pw.close();

    return true;
  }


  /**
   * <japanese>
   * o̓f[^݂܂B<p>
   *
   * ܂Ao̓t@C _objectName ǉ[hŃI[vA
   * I[vłȂꍇ return ܂B<p>
   * 
   * ɁAMf[^pbNz _dataPackArray ̐擪Ɋi[
   * Ăf[^pbN珇Ԃ(Ref
   * "OhOutputFile#makeStringData(java.lang.String,
   * java.lang.Object)") păf[^쐬Ao̓t@
   * CɋL܂B<p> Ӂ <p>
   *
   * Mf[^pbN͂̃\bh̒Ŏ擾܂B
   * f OhOutputFile ͎M[q̔hfłāA
   * f[^pbNXgbN@\͂܂B
   * </japanese>
   * <p>
   * Writing of output data.
   *
   * @exception IOException if an error occurs
   */
  public void outputData() throws IOException {    //; (*3)
    
    OhArrayList   dpArray      = new OhArrayList();
    OhArrayList[] dpArrayArray = {dpArray};

    int   numOfDps      = 0;
    int[] numOfDpsArray = {numOfDps};

    getDataPackArray(dpArrayArray, numOfDpsArray);

    dpArray  = dpArrayArray[0];
    numOfDps = numOfDpsArray[0];

    // _dataPackArray is not updated.
    //
    if (((OhDataPack)dpArray.get(numOfDps - 1)).getDataTime().
      compareTo(_lastDataTime) <= 0)
      return;    

    String buffer = "";

    /* initial output */

    if (_lastDataTime == OhTimePoint.OhTimePointMin)
      writeHeader();

    ////////////////////////////////////////////////
    // Additional writing.
    //
    // <japanese>
    // ݗpo̓t@C̒ǉI[vB
    // </japanese>
    //
    BufferedWriter bw = null;
    
    try {
      bw = new BufferedWriter(
        new FileWriter(_objectName, true));
    }
    catch(IOException e) {
      throw e;
    }

    PrintWriter pw = new PrintWriter(bw);  

    if (((OhDataPack)dpArray.get(0)).getDataTime().
      compareTo(OhTimePoint.OhTimePointMin) == 0) {

      makeStringData(buffer, dpArray.get(0));
      pw.print(buffer + NEWLINE);
    }

    
    int k = 0;

    for (k = 0; k < numOfDps; k++) {
      if (((OhDataPack)dpArray.get(k)).getDataTime().
        compareTo(_lastDataTime) > 0)
        break;
    }

    for (; k < numOfDps; k++) {
      makeStringData(buffer, dpArray.get(k));
      pw.print(buffer + NEWLINE);
    }

    pw.close();

    _lastDataTime =
      ((OhDataPack)dpArray.get(numOfDps - 1)).getDataTime();

    reduceDataPackStock();
  }


  /**
   * <japanese>
   * f[^KvԂ܂B<p>
   *
   * M̑M[q getNecessaryTimeFrom() ȍ~
   * getNecessaryTimeTo ȑÕf[^pbNȂ悤Ɏw
   * 邽߂ɒ`܂B<p>
   * <pre>
   * [RETURN]
   *  EOhTimePoint f[^Kv
   * </pre>
   * </japanese>
   * <p>
   * Returns necessary data time.
   *
   * @return an <code>OhTimePointImmutable</code> value
   */
  public OhTimePointImmutable getNecessaryTimeFrom() {
    return _lastDataTime;
  }


  /**
   * <japanese>
   * f[^KvԂ܂B<p>
   *
   * M̑M[q getNecessaryTimeFrom() ȍ~
   * getNecessaryTimeTo ȑÕf[^pbNȂ悤Ɏw
   * 邽߂ɒ`܂B<p>
   * <pre>
   * [RETURN]
   *  EOhTimePoint f[^Kv
   * </pre>
   * </japanese>
   * <p>
   * Returns necessary data time.
   *
   * @return an <code>OhTimePointImmutable</code> value
   */
  public OhTimePointImmutable getNecessaryTimeTo() {
    return new OhTimePoint(2100, 0);
  }


  // 3. Information output function.

  /**
   * <japanese>
   * IuWFNg(L҂̏܂܂Ȃ)o͂܂B<p>
   *
   * o̓t@CƂ̒[q̎MM[q̏
   * Xg[ aPwijɏo͂܂B
   *
   * <pre>
   *  [in] aPw  : IuWFNgo͂Xg[
   *  [in] aHead: Rg
   * </pre>
   * </japanese>
   * <p>
   * Outputs an object information.
   *
   * @param aPw a <code>PrintWriter</code> value
   * @param aHead a <code>String</code> value
   */
  public void cPrint(PrintWriter aPw, String aHead) {
    aPw.print(aHead + "OhInputFile : " + _className + "<" +
      dataType() + "> " + _objectName + " " + NEWLINE);
  }


  /**
   * <japanese>
   * IuWFNg(L҂̏܂)o͂܂B<p>
   *
   * o̓t@CƂ̒[q̎MM[q̏
   * Xg[ aPwijɏo͂܂B
   *
   * <pre>
   *  [in] aPw  : IuWFNgo͂Xg[
   *  [in] aHead: Rg
   * </pre>
   * </japanese>
   * <p>
   * Outputs an object information.
   * The information includes the "_owner"'s information.
   *
   * @param aPw a <code>PrintWriter</code> value
   * @param aHead a <code>String</code> value
   */
  public void oPrint(PrintWriter aPw, String aHead) {  //; (*5)
    cPrint(aPw, aHead);

    if (_sender == null)
      return;

    // Makes the header when outputting the information on the
    // input port.
    //
    String newHead = aHead + "  output < ";

    ((OhObject)_sender).oPrint(aPw, newHead);
  }

  // @end{methods} ---------------------------------------------

} // OhOutputFile
