/**
 * @(#) OhPort; class to express a port
 *
 * <a href="OhObject.html">OhObject</a>
 */

package ohymosj;

import java.util.*;
import java.io.*;

/**
 * OhPort is the base class from which all port classes are
 * derived. <p>
 *
 * We call an object of class OhPort or classes inherited from
 * class Port a <b>port</b>. <br>
 * In ohymosj, a port is used for
 * sending and/or receiving datapacks. <p>
 *
 * Basic member functions which a port should have are declared
 * as abstract member functions. <p>
 *
 * Part of those basic functions are declared as pure virtual.
 *
 * <japanese>
 * <p>
 * {^[qf̃NX <p>
 *
 * [q쐬邽߂̊ƂȂ钊ۃNXłB <p>
 *
 * NX OhPort ́ASĂ̒[qf̊NXłB
 * [qׂ{Iȃ\bh`Ă܂B
 * \bḧꕔ́ÃNXł͒`^Aۃ\bh
 * ȂĂ܂B
 * </japanese>
 *
 * @author satoh, shiiba
 * @version $Revision: 1.47 $ $Date: 2004/01/09 01:36:27 $
 */
public abstract class OhPort extends OhObject 
{
  // begin{abstract methods} ---------------------------------------------

  /**
   * Returns the string which shows the type of datapacks
   * treated by this port. <p>
   *
   * Because this function is declared as abstract, in a derived
   * class of class OhPort the
   * definition of this function must be given by a programmer.
   *
   * <japanese>
   * <p>
   * f[^̌^ԂB<p>
   *
   * ۃ\bhłB<p>
   * 
   * ̃\bh̎́A
   * OhSendDouble,
   * OhInputFileDouble,
   * OhRelayDouble,
   * OhReceiveDouble,
   * OhOutputFileDouble
   * ŗ^邱Ƃɂ܂B
   * int, long, float ɂĂlłB<p>
   *
   * Ⴆ΁AMf[^ int Ȃ "int" Ԃ悤ɒ`܂B
   * </japanese>
   *
   * @return a <code>String</code> value
   */
  public abstract String dataType();

  
  /**
   * Confirm whether all connection work has been done
   * successfully or not. <p>
   *
   * In class OhPort, which is an
   * abstract class, method areYouReady() is an abstract method.
   *
   *
   * <japanese>
   * vZƑO̊mF <p>
   *
   * ۃ\bhłB <p>
   *
   * ڑƌAvZƑO̊mFɗp܂B <p>
   *
   * ڑƂIAvZƂɐiނƂłꍇ true ԂA
   * ڑƂIĂȂꍇ false Ԃ܂B
   *
   * <pre>
   * RETURN:
   *   Etrue   ڑƂIĂB
   *   Efalse  ڑƂIĂȂB
   * </pre>
   * </japanese>
   *
   * @return a <code>boolean</code> value
   */
  public abstract boolean areYouReady();

  
  // 1. function for acquiring datapacks.
  
  /**
   * Abstract method.
   * Makes data pack array.
   *
   * <japanese>
   * <p>
   * f|^pbNz̍쐬 <p>
   *
   * ۃ\bhłB<p>
   *
   * ۂɂ́A_dataPackArray ̗vf _numberOfDataPacks ɐ
   * 肷邾łB
   * </japanese>
   *
   */
  public abstract void setDataPackArray();


  /**
   * Inquire whether the invoking port can get data from
   * <b>aTFrom</b> to <b>aTTo</b>.
   *
   * <japanese>
   * 擾\f[^pbN̊mF <p>
   *
   * ۃ\bhłB<p>
   *
   * canYouGetData(OhTimePointImmutable aTFrom, OhTimePointImmutable aTTo) ,
   * aTFrom ȑOAaTTo ȍ~܂ł̃f[^pbN̒[q
   * 擾\Ȃ΁Atrue ԂAs\Ȃ false Ԃ悤
   * ehNXŒ`܂.
   *
   * <pre>
   * ARGUMENT:
   *   EaTFrom    aTFrom ȑÕf[^pbN擾\ł邱ƂmF.
   *   EaTTo      aTTo ȍ~܂ł̃f[^pbN擾\ł邱ƂmF.
   * RETURN:
   *   EYES       aTFrom ȑO aTTo ȍ~܂ł̃f[^pbN擾\ȏꍇ.
   *               t ɂf[^擾\ȏꍇ.
   *   ENO        aTFrom ȑO aTTo ȍ~܂ł̃f[^pbN擾s\ȏꍇ.
   * </pre>
   * </japanese>
   *
   * @param aTFrom an <code>OhTimePointImmutable</code> value
   * @param aTTo an <code>OhTimePointImmutable</code> value
   * @return a <code>boolean</code> value
   */
  public abstract boolean canYouGetData(
    OhTimePointImmutable aTFrom, OhTimePointImmutable aTTo);
  

  /**
   * Gets a datapack array.
   *
   * <japanese>
   * <p>
   * f[^pbNz̎擾<p>
   *
   * hNXɂāAf[^pbNz擾郁\bh
   * `܂BaDpArray ɂ̓f[^pbNzi[AaNumOfDp 
   * ɂ́AaDpArray ̗vfi[܂B
   * </japanese>
   *
   * @param aDpArray an <code>OhArrayList[]</code> value
   * @param aNumOfDp an <code>int[]</code> value
   */
  public abstract void getDataPackArray(
    OhArrayList[] aDpArray, int[] aNumOfDp);


  /**
   * Returns the oldest data time.
   *
   * @return an <code>OhTimePoint</code> value
   */
  public abstract OhTimePointImmutable getOldestDataTime();

  /**
   * Returns the newest data time.
   *
   * @return an <code>OhTimePoint</code> value
   */
  public abstract OhTimePointImmutable getNewestDataTime();

  /**
   * Returns the number of data packs.
   *
   * @return an <code>int</code> value
   */
  public abstract int getNumberOfDataPacks();


  /**
   * Gets datapacks by specifying the datatime.
   * 
   * <japanese>
   * <p>
   * wɂf[^pbN̑Mi擾j <p>
   *
   * hNXɂāAw肵ăf[^pbN擾֐`܂B<p>
   *
   * aFormer ɂ́Af[^L^ aDataTime ȑO aDataTime ɍł߂^f[^pbN擾Ai[܂B<p>
   *
   * aLater ɂ́Af[^L^ aDataTime ȍ~ aDataTime ɍł߂^f[^pbN擾Ai[܂B<p>
   *
   * aFlag ͎擾̃[hI邽߂̃IvVŁAL̒ʂłB
   *
   * <pre>
   *   E"b"       aFormer, aLater ̗K擾B
   *               擾łȂꍇ́AƂ NULL ɐݒ肷B
   *   E"f"       aLater 擾łȂĂAaFormer ̂ݎ擾B
   *               aLater 擾łȂꍇANULL ɐݒ肷B
   *   E"l"       aFormer 擾łȂĂAaLater ̂ݎ擾B
   *               aFormer 擾łȂꍇANULL ɐݒ肷B
   *   Ȇ    aFormer, aLater ̂ȂƂ擾B
   *               擾łȂ̂ NULL ɐݒ肷B
   * </pre>
   *
   * aFlag ̃ftHgĺA <b>b</b> łB<p>
   * ܂A  aLater nȂꍇ́A aFlag  f Ƃ aFormer ̂ݎ擾A
   *  aFormer nȂꍇ́A aFlag  l Ƃ aLater ̂ݎ擾܂B<p>
   *
   * f[^L^x aDataTime łf[^pbN^Ăꍇ́A
   * aFormer, aLater Ƃi[܂B<p>
   *
   * 𖞂f[^pbN擾łȂꍇAaFormer, aLater  null ɐݒ肳܂B<p>
   *
   * f[^擾łꍇ́Atrue Ԃ܂B擾f[^́A
   * aFormer[0], aLater[0] Ɏ^܂B
   * </japanese>
   *
   * @param aFormer an <code>Object[]</code> value
   * @param aDataTime an <code>OhTimePoint</code> value
   * @param aLater an <code>Object[]</code> value
   * @param aMode a <code>char</code> value
   * @return a <code>boolean</code> value
   */
  public abstract boolean getDataPack(
    Object[] aFormer, OhTimePoint aDataTime, Object[] aLater,
    char aMode);

  /**
   * Returns the time from which on datapacks are required.
   * <p>
   *
   * <japanese>
   * f[^KvԂB<p>
   *
   * getNecessaryTimeFrom f[^L^f[^pbŃA
   * OhPort.reduceDataPackStock()
   * ɂ܂B
   * </japanese>
   *
   * @return an <code>OhTimePoint</code> value
   */
  public abstract OhTimePointImmutable getNecessaryTimeFrom();

  /**
   * Returns the time up to which datapacks are required.
   *
   * <japanese>
   * <p>
   * f[^KvԂB
   * </japanese>
   *
   * @return an <code>OhTimePoint</code> value
   */
  public abstract OhTimePointImmutable getNecessaryTimeTo();
 

  /**
   * Removes the unnecessary datapacks.
   *
   * <japanese>
   * svf[^pbN̏ <p>
   *
   * ̒[qŎ擾\ȃf[^pbN̂A
   * getNecessaryTimeFrom() f[^L^f[^pbN
   * 悤AhNXŒ`܂B
   *
   * </japanese>
   *
   */
  public abstract void reduceDataPackStock();
  
  // end{abstract methods} -----------------------------------------------


  // @begin{initialization} ----------------------------------------------

  /**
   * Initialization.
   *
   * @param aPortName a <code>String</code> value
   * @param aPortNum an <code>int</code> value
   * @exception OhError if an error occurs
   */
  public void init(String aPortName, int aPortNum) throws OhError {
    super.init(aPortName, aPortNum);
  }

  // @end{initialization} ------------------------------------------------


  // begin{constructors} -------------------------------------------------
  
  /**
   * Creates a new default <code>OhPort</code> instance.
   *
   */
  public OhPort() {
    super();
    _className = "OhPort";    //; _className = "Port"; is NG.
  }

  /**
   * Creates a new <code>OhPort</code> instance.
   *
   * @param aPortName a <code>String</code> value
   * @param aPortNumber an <code>int</code> value
   * @exception OhError if an error occurs
   */
  public OhPort(String aPortName, int aPortNumber) {
    super();
    _className = "OhPort";      //; "Port" is NG.
    _objectName = aPortName;
    _objectNumber = aPortNumber;
  }

  // end{constructors} ---------------------------------------------------


  // begin{methods} ------------------------------------------------------
  
  /**
   * Set the sender. There is only one sender for each
   * receiver. If it fails to set the sender, it throws an
   * OhError object.
   * <p>
   *
   * Although a default definition is given in the Port class, a
   * programmer must give his own definitions in derived
   * classes.
   *
   * <japanese>
   * <p>
   * M[q̓o^ <p>
   *
   * ̃NX̔hNXł́AM[q̃AhX aSender 
   * ƂĎ󂯂ƂA̒[q̓̃tB[hϐɋL
   * ܂B
   * <p>
   *
   * ̃\bh́AG[bZ[Wo͌ false Ԃ܂B
   * hNXŁAsetSender() Ē`Ȃꍇ́A
   * Adefault ̒`ɂȂ܂B
   * </japanese>
   *
   * @param aSender an <code>OhPort</code> value
   * @exception OhError if an error occurs
   */
  public void setSender(OhPort aSender) throws OhError {
    StringWriter buffer = new StringWriter();
    PrintWriter  out    = new PrintWriter(buffer);
    oPrint(out, " > ");

    if(aSender != null)
      (aSender).oPrint(out, " >   sender > ");

    String str = buffer.toString();
    out.close();

    throw new OhError(
      "This Port cannot set sender. " + NEWLINE + str);
  }
  

  /**
   * Register a receiver of this port. There may be more than
   * one receiver for a sender port. When the receiver cannot be
   * registered, this function throws an OhError object. <p>
   *
   * Although a default definition is given here, a programmer
   * must give his own definitions in derived classes.<p>
   *
   * <japanese>
   * M[q̓o^ <p>
   *
   * ̃NX̔hNXł́AObject ^ɃLXgꂽM
   * [q̃AhX aReceiver ƂĎ󂯂ƂAaReceiver 
   * MƂēo^ĂBo^łȂ OhError  
   * throw ĂB<p>
   *
   * ̃\bh́AG[bZ[Wo͌ false Ԃ܂B
   * hNXŁAregisterReceiver() Ē`Ȃꍇ́A
   * Adefault ̒`ɂȂ܂B
   * </japanese>
   *
   * @param aReceiver an <code>Object</code> value
   * @return a <code>boolean</code> value
   * @exception OhError if an error occurs
   */
  public void registerReceiver(OhObject aReceiver) throws OhError {    //; (*2)
    StringWriter buffer = new StringWriter();
    PrintWriter  out    = new PrintWriter(buffer);

    oPrint(out, " > ");

    if(aReceiver != null)
      aReceiver.oPrint(out, " >   receiver > ");

    String str = buffer.toString();
    out.close();
    
    throw new OhError(
      "This Port cannot register receiver. " + NEWLINE + str);
  }

  /**
   * Finish registration of receivers and count the number of
   * receivers. <p>
   *
   * In a derived class, do work which finishes connection
   * operations.
   *
   * <japanese>
   * <p>
   * [q̐ڑƂ̏Iуf[^MJn̂߂̏
   *  <p>
   *
   * ̃NX̔hNXŁA[q̐ڑƂ̏IƂ
   * d̏sȂ悤`܂B<p>
   *
   * Ɏs OhError IuWFNg𑗏o܂B
   * <p>
   *
   * OhReceivePort ł́AڑƌsȂׂ͓ɂ܂
   * B
   * <p>
   *
   * OhSendPort ł́A̎M[qɃf[^𑗐M邱Ƃ
   * ̂ŁAM port Xg _bufferList Ɏ^Ă
   * BŏIIɂ́AXgɓo^ꂽM port z 
   * _receiversArray Ɉڂւ܂B̍Ƃ 
   * OhSenderPort.canSwitch() Ŏs܂iOhSendPort ́A
   * OhSenderPort ̔hNXjB<p>
   *
   * ̌AڑIāAvZn߂܂B
   * ̎A_dataPackArray ̏܂B
   * Ȃ킿A_numberOfDataPacks ̒l -1  0 ɐݒ肵܂B
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public void doSwitch() throws OhError {
    StringWriter buffer = new StringWriter();
    PrintWriter out     = new PrintWriter(buffer);

    oPrint(out, " > ");

    String str = buffer.toString();
    out.close();

    throw new OhError(
      "This Port cannot switch. " + NEWLINE + str);
  }

   /**
    * Request of calculation. Transfer the request to a element
    * which does actual caluculation work or ports which send
    * datapack to the current port.
    *
    *
    * <japanese>
    * vZv󂯂֐łB
    * z֐łāANX <code>OhPort</code> ł́A
    * Ȃ悤ɒ`Ă܂B<p>
    *
    * <code>OhPort</code> ̔hNXł̂悤ɒ`
    * ܂B
    *
    * <ol>
    * <li> OhSendPort.calReq() ́A
    *       OhSendPort LĂvfɌvZv]܂B
    * <li> OhRelayPort.calReq() ́A
    *       OhRelayPort Ƀf[^𑗐MĂ[qɌvZv]܂B
    * <li> OhReceivePort.calReq() ́A
    *       OhReceivePort Ƀf[^𑗐MĂ[qɌvZv]܂B
    * </ol>
    * </japanese>
    *
    * @exception OhCannotGetDataError if an error occurs
    * @exception OhError if an error occurs
    */
   public void calReq() throws OhCannotGetDataError, OhError {}

  // Function for acquiring datapacks.
  
  /**
   * Inquire whether the invoking port can get data at 
   * OhTimePoint <b>aT</b>.
   *
   * <japanese>
   * <p>
   * ̃\bh́A<b>aT</b> ɂf[^擾\mF
   * \bhłB̃NX́Ã\bhĂяoĂ
   * ܂B
   * </japanese>
   *
   * @param aT an <code>OhTimePoint</code> value
   * @return a <code>boolean</code> value
   */
  public boolean canYouGetData(OhTimePoint aT) {
    return canYouGetData(aT, aT); // Abstract method call.
  }

  /**
   * Returns datapacks by specifying datatime.
   *
   * <japanese>
   * <p>
   * wɂf[^pbN̑Mi擾j <p>
   *
   * w肵ăf[^pbN擾郁\bh`܂B
   * ̃NX́A̒ۃ\bhĂяoĂ܂B<p>
   *
   * aFormer ɂ́Af[^L^ _aDataTime ȑO aDataTime 
   * ɍł߂^f[^pbN擾Ai[܂B<p>
   *
   * aFlag ͎擾̃[hI邽߂̃IvVŁAL̒
   * łB
   *
   * f[^L^x aDataTime łf[^pbN^
   * Ăꍇ́AaFormer ɂi[܂B<p>
   *
   * 𖞂f[^pbN擾łȂꍇAaFormer  null 
   * ɐݒ肳܂B<p>
   *
   * f[^擾łꍇ́A aFormer[0] Ɏ^Atrue 
   * Ԃ܂B
   * <p>
   * </japanese>
   *
   * @param aFormer an <code>Object[]</code> value
   * @param aDataTime an <code>OhTimePoint</code> value
   * @return a <code>boolean</code> value
   */
  public boolean getDataPack(Object[] aFormer,
    OhTimePoint aDataTime) {
    Object[] dummy = {null};
    return getDataPack(aFormer, aDataTime, dummy, 'f');
      // Abstract method call.
  }

  /**
   * Returns datapacks by specifying datatime.
   *
   * <japanese>
   * <p>
   * wɂf[^pbN̑Mi擾j <p>
   *
   * hNXɂāAw肵ăf[^pbN擾郁
   * \bh`܂B<p> ̃NX́A̒ۃ\bh
   * ĂяoĂ܂B<p>
   *
   * aLater ɂ́Af[^L^ aDataTime ȍ~ aDataTime 
   * ł߂^f[^pbN擾Ai[܂B<p>
   *
   * f[^L^x aDataTime łf[^pbN^
   * Ăꍇ́AaLater ɂi[܂B<p>
   *
   * 𖞂f[^pbN擾łȂꍇAaLater  null 
   * ɐݒ肳܂B<p>
   *
   * f[^擾łꍇ́A aLator[0] Ɏ^Atrue 
   * Ԃ܂B
   * </japanese>
   *
   * @param aDataTime an <code>OhTimePoint</code> value
   * @param aLater an <code>Object[]</code> value
   * @return a <code>boolean</code> value
   */
  public boolean getDataPack(OhTimePoint aDataTime,
    Object[] aLater) {
    Object[] dummy = {null};
    return getDataPack(dummy, aDataTime, aLater, 'l');        // Abstract method call.
  }

  // 2. Object information output function.

  /**
   * Prints the information about this OhPort object.
   *
   * <japanese>
   * <p>
   * IuWFNg̏o
   * </japanese>
   *
   * @param aPw a <code>PrintWriter</code> value
   * @param aHead a <code>String</code> value
   */
  public void cPrint(PrintWriter aPw, String aHead) {
    aPw.print(aHead + _className + 
      " " + _objectName +
      " " + _objectNumber + NEWLINE);
  }

  /**
   * Prints the information about connection with this port.
   *
   * <japanese>
   * <p>
   * ڑ̏o <p>
   *
   * hNXŐڑ\悤`܂B
   * ̃NXł́AG[bZ[Wo͂܂B
   * </japanese>
   *
   * @param aPw a <code>PrintWriter</code> value
   * @param aHead a <code>String</code> value
   */
  public void printConnection(PrintWriter aPw, String aHead) {
    oPrint(aPw, aHead);
    aPw.print(aHead + "    This port is not connected." + NEWLINE);
  }

  /**
   * Creates a port which has the class name specified by
   * the argument.<p>
   *
   * <japanese>
   * Ŏw肳ꂽNX OhPort IuWFNg쐬
   * ĕԂ܂Bw肳ꂽNX́ApbP[W ohymosj ɑ
   * ̂Ƃ܂B<p>
   *
   * . OhPort p = OhPort.create("OhSendDouble");
   * </japanese>
   *
   * @param aClassName a <code>String</code> value
   * @return an <code>OhPort</code> value
   */
  public static OhPort create(String aClassName) {
    OhPort port = null;
    try {
      port = (OhPort)Class.forName("ohymosj." + aClassName).newInstance();
    }
    catch (Exception e) {
      e.printStackTrace();
    }
    return port;
  }
  // end{methods} --------------------------------------------------------

} // OhPort
