/**
 * @(#) OhReceivePort; class to express a receive port
 *
 * <a href="OhObject.html">OhObject</a>
 * <-- <a href="OhPort.html">OhPort</a>
 */

package ohymosj;

import java.util.*;
import java.io.*;

/**
 * Class for models of receive ports. Derived from the class
 * OhPort. <p>
 * 
 * This class OhReceivePort has a instance variable for storing
 * a sender which sends datapacks to the current receive
 * port. <p> Member functions for
 *
 * <ul>
 *  <li> connecting the current receive port to a corresponding
 *       sender
 *  <li> receiving datapacks from the sender
 * </ul>
 *
 * are defined. <p>
 * 
 * NOTE: <p>
 * 
 * Only one send port can be connected to an object of class
 * OhReceivePort.
 *
 * <japanese>
 * <p>
 * {^M[qf̃NX <p>
 *
 * ۃNXłB<p>
 *
 * M[qɂ͑M[qPȂƂł܂B<p>
 *
 * NX OhReceivePort ł́A
 *
 * <ul>

 *  <li> f[^pbNM[q : _sender
 * </ul>
 *
 * LA
 *
 * <ul>
 *  <li> M[qƂ̐ڑ@\
 *  <li> _sender ̃f[^pbNM@\
 * </ul>
 *
 * `܂B
 *
 * </japanese>
 *
 * @author satoh, shiiba
 * @version $Revision: 1.37 $ $Date: 2004/01/09 02:23:10 $
 */
public abstract class OhReceivePort extends OhPort
{
  // @begin {instance variable or static variable} -------------

  /**
   * A sender which sends datapacks to the current object.
   *
   * <japanese>
   * <p>
   * OhSenderPort ^ɃLXgꂽM[q
   * </japanese>
   *
   */
  protected OhSenderPort _sender;
  
  // @end {instance variable or static variable} ---------------
  
  
  // @begin {abstract methods} ---------------------------------

  // 1. connection function.

  /**
   * Returns a string which expresses the type of datapacks
   * received by the current object. This is a abstract
   * method. This function must be defined in derived classes.
   *
   * <japanese>
   * <p>
   * Mf[^pbN^\Ԃ <p>
   *
   * NX OhPort ̒ۃ\bhłB<p>
   *
   *  OhRelayPort łۃ\bhƂĂ܂B
   * OhRelayPort ̔hNXŎMf[^pbN^\
   * Ԃ悤ɕK`ĉB<p>
   *
   * Ⴆ΁AMf[^pbN^ "double" ^łhNX
   * ́A
   *
   * <pre>
   *  public String dataType() {
   *    return "double";
   *  }
   * </pre>
   *
   * ̂悤ɒ`܂B<p>
   *
   * ڑɂ́A̒[qƑM[q̃f[^pbN^v
   * ƂmF܂B
   * </japanese>
   *
   * @return a <code>String</code> value
   */
  public abstract String dataType();

  
  // @end {abstract methods} -----------------------------------


  // @begin{initialization} ------------------------------------

  /**
   * Initialization with object name <b>aPortName</b> and object
   * number <b>aPortNum</b>.
   *
   * @param aPortName a <code>String</code> value
   * @param aPortNum an <code>int</code> value
   */
  public void init(String aPortName, int aPortNum)
    throws OhError {
    super.init(aPortName, aPortNum);
  }

  // @end{initialization} --------------------------------------
  
  
  // @begin {constructors} -------------------------------------

  /**
   * Creates a new default <code>OhReceivePort</code> instance.
   *
   */
  public OhReceivePort() {
    super();
    _sender = null;
    _className = "OhReceivePort";
  }


  /**
   * Creates a new <code>OhReceivePort</code> instance.
   *
   * @param aPortName a <code>String</code> value
   * @param aPortNumber an <code>int</code> value
   * @exception OhError if an error occurs
   */
  public OhReceivePort(String aPortName, int aPortNumber) {
    super();
    _className = "OhReceivePort";    
    _sender = null;
    _objectName = aPortName;
    _objectNumber = aPortNumber;
  }

  // @end {constructors} ---------------------------------------

  // @begin {methods} ------------------------------------------

  // 1. connection function.

  /**
   * Set "_sender" to "aSender" and return "true".
   * <p>
   * When one of the following cases occurs
   * <ul>
   *  <li> "aSender" is NULL.
   *  <li> "_sender" has been set already.
   *  <li> Data pack types between the current object and the
   *       sender are different. 
   * </ul>
   * this method throws an OhError object.
   *
   * <japanese>
   * <p>
   * M[q̐ݒ <p>
   *
   * NX OhPort ̒ۃ\bhłB̎M[qɃf[^
   * 𑗐MĂ[qւ̎Q _sender  aSender ɐݒ肵܂B
   * <p>
   *
   * L̏ꍇ̓G[bZ[W OhError IuWFNg throw 
   * ܂B
   *
   * <ul>
   *  <li> aSender  null ̏ꍇB
   *  <li> ̒[q̑M[q _sender ݒς݂̏ꍇB
   *  <li> ̒[q aSender Ńf[^pbN^قȂꍇB
   * </ul>
   *
   * </japanese>
   *
   * @param aSender an <code>OhPort</code> value
   */
  public void setSender(OhPort aSender) throws OhError {
    if(aSender == null) {
      throw new IllegalArgumentException(
        "Cannot set \"sender\"." + oPrintString(" > "));
    }
    
    if(_sender != null) {
      StringWriter buffer = new StringWriter();
      PrintWriter  out    = new PrintWriter(buffer);
      _sender.cPrint(out, " >    connected to  : ");
      aSender.cPrint(out, " >    connecting to  : ");
      oPrint(out, " > ");
      String str = buffer.toString();
      out.close();
      throw new OhError(
        "\"sender\" has already been set. Doubly setting?"
        + NEWLINE + str);
    }
    
    final String sendType    = ((OhPort)aSender).dataType();
    final String receiveType = dataType();

    if(!(sendType.equals(receiveType))) {
      throw new OhError(
        "'dataPackType' of this port and \"sendType\"(arg) are different"
        + NEWLINE
        + "sendType is " + sendType
        + " while receiveType is " + receiveType
        + NEWLINE
        + aSender.oPrintString(" >    s > ")
        + oPrintString(" > ")
      );
     }
    
     _sender = (OhSenderPort)aSender;
  }


  /**
   * Because no corresponding work is needed in the
   * <code>OhReceivePort</code> class, the switch method of the
   * <code>OhReceivePort</code> does nothing.
   * "<p>
   *
   * Note: <p> Switch of the
   * <code>OhSenderPort</code> class is defined differently.
   *
   * <japanese>
   * ڑƌ̏ <p>
   *
   * OhReceivePort ł͐ڑƌɍsȂׂ͓ɂ܂B
   * </japanese>
   *
   * @return a <code>boolean</code> value
   */
  public void doSwitch() {
  }

  /**
   * Answers whether the current receive port has been connected
   * to some sender or not.
   *
   * <japanese>
   * <p>
   * vZƑO̊mF <p>
   *
   * NX OhPort ̒ۃ\bhĂ܂B<br>
   * f[^pbNM邱ƂłԂł邱ƂmF܂B
   * ܂A_sender ݒ肳Ă邩`FbN܂B
   * _sender ݒ肳Ă΁A_sender f[^pbNM
   * 邱Ƃł܂B
   * </japanese>
   *
   * @return a <code>boolean</code> value
   */
  public boolean areYouReady() {
    return (_sender != null) ? true : false;
  }

  // 2. function for receiving datapacks.

  /**
   * Orders the sender to create a datapack array.
   *
   * <japanese>
   * f|^pbNz̍쐬 <p>
   *
   * NX OhSenderPort ̒ۃ\bhłB
   * _sender Ƀf[^pbNz쐬悤߂܂B
   * </japanese>
   *
   */
  public void setDataPackArray() {
    _sender.setDataPackArray();
  }


  /**
   * Returns the data time of the oldest data pack in the data
   * stored in the sender.
   *
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePointImmutable getOldestDataTime() {
    return _sender.getOldestDataTime();
  }

  /**
   * Returns the data time of the newest data pack in the data
   * stored in the sender.
   *
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePointImmutable getNewestDataTime() {
    return _sender.getNewestDataTime();
  }
  

  /**
   * Returns the number of datapacks stored in the sender.
   *
   * @return an <code>int</code> value
   */
  public int getNumberOfDataPacks() {
    return _sender.getNumberOfDataPacks();
  }
  

  /**
   * Tells whether the data from "aTFrom" to "aTTo" can be
   * obtained or not. <p> Accident: <p> "_sender" is "null" or
   * not OhPort type.
   *
   * <japanese>
   * 擾\f[^pbN̊mF <p>
   *
   * NX OhPort ̒ۃ\bhłB<p>
   *
   * _sender ɖ₢킹 aTFrom ȑO aTTo ȍ~܂ł̃f[
   * ^pbNM\mF܂BM\Ȃ true ԂA
   * Ms\Ȃ false Ԃ܂B<p>
   *
   * ̎M[q _owner ŌvZ̊mFƂɗp܂B
   *
   * </japanese>
   *
   * @param aTFrom an <code>OhTimePoint</code> value
   * @param aTTo an <code>OhTimePoint</code> value
   * @return a <code>boolean</code> value
   */
  public boolean canYouGetData(OhTimePointImmutable aTFrom,
    OhTimePointImmutable aTTo) {
    return _sender.canYouGetData(aTFrom, aTTo);
  }


  /**
   * Tells whether the data at "aTFromTo" can be obtained or not. <p>
   * 
   * Accident: <p>
   * 
   * "_sender" is "null" or not OhPort type.
   *
   * @param aTFromTo an <code>OhTimePoint</code> value
   * @return a <code>boolean</code> value
   */
  public boolean canYouGetData(OhTimePoint aTFromTo) {
    return _sender.canYouGetData(aTFromTo, aTFromTo);
  }


  /**
   * Get a data pack array from the sender. After the execution
   * of this function, "aDpArray" stores the data pack array and
   * "aNumOfDp" stores the dimention of "aDpArray". <p>
   *
   * Accident: <p>
   * 
   * "_sender" is "null" or not OhPort type.
   *
   * <japanese>
   * f[^pbNz̎M <p>
   *
   * NX OhPort ̒ۃ\bhłB<p>
   *
   * M[q _sender f[^pbNzM܂B<p>
   *
   *  aDpArray ̑vfɂ Object ^ɃLXgꂽ
   * OhDataPack ^ϐ̔z񂪊i[A
   * aNumOfDp ̑vfɂ aDpArray ̗vfi[܂B<p>
   *
   * aDpArray ̎Mf[^pbÑf[^L^AĂ邱Ƃ
   * OhSendPort.dataPackStock(OhDataPack dp)
   * ɂۏ؂܂B<p>
   *
   * Kvȃf[^pbNM\ۂ
   * canYouGetData(OhTimePoint, OhTimePoint) pĊmFł܂B<p>
   *
   * </japanese>
   *
   * @param aDpArray a <code>Vector</code> value
   * @param aNumOfDp an <code>int[]</code> value
   */
  public void getDataPackArray(OhArrayList[] aDpArray, int[] aNumOfDp) {
    _sender.getDataPackArray(aDpArray, aNumOfDp);
  }

  
  /**
   * Get a data pack at specified time "aDataTime".
   * The results are stored in "aFormer" and "aLater".
   *
   * <ul>
   *  <li> "aFormer" is the newest data before "aDataTime"
   *  <li> "aLater" is the pointer to the oldest data after "aDataTime"
   * </ul>
   *
   * When a data pack just at "aDataTime" exists, "aFormer" and "aLater" point to that data pack.
   * <p>
   * "aFlag" specifies the search option.
   *
   * <dl>
   *  <dt> "b"
   *  <dd>   Receive both of "former" and "later".
   *
   *  <dt> "f"
   *  <dd>   Receive "aFormer". Even if "aLater" cannot be obtained, the function returns "true".
   *
   *  <dt> "l"
   *  <dd>   Receive "aLater". Even if "aFormer" cannot be obtained, the function returns "true".
   *
   *  <dt> other character
   *  <dd>   Receive at lease one of "aFormer" and "aLater".
   * </dl>
   *
   * <japanese>
   * wɂf[^pbN̎M <p>
   *
   * NX OhSenderPort ̒ۃ\bhłB<p>
   *
   * w肵 _sender f[^pbNM܂B<p>
   *
   * aFormer ɂ́Af[^L^ aDataTime ȑO aDataTime ɍł߂^f[^pbNMAi[܂B<p>
   *
   * aLater ɂ́Af[^L^ aDataTime ȍ~ aDataTime ɍł߂^f[^pbNMAi[܂B<p>
   *
   * aFlag ͎擾̃[hI邽߂̃IvVŁAL̒ʂłB
   *
   * <pre>
   *   E"b"       aFormer, aLater ̗KMB
   *               MłȂꍇ́AƂ null ɐݒ肷B
   *   E"f"       aLater MłȂĂAaFormer ̂ݎMB
   *               aLater MłȂꍇAnull ɐݒ肷B
   *   E"l"       aFormer MłȂĂAaLater ̂ݎ擾B
   *               aFormer MłȂꍇAnull ɐݒ肷B
   *   Ȇ    aFormer, aLater ̂ȂƂ擾B
   *               MłȂ̂ null ɐݒ肷B
   * </pre>
   *
   * aFlag ̃ftHgĺA "b" łB<p> ܂A  aLater 
   * nȂꍇ́A aFlag  f Ƃ aFormer ̂ݎMA
   *  aFormer nȂꍇ́A aFlag  l Ƃ aLater 
   * ݎM܂B<p>
   *
   * f[^L^x aDataTime łf[^pbN^
   * Ăꍇ́AaFormer, aLater Ƃi[܂B<p>
   *
   * 𖞂f[^pbN擾łȂꍇAaFormer, aLater 
   *  null ɐݒ肳܂B<p>
   *
   * f[^pbN擾łꍇ́Atrue Ԃ܂B
   *
   * <pre>
   * [ARGUMENT]
   *  EaFormer ---> aDataTime ȑO aDataTime ɍł߂^f[^pbNvfɊi[B
   *  EaDataTime -> w莞 [sec]
   *                 ̎ɍł߂f[^L^^f[^pbNMB
   *  EaLater ----> aDataTime ȍ~ aDataTime ɍł߂^f[^pbNvfɊi[B
   *  EaMode -----> f[^pbN̎Mw肷tOB
   *
   * [RETURN]
   *  Etrue --> w̃f[^pbNMłꍇB
   *  Efalse -> _dataPackArray  null ̏ꍇB
   *             w̃f[^pbNMłȂꍇB
   * </pre>
   * </japanese>
   *
   * @param aFormer an <code>Object[]</code> value
   * @param aDataTime an <code>OhTimePoint</code> value
   * @param aLater an <code>Object[]</code> value
   * @param aMode a <code>char</code> value
   * @return a <code>boolean</code> value
   */
  public boolean getDataPack(
    Object[] aFormer, OhTimePoint aDataTime,
    Object[] aLater, char aMode) {
    return _sender.getDataPack(aFormer, aDataTime, aLater, aMode);
  }

  /**
   * By all means, Get a data pack at specified time
   * "aDataTime". The results are stored in "aFormer" and
   * "aLater".
   *
   * <japanese>
   * wɂf[^pbN̎M <p>
   *
   * NX OhSenderPort ̒ۃ\bhłB<p>
   *
   * w肵 _sender f[^pbNM܂B<p>
   *
   * aFormer ɂ́Af[^L^ aDataTime ȑO aDataTime 
   * ɍł߂^f[^pbNMAi[܂B<p>
   *
   * aLater ɂ́Af[^L^ aDataTime ȍ~ aDataTime 
   * ł߂^f[^pbNMAi[܂B<p>
   *
   * aFormer, aLater ̗KM܂B
   * MłȂꍇ́AƂ null ɐݒ肳܂B
   *
   * f[^L^x aDataTime łf[^pbN^
   * Ăꍇ́AaFormer, aLater Ƃi[܂B<p>
   *
   * f[^pbN擾łꍇ́Atrue Ԃ܂B
   * </pre>
   * </japanese>
   *
   * @param aFormer an <code>Object[]</code> value
   * @param aDataTime an <code>OhTimePoint</code> value
   * @param aLater an <code>Object[]</code> value
   * @return a <code>boolean</code> value
   */
  public boolean getDataPack(
    Object[] aFormer, OhTimePoint aDataTime, Object[] aLater) {
    return _sender.getDataPack(aFormer, aDataTime, aLater, 'b');
  }


  /**
   * Returns the "NecessaryTimeFrom" (this) of the owner of the
   * current receive port. You can redefine it in derived
   * classes.
   *
   * <japanese>
   * <p>
   * f[^KvԂ <p>
   *
   * M̑M[q getNecessaryTimeFrom() ȍ~ 
   * getNecessaryTimeTo() ȑÕf[^pbNȂ悤Ɏw
   * 邽߂ɒ`܂B<p>
   *
   * _owner 
   * OhElement.getNecessaryTimeFrom()
   * Ԃ܂B
   * _owner  null ̏ꍇ
   * OhTimePoint.OhTimePointMin
   * Ԃ܂B
   *
   * </japanese>
   *
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePointImmutable getNecessaryTimeFrom() {
    return (_owner != null) ? ((OhElement)_owner).getNecessaryTimeFrom(this) : OhTimePoint.OhTimePointMin;
  }

  /**
   * Returns the "NecessaryTimeTo" (this) of the owner of the
   * current receive port. You can redefine it in derived
   * classes.
   *
   * <japanese>
   * <p>
   * f[^KvԂ <p>
   *
   * M̑M[q getNecessaryTimeFrom() ȍ~ 
   * getNecessaryTimeTo() ȑÕf[^pbNȂ悤Ɏw
   * 邽߂ɒ`܂B<p>
   *
   * _owner 
   * OhElement.getNecessaryTimeTo()
   * Ԃ܂B
   * _owner  null ̏ꍇ
   * OhTimePoint.OhTimePointMax
   * Ԃ܂B<p>
   *
   * </japanese>
   *
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePointImmutable getNecessaryTimeTo() {
    return (_owner != null) ? ((OhElement)_owner).getNecessaryTimeTo(this) : OhTimePoint.OhTimePointMax;
  }


  /**
   * Relay the calculation request to the sender.
   *
   * <japanese>
   * M[qւ̌vZv͑M[qɓ]B
   * </japanese>
   *
   * @exception OhCannotGetDataError if an error occurs
   * @exception OhError if an error occurs
   */
  public void calReq() throws OhCannotGetDataError, OhError {
    _sender.calReq();
  }


  /**
   * Remove unnecessary data packs from the memory.
   *
   * <japanese>
   * svf[^pbN̏ <p>
   *
   * M[q _sender ɕsvf[^pbN悤w
   * B<p>
   *
   * ̒p[q _owner ŌvZIɗpAsv
   * ܂B<p>
   *
   * </japanese>
   *
   */
  public void reduceDataPackStock() {
    _sender.reduceDataPackStock();
  }


  // 3. Information output function.

  /**
   * Print the information about the connection to "aPw" with
   * the string "aHead".
   *
   * <japanese>
   * <p>
   * ڑ̏óB
   * </japanese>
   *
   * @param aPw a <code>PrintWriter</code> value
   * @param aHead a <code>String</code> value
   */
  public void printConnection(PrintWriter aPw, String aHead) {    //; (*4)
    oPrint(aPw, aHead);
    if(_sender == null) {
      aPw.print(aHead + "\tNot connected." + NEWLINE);
      return;
    }
    String tmp = aHead;
    tmp = tmp + "Connected > ";
    _sender.oPrint(aPw, tmp);
  }

  // @end {methods} --------------------------------------------
  
} // OhReceivePort
