/**
 * @(#) OhRelayPort; class to express relay ports.
 */

package ohymosj;

import ohymosj.*;
import java.util.*;
import java.io.*;

/**
 * The OhRelayPort class is a model of relay ports. <p>
 *
 * The OhRelayPort has a instance variable "_sender".
 * "_sender" is used to store the sender
 * which sends data packs to the current port. <p>
 *
 * Member functions for
 *
 * <ul>
 *  <li> Connecting the current port with the port which sends
 *       data packs to the current port
 *  <li> Connecting the current port with the port which
 *       receives data packs from the current port
 * </ul>
 *
 * are defined. <p>
 *
 * Only one port can be connected to the port for receiving data
 * packs. More than one ports can be connected to the ports
 * which sending data packs.
 *
 * <japanese>
 * <p>
 * {^p[qf̃NX <p>
 *
 * ۃNXłB<p>
 *
 * NX OhRelayPort ł́A
 *
 * <ul>
 *  <li> f[^pbNM[q : _sender
 * </ul>
 *
 * LA
 *
 * <ul>
 *  <li> M[qƂ̐ڑ@\
 *  <li> M[qƂ̐ڑ@\
 * </ul>
 *
 * `܂B
 * M[qƂ́A̒p[qɃf[^𑗂Ă鑤̒[qłB
 * M[qƂ́A̒p[q瑗f[^󂯂Ƃ鑤
 * [qłB
 *
 * OhRelayPort ɐڑł鑗M[q͂PłB
 * OhRelayPort ɂ͕̑M[qڑł܂B
 *
 * </japanese>
 *
 * @author satoh, shiiba
 * @version $Revision: 1.38 $ $Date: 2004/01/09 02:13:36 $
 */
public abstract class OhRelayPort extends OhSenderPort
{
  // @begin {instance variable or static variable} -------------

  /**
   * The sender which sends data packs to this OhRelayPort
   * casted to type OhSenderPort.
   *
   * <japanese>
   * <p>
   * <a href="OhSenderPort.html">OhSenderPort</a> ^̑M[q
   * </japanese>
   *
   */
  protected OhSenderPort _sender;

  // @end {instance variable or static variable} ---------------
  
  
  // @begin {abstract methods} ---------------------------------

  /**
   * Returns the string which shows the type of data packs
   * which are treated by this <code>OhRelayPort</code>. <p>
   *
   * In derived classes of this class <code>OhRelayPort</code>,
   * this funstion should be defined so that it returnts the
   * datapack type string.
   *
   * <japanese>
   * <p>
   * pf[^pbN^\Ԃ <p>
   *
   * NX <a href="OhSenderPort.html">OhSenderPort</a> ̒ۃ\bhłB<p>
   *
   *  OhRelayPort łۃ\bhƂĂ܂B
   * OhRelayPort ̔hNXŒpf[^pbN^\
   * Ԃ悤ɕK`ĉB<p>
   *
   * Ⴆ΁Apf[^pbN^ "double" ^łhNXł́A
   *
   * <pre>
   *  public String dataaType() {
   *    return "double";
   *  }
   * </pre>
   *
   * ̂悤ɒ`܂B<p>
   *
   * ڑɂ́Aڑ[qƂ̒[q̃f[^pbN^v
   * 邱ƂmF܂B
   * </japanese>
   *
   * @return a <code>String</code> value
   */
  public abstract String dataType();

  // @end {abstract methods} -----------------------------------


  // @begin{initialization} ------------------------------------

  /**
   * Initialization.
   *
   * @param aPortName a <code>String</code> value
   * @param aPortNum an <code>int</code> value
   * @return a <code>boolean</code> value
   */
  public void init(String aPortName, int aPortNum)
    throws OhError {
    ((OhPort)this).init(aPortName, aPortNum);
  }

  // @end{initialization} --------------------------------------

  
  // @begin {constructors} -------------------------------------
  
  /**
   * Creates a new default <code>OhRelayPort</code> instance.
   *
   */
  public OhRelayPort() {
    super();
    _sender    = null;
    _className = "OhRelayPort";
  }

  
  /**
   * Creates a new <code>OhRelayPort</code> instance.
   *
   * @param aPortName a <code>String</code> value
   * @param aPortNum an <code>int</code> value
   */
  public OhRelayPort(String aPortName, int aPortNum) {
    super();
    _sender    = null;
    _className = "OhRelayPort";
    _objectName = aPortName;
    _objectNumber = aPortNum;
  }

  // @end {constructors} ---------------------------------------

  
  // @begin {methods} ------------------------------------------

  // 1. connection facility
  
  /**
   * Returns the data connection type of the "_sender".
   *
   * @return a <code>Vector</code> value
   */
  public Vector getDctVec() {
    return _sender.getDctVec();
  }


  /**
   * Sets the sender of this OhRelayPort to aSender.<p>
   *
   * For the following cases, this method throws an OhError
   * object.
   *
   * <ul>
   *  <li> aSender is null.
   *  <li> already the sender has been set.
   *  <li> the type of datapacks which "aSender" treats is
   *       different from

   *       that of datapacks which this ports treats.
   * </ul>
   *
   * <japanese>
   * <p>
   * M[q̐ݒ <p>
   *
   * M[q _sender  aSender ɐݒ肵܂B
   * <p>
   *
   * L̏ꍇ OhError IuWFNg throw ܂B
   *
   * <ul>
   *  <li> aSender  null ̏ꍇB
   *  <li> ̒[q̑M[q _sender ݒς݂̏ꍇB
   *  <li> ̒[q aSender Ńf[^pbN^قȂꍇB
   * </ul>
   * </pre>
   * </japanese>
   *
   * @param aSender an <code>OhPort</code> value
   * @exception OhError if an error occurs
   */
  public void setSender(OhPort aSender)
    throws OhError {
    if (aSender == null) {
      throw new IllegalArgumentException(
        "cannot set sender." + oPrintString(" > "));
    }

    if (_sender != null) {
      StringWriter buffer = new StringWriter();
      PrintWriter  out    = new PrintWriter(buffer);
      _sender.cPrint(out, " >    connected to  : ");
      aSender.cPrint(out, " >    connecting to  : ");
      String str = buffer.toString();
      out.close();
      throw new OhError("sender has already been set."
        + NEWLINE +
        str);
    }

    // Gets the datapack type which this port receives, and the
    // datapack type which "aSender" receives.
    //
    String sType = aSender.dataType();
    String rType = dataType();

    if (!sType.equals(rType)) {
      throw new OhError(
        "data pack type of this port and that of " +
        "\"aSender\" are different." + NEWLINE +
        aSender.oPrintString(" > s > ") +
        oPrintString(" > ")
        );
    }

    _sender = (OhSenderPort)aSender;
  }
      

  /**
   * Reports whether this OhRelayPort is ready or not. When
   * "_sender" and "_numberOfReceivers" have been defined, it
   * returns true, otherwise it returns false.
   *
   * <japanese>
   * <p>
   * vZƑO̊mF <p>
   *
   * f[^pbN𒆌p\ȏԂł邱ƂmF܂B_sender 
   * ݒ肳Ă _sender f[^pbN𒆌pł܂B
   * _numberOfReceivers  0 ȏȂ _receiverArray ݒς
   * ŁAM[qɃf[^pbN𑗐Mł܂B<p>
   *
   * ̂߂ɂ́Ã\bhsOɁÃIuWFNg
   * AsetSender()  doSwitch() sĂKv܂B
   * </japanese>
   *
   * @return a <code>boolean</code> value
   */
  public boolean areYouReady() {
    return
      ((_sender != null) && (_numberOfReceivers >= 0)) ? true : false;
  }
  

  // 2. relay facility of data packs

  /**
   * Orders the sender to create a data pack array.
   *
   * <japanese>
   * f|^pbNz̍쐬 <p>
   *
   * NX OhSenderPort ̒ۃ\bhłB
   * _sender Ƀf[^pbNz쐬悤߂܂B
   * </japanese>
   *
   */
  public void setDataPackArray() {
    _sender.setDataPackArray();
  }


  /**
   * Returns the oldest data time of the data stocked in the
   * sender.
   *
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePointImmutable getOldestDataTime() {
    return _sender.getOldestDataTime();
  }

  /**
   * Returns the newest data time of the data stocked in the
   * sender.
   *
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePointImmutable getNewestDataTime() {
    return _sender.getNewestDataTime();
  }


  /**
   * Returns the number of data packs which stored in the
   * sender.
   *
   * @return an <code>int</code> value
   */
  public int getNumberOfDataPacks() {
    return _sender.getNumberOfDataPacks();
  }


  /**
   * Reports whether this OhRelayPort can get the data from
   * <b>aTFrom</b> to <b>aTTo</b>. <p>
   *
   * <japanese>
   * 擾\f[^pbN̊mF <p>
   *
   * NX OhSenderPort ̒ۃ\bhłB<p>
   *
   * _sender ɖ₢킹 aTFrom ȑO aTTo ȍ~܂ł̃f[
   * ^pbNM\mF܂BM\Ȃ true ԂA
   * Ms\Ȃ false Ԃ܂B<p>
   *
   * ̃\bh́AM[q canYouGetData(OhTimePoint,
   * OhTimePoint) Ăяo܂BɂAM[qŃf[
   * ^pbNMł邩mFł܂B<p>
   * </japanese>
   *
   * @param aTFrom an <code>OhTimePoint</code> value
   * @param aTTo an <code>OhTimePoint</code> value
   * @return a <code>boolean</code> value
   */
  public boolean canYouGetData(OhTimePointImmutable aTFrom,
    OhTimePointImmutable aTTo) {
    return _sender.canYouGetData(aTFrom, aTTo);
  }


  /**
   * The calculation request to a relay port is transmitted to a
   * send port.
   *
   * <japanese>
   * p[qւ̌vZv͑M[qɓ]B
   * </japanese>
   *
   * @exception OhCannotGetDataError if an error occurs
   * @exception OhError if an error occurs
   */
  public void calReq() throws OhCannotGetDataError, OhError {
    _sender.calReq();
  }


  /**
   * Get datapack array aDpArray (arg) and the number of
   * datapacks aNumOfDp (arg) from the sender.
   *
   * <ul>
   *  <li> [out] aDpArray  - datapack array
   *  <li> [out] aNumOfDp  - the number of datapacks
   * </ul>
   *
   * <japanese>
   * f[^pbNz̎M <p>
   *
   * NX OhSenderPort ̒ۃ\bhłB<p>
   *
   * M[q _sender f[^pbNzM܂B<p>
   *
   *  aDpArray ̑vfɂ Object ^ɃLXgꂽ
   * OhDataPack ^ϐ̔z񂪊i[A
   * aNumOfDp ̑vfɂ aDpArray ̗vfi[܂B<p>
   *
   * ̃\bh́AM[q getDataPackArray Ăяo
   * ܂BɂAM[qŃf[^pbNzMł
   * ܂B<p>
   *
   * </japanese>
   *
   * @param aDpArray an <code>OhArrayList[]</code> value
   * @param aNumOfDp an <code>int[]</code> value
   */
  public void getDataPackArray(OhArrayList[] aDpArray,
    int[] aNumOfDp) {
    _sender.getDataPackArray(aDpArray, aNumOfDp);
  }

  /**
   * Gets the datapack array.
   *
   * @return an <code>OhArrayList</code> value
   */
  public OhArrayList getDataPackArray() {
    return _sender.getDataPackArray();
  }


  /**
   * Get two datapacks <b>aFormer</b> and <b>aLater</b>. <br>
   * <b>aFormer</b> is the newest datapack before time <b>_dataTime</b>. <br>
   * <b>aLater</b>  is the oldest datapack after  time <b>_dataTime</b>. <br>
   * <b>aMode</b> specifies the option
   *
   * <ul>
   *  <li> "b" -- get both of <b>aFormer</b> and <b>aLater</b>
   *  <li> "f" -- get <b>aFormer</b> at least
   *  <li> "l" -- get <b>aLater</b> at least
   * </ul>
   *
   * When the datapack(s) specified by <b>aMode</b> can be got,
   * it returns "true", otherwise "false". <p>
   *
   * [out] aFormer --- pointer to a datapack
   * [in]  _dataTime - specified time
   * [out] aLater ---- pointer to a datapack
   * [in]  aMode ----- option to get pointers to datapacks
   *
   * <japanese>
   * wɂf[^pbN̎M <p>
   *
   * NX OhSenderPort ̒ۃ\bhłB<p>
   *
   * w肵 _sender f[^pbNM܂B<p>
   *
   * aFormer ɂ́Af[^L^ _dataTime ȑO _dataTime 
   * ɍł߂^f[^pbNMAi[܂B<p>
   *
   * aLater ɂ́Af[^L^ _dataTime ȍ~ _dataTime 
   * ł߂^f[^pbNMAi[܂B<p>
   *
   * aFlag ͎擾̃[hI邽߂̃IvVŁAL̒
   * łB
   *
   * <pre>
   *   E"b"       aFormer, aLater ̗KMB
   *               MłȂꍇ́AƂ null ɐݒ肷B
   *   E"f"       aLater MłȂĂAaFormer ̂ݎMB
   *               aLater MłȂꍇAnull ɐݒ肷B
   *   E"l"       aFormer MłȂĂAaLater ̂ݎ擾B
   *               aFormer MłȂꍇAnull ɐݒ肷B
   *   Ȇ    aFormer, aLater ̂ȂƂ擾B
   *               MłȂ̂ null ɐݒ肷B
   * </pre>
   *
   * aFlag ̃ftHgĺA "b" łB<p>
   *
   * ܂A  aLater nȂꍇ́A aFlag  f Ƃ 
   * aFormer ̂ݎMA aFormer nȂꍇ́A aFlag 
   *  l Ƃ aLater ̂ݎM܂B<p>
   *
   * f[^L^x _dataTime łf[^pbN^
   * Ăꍇ́AaFormer, aLater Ƃi[܂B<p>
   *
   * 𖞂f[^pbN擾łȂꍇAaFormer, aLater 
   *  null ɐݒ肳܂B<p>
   *
   * f[^pbN擾łꍇ́Atrue Ԃ܂B<p>
   *
   * ̃\bh͑M[q getDataPack(Object[],
   * OhTimePoint, Object[], char) Ăяo܂Bɂ
   * M[qŃf[^pbNMł܂B
   *
   * </japanese>
   *
   * @param aFormer an <code>Object[]</code> value
   * @param _dataTime an <code>OhTimePoint</code> value
   * @param aLater an <code>Object[]</code> value
   * @param aMode a <code>char</code> value
   * @return a <code>boolean</code> value
   */
  public boolean getDataPack(Object[] aFormer,
    OhTimePoint _dataTime, Object[] aLater, char aMode) {
    return
      _sender.getDataPack(aFormer, _dataTime, aLater, aMode);
  }


  /**
   * By all means, Get two datapacks <b>aFormer</b> and
   * <b>aLater</b>.
   *
   * <japanese>
   * wɂf[^pbN̎M <p>
   *
   * NX OhSenderPort ̒ۃ\bhłB<p>
   *
   * w肵 _sender f[^pbNM܂B<p>
   *
   * aFormer ɂ́Af[^L^ _dataTime ȑO _dataTime 
   * ɍł߂^f[^pbNMAi[܂B<p>
   *
   * aLater ɂ́Af[^L^ _dataTime ȍ~ _dataTime 
   * ł߂^f[^pbNMAi[܂B<p>
   *
   * aFormer, aLater ̗KM܂B
   * MłȂꍇ́AƂ null ɐݒ肳܂B
   *
   * f[^L^x _dataTime łf[^pbN^
   * Ăꍇ́AaFormer, aLater Ƃi[܂B<p>
   *
   * f[^pbN擾łꍇ́Atrue Ԃ܂B
   *
   * ̃\bh͑M[q getDataPack(Object[],
   * OhTimePoint, Object[], char) Ăяo܂Bɂ
   * M[qŃf[^pbNMł܂B
   *
   * <pre>
   * [ARGUMENT]
   *  EaFormer ---> _dataTime ȑO _dataTime ɍł߂^f[^pbNvfɊi[B
   *  E_dataTime -> w莞 [sec]
   *                 ̎ɍł߂f[^L^^f[^pbNMB
   *  EaLater ----> _dataTime ȍ~ _dataTime ɍł߂^f[^pbNvfɊi[B
   *
   * [RETURN]
   *  Etrue --> w̃f[^pbNMłꍇB
   *  Efalse -> _dataPackArray  null ̏ꍇB
   *             w̃f[^pbNMłȂꍇB
   * </pre>
   * </japanese>
   *
   * @param aFormer an <code>Object[]</code> value
   * @param _dataTime an <code>OhTimePoint</code> value
   * @param aLater an <code>Object[]</code> value
   * @return a <code>boolean</code> value
   */
  public boolean getDataPack(Object[] aFormer,
    OhTimePoint _dataTime, Object[] aLater) {
    return
      _sender.getDataPack(aFormer, _dataTime, aLater, 'b');
  }


  /**
   * Forwards the message "reduceDataPackStock()" to the sender.
   *
   * <japanese>
   * svf[^pbN̏ <p>
   *
   * M[q _sender ɕsvf[^pbN悤w
   * B<p>
   *
   * ̒p[q _owner ŌvZIɗpAsv
   * ܂B<p>
   *
   * ̃\bh͑M[q getDataPack(Object[],
   * OhTimePoint, Object[], char) Ăяo܂Bɂ
   * M[qŃf[^pbNMł܂B
   * </japanese>
   *
   */
  public void reduceDataPackStock() {
    _sender.reduceDataPackStock();
  }


  
  // 3. information output function.

  /**
   * Prints the information about connection of this
   * Relay_port. This function is used for debugging.
   *
   * <japanese>
   * <p>
   * ڑ̏óB
   * </japanese>
   *
   * @param aPw a <code>PrintWriter</code> value
   * @param aHead a <code>String</code> value
   */
  public void printConnection(PrintWriter aPw, String aHead) {
    oPrint(aPw, aHead);


    if (_sender == null) {
      aPw.print(aHead + "\tNot connected to sender." + NEWLINE);
      return;
    }
  
    StringBuffer sb = new StringBuffer("");
    if (aHead != null)
      sb.append(aHead);
    sb.append("Connected > ");
    String str = sb.toString();
    ((OhObject)_sender).oPrint(aPw, str);

    if (_numberOfReceivers <= 0) {
      aPw.print(aHead + "\tNot connected to receiver." + NEWLINE);
    }
    else {
      for (int k = 0; k < _numberOfReceivers; k++)
        ((OhObject)_receiverArray[k]).oPrint(aPw, str);
    }
  }

  /**
   * Prints the information about connection of this
   * Relay_port. This function is used for debugging.
   *
   * <japanese>
   * <p>
   * ڑ̏óB
   * </japanese>
   *
   * @param aPw a <code>PrintWriter</code> value
   */
  public void printConnection(PrintWriter aPw) {
    oPrint(aPw, "    ");

    if (_sender == null) {
      aPw.print("    \tNot connected to sender." + NEWLINE);
      return;
    }

    StringBuffer sb = new StringBuffer("    ");
    sb.append("Connected > ");
    String str = sb.toString();
    ((OhObject)_sender).oPrint(aPw, str);

    if (_numberOfReceivers <= 0) {
      aPw.print("    \tNot connected to receiver." + NEWLINE);
    }
    else {
      for (int k = 0; k < _numberOfReceivers; k++)
        ((OhObject)_receiverArray[k]).oPrint(aPw, str);
    }
  }

  // @end {methods} --------------------------------------------

} // OhRelayPort
