/**
 * @(#) Class to express time points
 */

package ohymosj;

import java.io.*;
import java.util.*;
import java.text.DecimalFormat;

/**
 * The OhTimePoint class is the class which expresses time
 * points. The OhTimePoint class espresses a time point by
 * <b>year</b>, <b>tsec</b>, and <b>frac</b>. <p>
 *
 * <b>tsec</b> is the time measured in seconds from the start
 * of the year. <b>frac</b> is the fraction part of the time
 * which is greater than or equal to -0.5 and less than 0.5.<p>
 *
 * <japanese>
 * <p>
 * OhTImePoint ́A\킷NXłBIɂ́A
 * N <b>year</b>, ̔N̓̌oߕb <b>tsec</b>A
 *  <b>frac</b> ŋL^܂B<b>frac</b> ́A-0.5 ȏA
 * 0.5 Ƃ܂B<p>
 *
 * Ⴆ΁A2003 N̔N 10.3 bo߂\킷 
 * OhTimePoint IuWFNgł́A<b>year</b>  2003, <b>tsec</b> 
 * 10, <b>frac</b> ́A0.3 łB
 * </japanese>
 *
 * @author satoh, shiiba
 * @version $Revision: 1.5 $ $Date: 2003/12/16 11:43:57 $
 */
final public class OhTimePoint
  implements Cloneable, Comparable, Serializable, OhTimePointImmutable {
  int _year;
  long _tsec;
  double _frac;
  long _oneYearInSec;
  static int[] nDay = { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };
  static int[] lDay = { 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

  /**
   * The index of the element of the calendar array which
   * contains the datum of year. The calendar array is obtained
   * by toCalendarArray()
   */
  public static final int YEAR = 0;
  /**
   * The index of the element of the calendar array which
   * contains the datum of month. The calendar array is obtained
   * by toCalendarArray()
   */
  public static final int MONTH = 1;
  /**
   * The index of the element of the calendar array which
   * contains the datum of day. The calendar array is obtained
   * by toCalendarArray()
   */
  public static final int DAY = 2;
  /**
   * The index of the element of the calendar array which
   * contains the datum of hour. The calendar array is obtained
   * by toCalendarArray()
   */
  public static final int HOUR = 3;
  /**
   * The index of the element of the calendar array which
   * contains the datum of munute. The calendar array is
   * obtained by toCalendarArray()
   */
  public static final int MINUTE = 4;
  /**
   * The index of the element of the calendar array which
   * contains the datum of second. The calendar array is
   * obtained by toCalendarArray()
   */
  public static final int SECOND = 5;
  
  /**
   * Creates a new default <code>OhTimePoint</code> instance.
   * The default time is the beginning of the year 0.
   *
   * <japanese>
   * <p>
   * w肵ȂRXgN^ł́A0N̎𐶐
   * ܂B
   * </japanese>
   */
  public OhTimePoint (){
    _year = 0;  _tsec = 0;  _frac = 0;
    _oneYearInSec = 31622400;
  }
  
  /**
   * Creates a new <code>OhTimePoint</code> instance.
   *
   * <japanese>
   * <p>
   * aYear N̏ aTsec b̎ OhTimePoint IuWFN
   * g𐶐܂B
   * </japanese>
   *
   * @param aYear an <code>int</code> value.
   * @param aTsec a <code>long</code> value. Time from the beginning
   * of the year.
   */
  public OhTimePoint (int aYear, long aTsec) {
    _year = aYear;
    if (_year == 0) { _oneYearInSec = 31622400; } // 366 * 24 * 3600
    else {
      _oneYearInSec = 31536000; // 365 * 24 * 3600
      if (isLeapYear(_year)) _oneYearInSec = 31622400;
    }
    if (aTsec < 0 || aTsec >= _oneYearInSec) {
      throw new IllegalArgumentException("illegal aTsec");
    }
    _tsec = aTsec;
    _frac = 0;
  }
  
  /**
   * Creates a new <code>OhTimePoint</code> instance.
   *
   * <japanese>
   * <p>
   * aYear N̏ aTsec + aFrac b̎ OhTimePoint I
   * uWFNg𐶐܂B
   * </japanese>
   * 
   * @param aYear an <code>int</code> value.
   * @param aTsec a <code>long</code> value. Time in seconds
   *        from the beginning.
   * @param aFrac an <code>double</code> value. aFrac must be
   *        greater than or equal to -0.5 and less than 0.5.
   */
  public OhTimePoint (int aYear, long aTsec, double aFrac) {
    _year = aYear;
    if (_year == 0) { _oneYearInSec = 31622400; } // 366 * 24 * 3600
    else {
      _oneYearInSec = 31536000; // 365 * 24 * 3600
      if (isLeapYear(_year)) _oneYearInSec = 31622400;
    }
    if (aTsec < 0 || aTsec >= _oneYearInSec) {
      throw new IllegalArgumentException("illegal aTsec");
    }
    _tsec = aTsec;
    _frac = aFrac;
  }
  
  /**
   * Creates a new <code>OhTimePoint</code> instance.
   *
   * <japanese>
   * <p>
   * aYear N aMonth  aDay  aHour  aMin  aSec bɑΉ
   * OhTimePoint IuWFNg𐶐܂B
   * </japanese>
   * 
   * @param aYear an <code>int</code> value. year ex. 2003
   * @param aMonth an <code>int</code> value. month from 1 to 12. 1 for January
   * @param aDay an <code>int</code> value. day 1 for the first day of
   * the month
   * @param aHour an <code>int</code> value. hour from 0 to 23
   * @param aMin an <code>int</code> value. min from 0 to 59
   * @param aSec an <code>int</code> value. second from 0 to 59
   */
  public OhTimePoint (int aYear, int aMonth, int aDay,
                      int aHour, int aMin, int aSec) {
    _year = aYear;
    boolean wLeapYearP = isLeapYear(aYear);
    _oneYearInSec = 31536000; // 365 * 24 * 3600
    if (wLeapYearP) _oneYearInSec += 86400;

    _tsec = 0;
    
    if (aMonth < 1 || aMonth > 12) {
      throw new IllegalArgumentException("illegal aMonth");
    }
    for (int i = 0; i < aMonth -1; i++) {
      _tsec += (wLeapYearP ? lDay[i] : nDay[i]) * 86400;
    }
    if (aDay < 1 ||
      aDay > (wLeapYearP ? lDay[aMonth-1] : nDay[aMonth-1])) {
      throw new IllegalArgumentException("illegal aDay");
    }
    _tsec += (aDay - 1) * 86400;
    if (aHour < 0 || aHour >= 24) {
      throw new IllegalArgumentException("illegal aHour");
    }
    _tsec += aHour * 3600;
    if (aMin < 0 || aMin >= 60) {
      throw new IllegalArgumentException("illegal aMin");
    }
    _tsec += aMin * 60;
    if (aSec < 0 || aSec >= 60) {
      throw new IllegalArgumentException("illegal aSec");
    }
    _tsec += aSec;
    _frac = 0.;
  }
    
  /**
   * Creates a new <code>OhTimePoint</code> instance.
   *
   * <japanese>
   * <p>
   * aYear N aMonth  aDay  aHour  aMin  aSec + aFrac b
   * ɑΉOhTimePoint IuWFNg𐶐܂B
   * </japanese>
   * 
   * @param aYear an <code>int</code> value. year ex. 2003
   * @param aMonth an <code>int</code> value. month from 1 to 12. 1 for January
   * @param aDay an <code>int</code> value. day 1 for the first day of
   * the month
   * @param aHour an <code>int</code> value. hour from 0 to 23
   * @param aMin an <code>int</code> value. min from 0 to 59
   * @param aSec an <code>int</code> value. second from 0 to 59
   * @param aFrac an <code>double</code> value. aFrac must be
   *        greater than or equal to -0.5 and less than 0.5.
   */
  public OhTimePoint (int aYear, int aMonth, int aDay,
                      int aHour, int aMin, int aSec,
                      double aFrac) {
    _year = aYear;
    boolean wLeapYearP = isLeapYear(aYear);
    _oneYearInSec = 31536000; // 365 * 24 * 3600
    if (wLeapYearP) _oneYearInSec += 86400;

    _tsec = 0;
    
    if (aMonth < 1 || aMonth > 12) {
      throw new IllegalArgumentException("illegal aMonth");
    }
    for (int i = 0; i < aMonth -1; i++) {
      _tsec += (wLeapYearP ? lDay[i] : nDay[i]) * 86400;
    }
    if (aDay < 1 ||
      aDay > (wLeapYearP ? lDay[aMonth-1] : nDay[aMonth-1])) {
      throw new IllegalArgumentException("illegal aDay");
    }
    _tsec += (aDay - 1) * 86400;
    if (aHour < 0 || aHour >= 24) {
      throw new IllegalArgumentException("illegal aHour");
    }
    _tsec += aHour * 3600;
    if (aMin < 0 || aMin >= 60) {
      throw new IllegalArgumentException("illegal aMin");
    }
    _tsec += aMin * 60;
    if (aSec < 0 || aSec >= 60) {
      throw new IllegalArgumentException("illegal aSec");
    }
    _tsec += aSec;
    _frac = aFrac;
  }
  
  /**
   * Creates a new <code>OhTimePoint</code> instance from another
   * <code>OhTimePoint</code> instance.
   *
   * <japanese>
   * <p>
   * Rs[RXgN^B
   * </japanese>
   *
   * @param tp an </code>OhTimePointImmutable</code> value
   */
  public OhTimePoint(OhTimePointImmutable tp) {
    _year = tp.getYear();
    _tsec = tp.getTsec();
    _oneYearInSec = tp.getOneYearInSec();
    _frac = tp.getFrac();
  }
  
  /**
   * Returns a deep copy of the current object.
   *
   * <japanese>
   * <p>
   * ݂̃IuWFNgƓԂ OhTimePoint ^̐VKIu
   * WFNg쐬ĕԂ܂B
   * </japanese>
   *
   * @return an <code>Object</code> value
   */
  public Object clone() {
    return new OhTimePoint(_year, _tsec, _frac);
  }
  
  /**
   * Copies fields from those of the
   * <code>OhTimePointImmutable</code> argument <b>tp</b>.
   *
   * <japanese>
   * <p>
   * ݂ OhTimePoint IuWFNg̏Ԃ OhTimePoint 
   * IuWFNgƓɂȂ悤ɂ܂B
   * </japanese>
   * @param tp an <code>OhTimePointImmutable</code> value
   */
  public void copy(OhTimePointImmutable tp) {
    _year = tp.getYear();
    _tsec = tp.getTsec();
    _oneYearInSec = tp.getOneYearInSec();
    _frac = tp.getFrac();
  }    
  
  /**
   * Tests if the specified Object is an instance of
   * OhTimePointImmutable and if it is equal to this
   * OhTimePoint.
   *
   * <japanese>
   * <p>
   * Ɏw肵IuWFNg OhTimePointImmutable IuWF
   * NgłāAA݂ OhTimePointImmutable IuWFN
   * gƓԂł邩eXg܂B
   * </japanese>
   *
   * @param arg an <code>Object</code> value
   * @return a <code>boolean</code> value
   */
  public boolean equals(Object arg) {
    return ( (arg != null) &&
      (arg instanceof OhTimePointImmutable) &&
      _year == ((OhTimePoint) arg)._year &&
      _tsec == ((OhTimePoint) arg)._tsec &&
      _frac == ((OhTimePoint) arg)._frac ) ;
  }
  
  /**
   * Tests if the specified Object is an instance of
   * OhTimePointImmutable and if the time differenc between the
   * current time point and <b>anObj</b> is less than <b>anEps</b>.
   * <p>
   *
   * <japanese>
   * Ɏw肵IuWFNg <b>anObj</b>  
   * OhTimePointImmutable IuWFNgłāAA݂ 
   * OhTimePointImmutable IuWFNgƂ̎ԍ <b>anEps</b> 
   * 菬eXg܂B
   *
   * <b>anEps</b> ̒lƂāA
   * <a href="OhTime.html#field_summary">OhTime#NEGLIGIBLY_SMALL_TIME</a> gpƂ悢ł
   * B
   * <p>
   * </japanese>
   *
   * @param anObj an <code>Object</code> value
   * @param anEps a <code>double</code> value
   * @return a <code>boolean</code> value
   */
  public boolean equals(Object anObj, double anEps) {
    return ( (anObj != null) &&
      (anObj instanceof OhTimePointImmutable) &&
      _year == ((OhTimePoint) anObj)._year &&
      _tsec == ((OhTimePoint) anObj)._tsec &&
      Math.abs(((OhTimePoint) anObj)._frac - _frac) < anEps);
  }
  
  /**
   * Gets the year
   *
   * @return an <code>int</code> value
   */
  public int getYear() {
    return _year;
  }
  
  /**
   * Gets the time from the beginning of the year of the time
   * expressed by this OhTimePoint object.<p>
   *
   * <japanese>
   * year tB[hŎw肳ĂN̔ŇoߕbԂ
   * ܂B
   * </japanese>
   *
   * @return a <code>long</code> value
   */
  public long getTsec() {
    return _tsec;
  }
  
  /**
   * Gets the number of seconds contained in the year of the
   * time expressed by this OhTimePoint object.<p>
   *
   * <japanese>
   * year tB[hŎw肳ĂN1N̒bŕ\B
   * </japanese>
   *
   * @return a <code>long</code> value
   */
  public long getOneYearInSec() {
    return _oneYearInSec;
  }
  
  /**
   * Gets the fraction part <b>frac</b>.
   *
   * <japanese>
   * <p>
   * frac tB[h̒lB-0.5 ȏA0.5 ̒lB
   * </japanese>
   *
   * @return a <code>double</code> value
   */
  public double getFrac() {
    return _frac;
  }
  
  /**
   * Tests if the year <b>aYear</b> is a leap year or not.<p>
   *
   * <japanese>
   * Ŏw肷N aYear 邤N (true) ǂ (false) 
   * 𔻒肵܂B
   * </japanese>
   *
   * @param aYear an <code>int</code> value
   * @return a <code>boolean</code> value
   */
  public static boolean isLeapYear(int aYear) {
    if (aYear % 4 != 0) return false;
    if (aYear % 100 != 0) return true;
    if (aYear % 400 != 0) return false;
    return true;
  }
  
  /**
   * Converts the current OhTimePoint object to the Calendar array.
   * When cal is the returned array, cal[OhTimePoint.YEAR],
   * cal[OhTimePoint.MONTH], cal[OhTimePoint.DAY],
   * cal[OhTimePoint.HOUR], cal[OhTimePoint.MINUTE],
   * cal[OhTimePoint.SECOND] contain year, month, hour, minute and
   * second data.
   *
   * @return an <code>int[]</code> value
   */
  public int[] toCalendarArray() {
    int[] cal = new int[6];
    
    int year = _year;
    int month = 1;
    int day = 1;
    int hour = 0;
    int minute = 0;
    int second = 0;
    int wT = (int)_tsec;
    int yearSec;
    boolean wLeapYearP;
    while (true) {
      wLeapYearP = isLeapYear(year);
      yearSec = 31536000; // 365 * 24 * 3600;
      if (wLeapYearP) { yearSec += 86400; }
      if (wT <= yearSec) break;
      year += 1;
      wT -= yearSec;
    }
    long monthSec;
    while (true) {
      monthSec = (wLeapYearP ? lDay[month] : nDay[month]) * 86400; 
      if (wT <= monthSec) break;
      month += 1;
      wT -= monthSec;
    }
    day = wT / 86400 + 1;
    wT -= (day - 1) * 86400;
    hour = wT / 3600;
    wT -= hour * 3600;
    minute = wT / 60;
    wT -= minute * 60;
    second = wT;

    cal[YEAR] = year; cal[MONTH] = month; cal[DAY] = day;
    cal[HOUR] = hour; cal[MINUTE] = minute; cal[SECOND] = second;

    return cal;
  }
  
  /**
   * Sets the time of this OhTimePoint object to the beginning
   * of the year aYear.<p>
   *
   * <japanese>
   * ݂ OhTimePoint IuWFNg aYear ̔Nɐݒ肵܂B
   * year tB[hݒ肷邾łȂAtsec, frac tB[h
   *  0 ɐݒ肵܂B
   * </japanese>
   *
   * @param aYear an <code>int</code> value
   */
  public void setYear(int aYear) {
    if (aYear < -1) {
      throw new IllegalArgumentException("illega aYear");
    }
    _year = aYear;
    _oneYearInSec = 365 * 24 * 3600;
    if (isLeapYear(_year)) _oneYearInSec += 86400;
    _tsec = 0;
    _frac = 0.;
  }
  
  /**
   * Set this OhTimePoint time at the time point when <b>aTsec</b>
   * seconds have elapsed from the beginning of the year.<p>
   *
   * <japanese>
   * ݂ OhTimePoint IuWFNg\킷Ałɐݒ肳
   * Ă year tB[h̔NAŎw肵bo
   * ߂Ď_ɐݒ肵܂B
   * </japanese>
   *
   * @param aTsec a <code>long</code> value
   */
  public void setTsec(long aTsec) {
    if (aTsec < 0 || aTsec > _oneYearInSec) {
      throw new IllegalArgumentException("illega aTsec");
    }
    _tsec = aTsec;
  }
  
  /**
   * Set the fraction part <b>frac</b> of the current object.<p>
   *
   * <japanese>
   * ̒[ frac  aFrac ɐݒ肵܂B
   * </japanese>
   *
   * @param aFrac a <code>double</code> value
   */
  public void setFrac(double aFrac) {
    if (aFrac < -0.5 || aFrac >= 0.5) {
      throw new IllegalArgumentException("illega aTsec");
    }
    _frac = aFrac;
  }
  
  /**
   * Compares the current OhTimePoint object to another 
   * object <b>arg</b> and returns the result.
   * The returend value is 1, 0, -1 
   *
   * <japanese>
   * <p>
   * ݂ OhTimePoint IuWFNgƈŗ^ OhTimePoint 
   * IuWFNgƂ̑Orʂ𐮐l 1, 0, -1 ̂
   * ꂩŕԂ܂B0 ́AB-1 ́A݂ OhTimePoint I
   * uWFNg̕OA1 ́AӖ܂B
   * </japanese>
   *
   * @param arg an <code>Object</code> value
   * @return an <code>int</code> value.
   *   the value 0 if this OhTimePoint object expresses the same time as
   *   the argument OhTimePoint object; 
   *   the value -1 0 if this OhTimePoint object expresses earlier
   *   time than the
   *   argument OhTimePoint object;
   *   the value 1 if this OhTimePoint object expresses later time than the
   *   argument OhTimePoint object.
   */
  public int compareTo(Object arg) {
    OhTimePoint a = (OhTimePoint)arg;
    if (_year > a._year) return 1;
    if (_year < a._year) return -1;
    if (_tsec > a._tsec) return 1;
    if (_tsec < a._tsec) return -1;
    if (_frac > a._frac) return 1;
    if (_frac < a._frac) return -1;
    return 0;
  }
  
  public int compareTo(OhTimePointImmutable arg) {
    if (_year > arg.getYear()) return 1;
    if (_year < arg.getYear()) return -1;
    if (_tsec > arg.getTsec()) return 1;
    if (_tsec < arg.getTsec()) return -1;
    if (_frac > arg.getFrac()) return 1;
    if (_frac < arg.getFrac()) return -1;
    return 0;
  }
  
  /**
   * Adds the specified amount of seconds to the current object.
   * <p>
   *
   * <japanese>
   * <p>
   * ݂ OhTimePoint IuWFNg timeSec b܂B
   * </japanese>
   *
   * @param timeSec a <code>long</code> value. 
   */
  public void add(long timeSec) {
    if (timeSec >= 0) {
      _tsec += timeSec;
      while (_tsec >= _oneYearInSec) {
        _tsec -= _oneYearInSec;
        _year += 1;
        _oneYearInSec = 31536000L; // 365 * 24 * 3600;
        if (isLeapYear(_year)) _oneYearInSec = 31622400L;
      }
    }
    else {
      _tsec += timeSec;
      while (_tsec < 0) {
        _year -= 1;
        _oneYearInSec = 31536000L; // 365 * 24 * 3600;
        if (isLeapYear(_year)) _oneYearInSec = 31622400L;
        _tsec += _oneYearInSec;
      }
    }
  }
  
  /**
   * Adds the specified amount of seconds to the current object.
   * <p>
   *
   * <japanese>
   * <p>
   * ݂ OhTimePoint IuWFNg OhTimeImmutable time ŕ\
   * 킳鎞Ԃ𑫂܂B
   * </japanese>
   *
   * @param time a <code>OhTimeImmutable</code> value. 
   */
  public void add(OhTimeImmutable time) {
    double frac = _frac + time.getFrac();
    long n = Math.round(frac);
    _frac = frac - n;
    long timeSec = time.getSec() + n;
    if (timeSec >= 0) {
      _tsec += timeSec;
      while (_tsec >= _oneYearInSec) {
        _tsec -= _oneYearInSec;
        _year += 1;
        _oneYearInSec = 31536000L; // 365 * 24 * 3600;
        if (isLeapYear(_year)) _oneYearInSec = 31622400L;
      }
    }
    else {
      _tsec += timeSec;
      while (_tsec < 0) {
        _year -= 1;
        _oneYearInSec = 31536000L; // 365 * 24 * 3600;
        if (isLeapYear(_year)) _oneYearInSec = 31622400L;
        _tsec += _oneYearInSec;
      }
    }
  }
  
  /**
   * Adds <b>aTime</b> to <b>aTp</b> and returns the resultant
   * object.
   *
   * <japanese>
   * <p>
   *  aTp ɁA aTime 𑫂Ԃ܂B
   * </japanese>
   *
   * @param aTp an <code>OhTimePointImmutable</code> value
   * @param aTime an <code>OhTimeImmutable</code> value
   * @return an <code>OhTimePoint</code> value
   */
  public static OhTimePoint add(OhTimePointImmutable aTp,
    OhTimeImmutable aTime) {
    OhTimePoint tp = new OhTimePoint(aTp.getYear(), aTp.getTsec(),
      aTp.getFrac());
    tp.add(aTime);
    return tp;
  }
  
  /**
   * Sutracts aLongTime from the current OhTimePoint object and
   * returnts the updated object.
   *
   * <japanese>
   * <p>
   * ݂ OhTimePoint IuWFNg aLongTime Ŏw肷
   * 鎞Ԃ܂B
   * </japanese>
   *
   * @param aLongTime a <code>long</code> value
   */
  public void sub(long aLongTime) {
    add(-aLongTime);
  }
  
  /**
   * Subtracts aTime from the current OhTimePoint object.
   *
   * <japanese>
   * <p>
   * ݂ OhTimePoint IuWFNg aTime Ŏw肷鎞
   * Ԃ܂B
   * </japanese>
   *
   * @param aTime an <code>OhTimeImmutable</code> value
   */
  public void sub(OhTimeImmutable aTime) {
    OhTime t = new OhTime(-aTime.getSec(), -aTime.getFrac());
    t.adjust();
    add(t);
  }
  
  /**
   * Subtract <b>aTime</b> from <b>aTp</b> and returns the
   * resultant OhTimePoint object.
   *
   * <japanese>
   * <p>
   *  aTp A aTime Ԃ܂B
   * </japanese>
   *
   * @param aTp an <code>OhTimePointImmutable</code> value
   * @param aTime an <code>OhTimeImmutable</code> value
   * @return an <code>OhTimePoint</code> value
   */
  public static OhTimePoint sub(OhTimePointImmutable aTp,
    OhTimeImmutable aTime) {
    OhTimePoint tp = new OhTimePoint(aTp.getYear(), aTp.getTsec(),
      aTp.getFrac());
    tp.sub(aTime);
    return tp;
  }
  
  /**
   * Returns the time difference between the two
   * OhTimePointImmutable objects.<p>
   *
   * <japanese>
   * <p>
   * ŗ^2 OhTimePointImmutable IuWFNg̎ԍ
   * Ԃ܂B
   * </japanese>
   *
   * @param aT an <code>OhTimePointImmutable</code> value
   * @param bT an <code>OhTimePointImmutable</code> value
   * @return a <code>OhTime</code> value
   */
  public static OhTime sub(OhTimePointImmutable aT, OhTimePointImmutable bT) {
    long t    = 0L;
    int  yeara = aT.getYear();
    long tseca = aT.getTsec();
    int  yearb = bT.getYear();
    long tsecb = bT.getTsec();
    
    if (yeara == yearb) {
      t = Math.abs(tseca - tsecb);
    }
    else if (yeara < yearb) {
      int y = yeara;
      while (y < yearb) {
        if (isLeapYear(y))
          t = t + 31622400L;
        else
          t = t + 31536000L;
        y++;
      }
      t = t - tseca + tsecb;
    }
    else {
      int y = yearb;
      while (y < yeara) {
        if (isLeapYear(y))
          t = t + 31622400L;
        else
          t = t + 31536000L;
        y++;
      }
      t = t - tsecb + tseca;
    }
    double frac = aT.getFrac() - bT.getFrac();
    OhTime time = OhTime.adjust(t, frac);
    return time;
  }
  
  /**
   * Returns the time difference between the current OhTimePoint
   * and the OhTimePointImmutable <b>aT</b>.
   *
   * <japanese>
   * <p>
   * ݂ OhTimePoint IuWFNgƁAŎw肵 OhTimePoint 
   * IuWFNg aT ̎ԍvZĖ߂܂B
   * </japanese>
   *
   * @param aT an <code>OhTimePointImmutable</code> value
   * @return an <code>OhTime</code> value
   */
  public OhTime sub(OhTimePointImmutable aT) {
    long t    = 0L;
    int  year = aT.getYear();
    long tsec = aT.getTsec();
    
    if (_year == year) {
      t = Math.abs(_tsec - tsec);
    }
    else if (_year < year) {
      int y = _year;
      while (y < year) {
        if (isLeapYear(y))
          t = t + 31622400L;
        else
          t = t + 31536000L;
        y++;
      }
      t = t - _tsec + tsec;
    }
    else {
      int y = year;
      while (y < _year) {
        if (isLeapYear(y))
          t = t + 31622400L;
        else
          t = t + 31536000L;
        y++;
      }
      t = t - tsec + _tsec;
    }
    double frac = _frac - aT.getFrac();
    OhTime time = OhTime.adjust(t, frac);
    return time;
  }
  
  /**
   * Compares two OhTimePointImmutable objects <b>aTpa</b> and
   * <b>aTpb</b>. When the time difference of the two objects
   * is less than <b>aEps</b>, then 0 is returned.
   *
   * <japanese>
   * <p>
   * ̎ <b>aTpa</b>  <b>aTpb</b> ̎ԍ𒲂ׂ܂B
   * ԍ̐Βl <b>aEps</b> 菬ꍇ́AIɓ
   * ƌȂāA0 Ԃ܂B<b>aTpa</b> 
   * <b>aTpb - aEps</b> ÔƂ́A-1 Ԃ܂B
   * <b>aTpa</b>  <b>aTpb + aEps</b> ̂Ƃ́A
   * +1 Ԃ܂B
   * </japanese>
   *
   * @param aTpa an <code>OhTimePointImmutable</code> value
   * @param aTpb an <code>OhTimePointImmutable</code> value
   * @param aEps a <code>double</code> value
   * @return an <code>int</code> value
   */
  public static int compare(
    OhTimePointImmutable aTpa, OhTimePointImmutable aTpb, double aEps) {
    OhTimePoint ta = (OhTimePoint) aTpa.clone();
    OhTime d = ta.sub(aTpb);
    double diff = d.toDouble();
    if (diff > aEps ) return 1;
    if (diff < -aEps) return -1;
    return 0;
  }
  
  /**
   * Returns "" + _year + " " + _tsec + " " + _frac.
   *
   * @return a <code>String</code> value
   */
  public String toString() {
    StringBuffer a = new StringBuffer("");
    a.append(_year).append(" ").append(_tsec);
    a.append(" ").append(_frac);
    return a.toString();
  }
  
  /**
   * Returns the Calendar String. For example, this method returns
   * "20030101010101 0.0" for 2003 January 1st, 01:01:01 0.0.
   *
   * @return a <code>String</code> value
   */
  public String toCalendarString() {
    int[] cal;
    cal = toCalendarArray();
    StringBuffer a = new StringBuffer("");
    DecimalFormat df = new DecimalFormat("00");
    a.append(cal[0]).
       append(df.format(cal[1])).
       append(df.format(cal[2])).
       append(df.format(cal[3])).
       append(df.format(cal[4])).
       append(df.format(cal[5])).
       append(" ").
       append(_frac);
    return a.toString();
  }
  
  /**
   * Writes the state of the current object into aOut. The
   * readObject method restore the object by reading the result.
   *
   * <japanese>
   * <p>
   * aOut ɁA݂ OhTime IuWFNg̏Ԃo܂B
   * readObject AŁAǂݎăIuWFNg̏Ԃ
   * ܂B
   * </japanese>
   *
   * @param aOut an <code>ObjectOutputStream</code> value
   * @exception IOException if an error occurs
   */
  private void writeObject(ObjectOutputStream aOut) 
    throws IOException {
    aOut.defaultWriteObject();
    aOut.writeInt(_year);
    aOut.writeLong(_tsec);
    aOut.writeDouble(_frac);
    aOut.writeLong(_oneYearInSec);
  }
  
  /**
   * Restores the state of the object from aIn.
   *
   * <japanese>
   * <p>
   * aIn AOhTime IuWFNg̏Ԃǂݎ܂B
   * </japanese>
   *
   * @param aIn an <code>ObjectInputStream</code> value
   * @exception IOException if an error occurs
   * @exception ClassNotFoundException if an error occurs
   */
  private void readObject(ObjectInputStream aIn)
    throws IOException, ClassNotFoundException {
    aIn.defaultReadObject();
    _year = aIn.readInt();
    _tsec = aIn.readLong();
    _frac = aIn.readDouble();
    _oneYearInSec = aIn.readLong();
  }
  
//////////////////////////////////////////////////////////////////////////
// added by Y.Satoh (2003/10/03)
  /**
   * Minimun value of OhTimePoint objects,  OhTimePoint(-1, 0L).
   *
   */
  public static OhTimePoint OhTimePointMin = new OhTimePoint(-1, 0L);
  
  /**
   * Maximum value of OhTimePoint objects,  OhTimePoint(10000, 0L).
   *
   */
  public static OhTimePoint OhTimePointMax = new OhTimePoint(10000, 0L);
  
  /**
   * OhTimePoint object with <b>year</b> = 0, <b>sec</b> = 0 and
   * <b>frac</b> = 0.0.
   *
   */
  public final static OhTimePoint ZERO = new OhTimePoint(0, 0L, 0.0);
  
}// ohtimepoint
