/**
 * @(#)  Class to express Power Law Kinematic Wave model for Stream Flow.
 *
 */

package ohymosj;

import ohymosj.*;
import java.util.*;
import java.io.*;


/**
 * <b>classname</b>: <br>
 * PowerLawKinematicWave<p>
 *
 * <b>Receiveports</b>: <br>
 * (1) OhReceiveDouble("rpUpIn", i), i = 0, ..., _numOfUpIn-1.
 * _numOfUpIn is variable. Inflow from upstream end of the
 * stream segment.
 * The dimension of _cUpIn[i]*rpUpIni must be m**3/sec where
 * rpUpIni is the value obtained from the port ("rpUpIn",
 * i). See: <a href="OhReceiveDouble.html">OhReceiveDouble</a> <p>
 *
 * (2) OhReceiveDouble("rpLateralIn", i), i = 0, ...,
 * _numOfLateralIn-1. _numOfLateralIn is variable. Lateral
 * inflow to the stream segment.
 * The dimension of _cLateralIn[i]*rpLateralIni must be m**3/sec
 * where rpLateralIni is the value obtained from the port
 * ("rpLateralIn", i). See: <a href="OhReceiveDouble.html">OhReceiveDouble</a> <p>
 *
 * <b>Sendports</b>: <br>
 * OhSendDouble("sp", 0) -- discharge (m**3/sec) at the
 *   downstream end of the stream segment. brokenline. <br>
 * See: <a href="OhSendDouble.html">OhSendDouble</a> <p>
 *
 * <b>parameters</b>: 
 * <dl>
 *  <dt> _a <dd>
 *  double. kinematic constant in Q = _a A^(_m) (m-sec system)
 *  <dt> _m <dd>
 *  double. kinematic constant in Q = _a A^(_m) (-)
 *  <dt> _lengthOfStream <dd>
 *  double. length of the stream segment (m)
 *  <dt> _numOfSepStream <dd>
 *  int. number of divisions of the stream segement. The stream
 *  segment is devided into _numOfSepStream equal length
 *  divisions so that each division has length
 *  _lengthOfStream/_numOfSepStream.
 *  <dt> dt    <dd>
 *  double. time step (sec)
 *  <dt> _theta <dd>
 *  double. bed slope of the stream segment (-)
 *  <dt> _qLevel <dd>
 *  double. standard level value of the discharge (m^3/sec).
 *  <dt> _epsilon <dd>
 *  double. small value (-).
 *  <dt> _cUpIn[i], i=0, ..., _numOfUpIn-1 <dd>
 *  double[].  _cUpIn[i] is the weight to the i-th inflow receive port.
 *  <dt> _cLateralIn[i], i=0, ..., _numOfLateralIn-1 <dd>
 *  double[]. _cLateralIn[i] is the weight to the i-th lateral
 *  inflow receive port.
 * </dl>
 *
 * <b>state variables</b>:<br>
 * <dl>
 *  <dt> _currentTime <dd>
 *  OhTimePoint.
 *  <dt> _qCrnt[i], i = 0, ..., _numOfSepStream <dd>
 *  double[]. the dischareges in the divisions of the stream segement
 *  (m^3/sec)
 * </dl>
 * <p>
 *
 * The <b>init(String, int, OhBufferedReader)</b> method looks like:
 *
 * <pre>
 *  public void init(String aObjName, int aObjNum,
 *    OhBufferedReader aBr) throws OhError {
 *    int numOfUpIn, numOfLateralIn;
 *    try {
 *      aBr.skipWhites();
 *      numOfUpIn      = Integer.parseInt(aBr.readWord());
 *      aBr.skipWhites();
 *      numOfLateralIn = Integer.parseInt(aBr.readWord());
 *    }
 *    catch (Exception e) {
 *      ...
 *    }
 *    init(aObjName, aObjNum, null, numOfUpIn, numOfLateralIn);
 *  }
 * </pre>
 * 
 * The <b>setParemeter(OhBufferedReader)</b> method used for
 * reading parameters from a file looks like:
 *
 * <pre>
 *  public void setParameter(OhBufferedReader aBr) throws OhError {
 *    double dt = 0.0;
 *    try {
 *      aBr.skipWhites();
 *      _a = Double.parseDouble(aBr.readWord());
 *      _m = Double.parseDouble(aBr.readWord());
 *      _lengthOfStream = Double.parseDouble(aBr.readWord());
 *      _numOfSepStream = Integer.parseInt(aBr.readWord());
 *      dt = Double.parseDouble(aBr.readWord());
 *      _dt = new OhTime(dt);
 *      _theta = Double.parseDouble(aBr.readWord());
 *      _qLevel = Double.parseDouble(aBr.readWord());
 *      _epsilon = Double.parseDouble(aBr.readWord());
 *    }
 *    catch (IOException e) {
 *      throw new OhError(e.toString() + NEWLINE + oPrintString("\t"));
 *    }
 *    catch (NumberFormatException e) {
 *      throw new OhError(e.toString() + NEWLINE + oPrintString("\t"));
 *    }
 *
 *    _qCrnt = new double[_numOfSepStream + 1];
 *    _qNext = new double[_numOfSepStream + 1];
 *
 *    try {
 *      if (_numOfUpIn > 0) {
 *        for (int i = 0; i < _numOfUpIn; i++) {
 *          aBr.skipWhites();
 *          _cUpIn[i] = Double.parseDouble(aBr.readWord());
 *        }
 *      }
 *
 *      if (_numOfLateralIn > 0) {
 *        for (int i = 0; i < _numOfLateralIn; i++) {
 *          aBr.skipWhites();
 *          _cLateralIn[i] = Double.parseDouble(aBr.readWord());
 *        }
 *      }
 *    }
 *    catch (IOException e) {
 *      throw new OhError(e.toString() + NEWLINE + oPrintString("\t"));
 *    }
 *    catch (NumberFormatException e) {
 *      throw new OhError(e.toString() + NEWLINE + oPrintString("\t"));
 *    }
 *
 *    _dx = _lengthOfStream / _numOfSepStream;
 *  }
 * </pre>
 *
 * Note: <a href="OhTime.html#field_summary">OhTime.NEGLIGIBLY_SMALL_TIME</a> <p>
 *
 * The <b>setInitialState(OhBufferedReader)</b> method used for
 * reading initial states from a file looks like:<p>
 *
 * <pre>
 *  public void setInitialState(OhBufferedReader aBr) throws OhError {
 *    int  year = 0;
 *    long tsec = 0L;
 *    double frac = 0.0;
 *
 *    // System.out.println("Kinematic#setInitialState#1");
 *    
 *    try {
 *      aBr.skipWhites();
 *      year = Integer.parseInt(aBr.readWord());
 *      tsec = Long.parseLong(aBr.readWord());
 *      frac = Double.parseDouble(aBr.readWord());
 *    }
 *    catch (IOException e) {
 *      // System.out.println("Kinematic#setInitialState#2");
 *      throw new OhError("Error in setInitialState()." + NEWLINE +
 *       oPrintString("\t"));
 *    }
 *    catch (NumberFormatException e) {
 *      // System.out.println("Kinematic#setInitialState#3");
 *      throw new OhError("Error in setInitialState().",
 *        oPrintString("\t"));
 *    }
 *
 *    if (year > 0) {
 *      _currentTime = new OhTimePoint(year, tsec, frac);
 *    }
 *
 *    int dataType;  // 0:ONLY_DOWN_STREAM_END 1:EVERY_SECTION
 *
 *    try {
 *      aBr.skipWhites();
 *      dataType = Integer.parseInt(aBr.readWord());
 *    }
 *    catch (IOException e) {
 *      throw new OhError(oPrintString("\t"));
 *    }
 *    catch (NumberFormatException e) {
 *      throw new OhError(oPrintString("\t"));
 *    }
 *
 *    if (dataType == 0)
 *      _isdt = InitialStateDataType.ONLY_DOWN_STREAM_END;
 *    else
 *      _isdt = InitialStateDataType.EVERY_SECTION;
 *
 *    try {
 *      if (_isdt == InitialStateDataType.ONLY_DOWN_STREAM_END) {
 *        aBr.skipWhites();
 *        _qCrnt[_numOfSepStream] = Double.parseDouble(aBr.readWord());
 *      }
 *      else {
 *        for (int i = 1; i <= _numOfSepStream; i++) {
 *          _qCrnt[_numOfSepStream] = Double.parseDouble(aBr.readWord());
 *        }
 *      }
 *    }
 *    catch (IOException e) {
 *      throw new OhError(oPrintString("\t"));
 *    }
 *    catch (NumberFormatException e) {
 *      throw new OhError(oPrintString("\t"));
 *    }
 *  }
 * </pre>
 *
 * <b>Samples of a SCF file, a PMT file, and a IST file</b>:<p>
 *
 * <b>a part of a SCF file</b>
 * <pre>
 * [InputPorts]
 * # id objname classname
 * 
 * 0 tplkinArakura.i OhInputFileDouble
 *
 * [Components]
 * # id objname objnum classname basin_structure_file nin ...
 * 
 * 0   stream 0   PowerLawKinematicWave  0  1
 * 1   stream 1   PowerLawKinematicWave  0  1
 * ...
 * 102 stream 102 PowerLawKinematicWave  2  1
 * ...
 *
 *  [ConnectPorts]
 * r 0  c 0 rpLateralIn 0  # connect input port 0 to rpLateralIn 0 of component 0
 * r 0  c 1 rpLateralIn 0
 * ...
 * </pre>
 *
 * <b>A part of a parameter file</b><p>
 * <pre>
 *   PowerLawKinematicWave stream 0
 *   {
 *     0.01 1.67 1700. 20 60 0.5 100 0.001  # _a, _m, _lengthOfStream, _numOfSepStream, _dt, _theta, _qLevel, _epsilon
 *     1.0                                  # _cLateralIn[0]
 *   }
 * </pre>
 *
 * <b>A part of a initial state file</b><p>
 * <pre>
 *   PowerLawKinematicWave stream 0
 *   {
 *     2000 0 0.0  # _currentTime
 *     0           # datatype 0:ONLY_DOWN_STREAM_END
 *     0.          # m^3/sec
 *   }
 * </pre>
 *
 * <japanese>
 * PowerLawKinematicWave NX́Aʂς㰏ŕ\킳
 * ͓̃fłBЂƂ͓̉Ԃ̗ΏۂƂ܂B<p>
 *
 * ŁA͓ԂƂ͂̓ɍ_ȂP{͓̉
 * ̂Ƃ܂B Q ́Aʐfʐ A ̎w֐ a*A^m 
 * \ł̂Ƃ܂B<p>
 *
 * <b>M[q</b>:<br>
 * (1) ͓̉̃f́A㗬[畡̗ʂ
 * Ƃ邽߂̎M[qĂ܂BM[q̌ _numOfUpIn 
 * ́ACӂɕύXł܂BM[q́A
 *
 * <pre>
 *   OhReceiveDouble("rpUpIn", i), i = 0, ..., _numOfUpIn-1
 * </pre>
 *
 * łB(Q <a href="OhReceiveDouble.html">OhReceiveDouble</a>) M[qMf[
 * ^ɏ搔 _cUpIn[i] |̂A (m^3/sec) ƂȂ
 * B<p>
 *
 * (2) ܂A̂̉ʂ󂯂Ƃ邽߂̎M[q
 * Ă܂BM[q̌ _numOfLateralIn ́ACӂɕύXł
 * BM[q́A
 *
 * <pre>
 *   OhReceiveDouble("rpLateralIn", i), i = 0, ..., _numOfLateralIn-1
 * </pre>
 *
 * łB(Q <a href="OhReceiveDouble.html">OhReceiveDouble</a>) M[qMf[
 * ^ɏ搔 _cLateralIn[i] |̂A (m^3/sec) ƂȂ
 * B<p>
 *
 * <b>M[q</b>:<br>
 * M[q́A1ŁA
 *
 * <pre>
 *   OhSendDouble("sp", 0)
 * </pre>
 *
 * łB(Q <a href="OhSendDouble.html">OhSendDouble</a>)B
 * <a href="DataConnectionType.html">DataConnectionType</a> ́Abrokenline łBM[q 
 * ("sp", 0) 瑗Mf[^͉͓敪[̗ʂŁAPʂ́A
 * m**3/sec łB
 * <p>
 *
 * <b>p^</b>:<br>
 * <dl>
 *  <dt> _a <dd>
 *  double. ʗϊ֌W Q = _a A^(_m) ̌W (m-sec system)
 *  <dt> _m <dd>
 *  double. ʗϊ֌W Q = _a A^(_m) ̎w (-)
 *  <dt> _lengthOfStream <dd>
 *  double. ͓敪 (m)
 *  <dt> _numOfSepStream <dd>
 *  int. ͓敪̕. ͓敪𓙕̂ŁAꂽ͓̉fЂ
 *  ́A_lengthOfStream/_numOfSepStream łB
 *  <dt> dt    <dd>
 *  double. vZԊԊuBŒłB(sec)
 *  <dt> _theta <dd>
 *  double. ͓̒z (-)
 *  <dt> _qLevel <dd>
 *  double. WIȗ (m^3/sec).
 *  <dt> _epsilon <dd>
 *  double. l (-).
 *  <dt> _cUpIn[i], i=0, ..., _numOfUpIn-1 <dd>
 *  double[].  _cUpIn[i] ́Ai Ԗڂ͓̉[̗ʂւ̔{.
 *  <dt> _cLateralIn[i], i=0, ..., _numOfLateralIn-1 <dd>
 *  double[]. _cLateralIn[i] ́Ai Ԗڂ̉ʂւ̔{.
 * </dl>
 *
 * <b>ԗ</b>:<br>
 * <a href="OhTimePoint.html">OhTimePoint</a> ^̎_numOfSepStream +1 ̒fʂ̗ (m**3)<p>
 *
 * _numOfUpIn, _numOfLateralIn ́A
 * <a href="PowerLawKinematicWave.html#method_summary">init(String, int, OhBufferedReader)</a> Őݒ
 * ܂Bp^́A<a href="PowerLawKinematicWave.html#method_summary">setParameter()</a> œǂ
 * ܂܂Bԗʂ́A<a href="PowerLawKinematicWave.html#method_summary">setInitialState()</a> 
 * ǂݍ܂܂B
 * </japanese>
 *
 * @author satoh, shiiba
 * @version $Revision: 1.38 $ $Date: 2004/02/10 20:08:11 $
 */

/*
 * <japanese>
 * 1. NX OhElement pB  
 * </japanese>
 */
public class PowerLawKinematicWave extends OhElement
{
  // @begin {fields} -------------------------------------------

/*
 * <japanese>
 * 2. M[qEM[qA܂͎M[qEM[q̔zo
 *    ϐƂĒ`B
 * </japanese>
 */
  
  /**
   * Structural variable:
   * number of inflows from the upstream ends.
   * _numOfUpIn >= 0.
   *
   */
  protected int _numOfUpIn;

  /**
   * Structural variable:
   * number of lateral inflows.
   * _numOfLateralIn >= 0.
   *
   */
  protected int _numOfLateralIn;

  /**
   * inflow ports from the upstream end.
   *
   */
  protected OhReceiveDouble[] _rpUpIn;

  /**
   * lateral inflow ports.
   * Lateral inflow rate per unit length of stream will be
   * calculated by multiplying constant _cLateralIn[i] with the
   * value obtained from _rpLateralIn[i]
   *
   */
  protected OhReceiveDouble[] _rpLateralIn;

  /**
   * Send port.
   * Runoff discharge (m^3/sec) at the slopeend or the stream end. 
   *
   * <japanese>
   * <p>
   * M[q͂PłB
   * f[^ <a href="OhDataPackDouble.html">OhDataPackDouble</a> ^łB
   * f[^Ԍ^
   * <a href="DataConnectionType.html#field_summary">DataConnectionType#BROKENLINE</a>
   * łB<p>
   *
   * ͓[̗oʂ܂B
   * vZꂽoʂ̎ [m^3/sec] łB
   * </japanese>
   *
   */
  protected OhSendDouble _sp;

/*
 * <japanese>
 * 3. p^Eԗʂ̋LɗpϐoϐƂĒ`
 *    B
 * </japanese>
 */

  // Parameters.

  /**
   * Parameter:
   * kinematic constant <b>a</b> [m/sec] in
   *
   * <pre>
   *   Q = a*A^m
   * </pre>
   *
   */
  protected double _a;

  /**
   * Parameter:
   * kinematic constant <b>m</b> [-] in
   *
   * <pre>
   *   Q = a*A^m
   * </pre>
   *
   */
  protected double _m;

  /**
   * Parameter:
   * length of the stream segment. [m]
   *
   * <japanese>
   * <p>
   * ͓Ԃ̒B[m]
   * </japanese>
   *
   */
  protected double _lengthOfStream;

  /**
   * Parameter:
   * number of divisions of the stream. [-]
   *
   * <japanese>
   * <p>
   * ͓Ԃ𓙕B[-]
   * </japanese>
   *
   */
  protected int _numOfSepStream;

  /**
   * Parameter:
   * time step [OhTime] for finite difference calculation.
   * Time step is fixed in this calculation model.
   *
   * <japanese>
   * <p>
   * vZԊԊuB[OhTime] <p>
   *
   * ̃fł͌ŒlƂĈ܂B
   * </japanese>
   *
   */
  protected OhTime _dt;

  /**
   * Parameter:
   * time weighting parameter [-].
   * The standard value of theta is 0.5. 
   * If theta = 0, the finite difference calculation scheme
   * becomes an explicit one.
   *
   */
  protected double _theta;

  /**
   * Parameter:
   * tolerance level of diffrence of two consecutive caluculated
   * discharges to stop the iteration.
   *
   */
  protected double _epsilon;

  /**
   * Parameter:
   * tolerance level of diffrence of two consecutive caluculated
   * discharges to stop the iteration.
   *
   */
  protected double _qLevel;

  /**
   * Parameter:
   * _cUpIn[i] is a multiplication factor [-] to the i-th inflow
   * port from upstream end, _rpUpIn[i].
   *
   */
  protected double[] _cUpIn;

  /**
   * Parameter:
   * _cLateralIn[i] is a multiplication factor [-] to the i-th
   * lateral inflow port _rpLateralIn[i].
   *
   */
  protected double[] _cLateralIn;


  // State variable.

  /**
   * State variable:
   * discharge Q [m^3/sec] ( = q * w) at current time.
   *
   */
  protected double[] _qCrnt;


  // Working variables.

  /**
   * Working variable:
   * length of the division. [m]
   *
   * <pre>
   *   _dx = _lengthOfStream / _numOfSepStream
   * </pre>
   *
   */
  protected double _dx;

  /**
   * Working variable:
   * discharge Q [m^3/sec] at next time.
   *
   */
  protected double[] _qNext;

  /**
   * <japanese>
   * ł̗ʂ̕zǂ̂悤ɗ^邩w肷ϐB
   * ONLY_DOWN_STREAM_END ȂA[ŗʂ^āA͓}B
   * EVERY_SECTION ȂAx = _dx, 2*_dx, ..., _numOfSepStream *
   * _dx ̕_ŗʂ^B
   * </japanese>
   * <p>
   * Supplemental variable for setting initial state.
   *
   */
  protected InitialStateDataType _isdt;

  // @end {fields} ---------------------------------------------


  // @begin {initialization} -----------------------------------

  /**
   * Initialization. <p>
   *
   * Uses init(String, int, String, int, int).
   *
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @param aBr an <code>OhBufferedReader</code> value
   * @return a <code>boolean</code> value
   * @exception OhError if an error occurs
   */
  public void init(String aObjName, int aObjNum,
    OhBufferedReader aBr) throws OhError {
    int numOfUpIn;
    int numOfLateralIn;

    try {
      aBr.skipWhites();
      numOfUpIn      = Integer.parseInt(aBr.readWord());
      aBr.skipWhites();
      numOfLateralIn = Integer.parseInt(aBr.readWord());
    }
    catch (IOException e) {
      throw new OhError(oPrintString("\t"));
    }
    catch (NumberFormatException e) {
      throw new OhError(oPrintString("\t"));
    }

    init(aObjName, aObjNum, null, numOfUpIn, numOfLateralIn);
  }

  /**
   * Initialization. <p>
   *
   * Uses OhElement.init(String, int, int, String).
   *
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @param aDataStr a <code>String</code> value
   * @param aNumOfUpIn an <code>int</code> value
   * @param aNumOfLateralIn an <code>int</code> value
   * @return a <code>boolean</code> value
   * @exception OhError if an error occurs
   */
  public void init(String aObjName, int aObjNum, String aDataStr,
    int aNumOfUpIn, int aNumOfLateralIn) throws OhError {

    _numOfUpIn      = aNumOfUpIn;
    _numOfLateralIn = aNumOfLateralIn;

    if (_numOfUpIn < 0)
      _numOfUpIn = 0;

    if (_numOfLateralIn < 0)
      _numOfLateralIn = 0;

    _rpUpIn      = new OhReceiveDouble[_numOfUpIn];
    _rpLateralIn = new OhReceiveDouble[_numOfLateralIn];

    _cUpIn      = new double[_numOfUpIn];
    _cLateralIn = new double[_numOfLateralIn];

    if (_numOfUpIn > 0) {
      for (int i = 0; i < _numOfUpIn; i++) {
        _rpUpIn[i] = new OhReceiveDouble();  //; added by Satoh.
        _rpUpIn[i].init("rpUpIn", i);
      }
    }

    if (_numOfLateralIn > 0) {
      for (int i = 0; i < _numOfLateralIn; i++) {
        _rpLateralIn[i] = new OhReceiveDouble();  //; added by Satoh.
        _rpLateralIn[i].init("rpLateralIn", i);
      }
    }

    _sp = new OhSendDouble();  //; added by Satoh.
    _sp.init("sp", 0);

    super.init(aObjName, aObjNum, 0, aDataStr);
  }
  
  // @end {initialization} -------------------------------------

/*
 * <japanese>
 * 4. RXgN^`B
 * </japanese>
 */

  // @begin {constructors} -------------------------------------

  /**
   * Creates a new default <code>PowerLawKinematicWave</code>
   * instance.
   *
   * @exception OhError if an error occurs
   */
  public PowerLawKinematicWave() {
    super();

    _cUpIn       = null;
    _cLateralIn  = null;
    _qCrnt       = null;
    _qNext       = null;
    _rpUpIn      = null;
    _rpLateralIn = null;
    
    _className = "PowerLawKinematicWave";
  }

  /**
   * Creates a new <code>SimpleDam</code> instance.
   *
   * Uses init(String, int, String, int, int).
   *
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @param aDataStr a <code>String</code> value
   * @param aNumOfUpIn an <code>int</code> value
   * @param aNumOfLateralIn an <code>int</code> value
   * @exception OhError if an error occurs
   */
  public PowerLawKinematicWave(
    String aObjName, int aObjNum, String aDataStr,
    int aNumOfUpIn, int aNumOfLateralIn) throws OhError {
    super();

    _cUpIn       = null;
    _cLateralIn  = null;
    _qCrnt       = null;
    _qNext       = null;
    _rpUpIn      = null;
    _rpLateralIn = null;

    _className = "PowerLawKinematicWave";

    init(aObjName, aObjNum, aDataStr, aNumOfUpIn, aNumOfLateralIn);
  }
  
  // @end {constructors} ---------------------------------------

  // @begin {methods} ------------------------------------------

/*
 * <japanese>
 * 5. M[qEM[qo^郁\bh 
 *    registerReceivePorts(), registerSendPorts() `B
 * </japanese>
 */

  /**
   * <japanese>
   * SM[q̓o^ <p>
   *
   * ̃\bh́A̗vffׂĂ̎M[q
   * <a href="OhElement.html#method_summary">OhElement#register(Object)</a>
   * <a href="OhElement.html#method_summary">OhElement#register(Object, int)</a>
   * ɂo^܂B<p>
   *
   * o^ĂȂ[q͐ڑłȂ̂őMɗpł܂B
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public void registerReceivePorts() throws OhError { 
    for (int i = 0; i < _numOfUpIn; i++) {
      register(_rpUpIn[i]);
    }

    for (int i = 0; i < _numOfLateralIn; i++) {
      register(_rpLateralIn[i]);
    }
  }

  /**
   * <japanese>
   * hNX̑SM[q̓o^ <p>
   *
   * ̃\bh́A̗vffׂĂ̑M[q
   * <a href="OhElement.html#method_summary">OhElement#register(Object)</a>
   * <a href="OhElement.html#method_summary">OhElement#register(Object, int)</a>
   * ɂo^܂B<p>
   *
   * o^ĂȂ[q͐ڑłȂ̂őMɗpł܂B
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public void registerSendPorts() throws OhError {
    register(_sp);
  }

/*
 * <japanese>
 * 6. p^Eԗʂǂݍރ\bh setParameter(),
 *    setInitialState() `B
 * </japanese>
 */

  /**
   * <japanese>
   * p^̐ݒ <p>
   *
   * ̓Xg[ aBr Ŏw肳t@Cf[^ǂ݂Ƃ
   * ăp^ݒ肵܂B
   * </japanese>
   *
   * @param aBr an <code>OhBufferedReader</code> value
   * @exception OhError if an error occurs
   */
  public void setParameter(OhBufferedReader aBr) throws OhError {
    double dt = 0.0;
    try {
      aBr.skipWhites();
      _a = Double.parseDouble(aBr.readWord());
      _m = Double.parseDouble(aBr.readWord());
      _lengthOfStream = Double.parseDouble(aBr.readWord());
      _numOfSepStream = Integer.parseInt(aBr.readWord());
      dt = Double.parseDouble(aBr.readWord());
      _dt = new OhTime(dt);
      _theta = Double.parseDouble(aBr.readWord());
      _qLevel = Double.parseDouble(aBr.readWord());
      _epsilon = Double.parseDouble(aBr.readWord());
    }
    catch (IOException e) {
      throw new OhError(e.toString() + NEWLINE + oPrintString("\t"));
    }
    catch (NumberFormatException e) {
      throw new OhError(e.toString() + NEWLINE + oPrintString("\t"));
    }

    _qCrnt = new double[_numOfSepStream + 1];
    _qNext = new double[_numOfSepStream + 1];

    try {
      if (_numOfUpIn > 0) {  //; 㗬[̗ꍇAf[^n̏d݂ǂݎB
        for (int i = 0; i < _numOfUpIn; i++) {
          aBr.skipWhites();
          _cUpIn[i] = Double.parseDouble(aBr.readWord());
        }
      }

      if (_numOfLateralIn > 0) {  //; ̗ꍇAf[^n̏d݂ǂݎB
        for (int i = 0; i < _numOfLateralIn; i++) {
          aBr.skipWhites();
          _cLateralIn[i] = Double.parseDouble(aBr.readWord());
        }
      }
    }
    catch (IOException e) {
      throw new OhError(e.toString() + NEWLINE + oPrintString("\t"));
    }
    catch (NumberFormatException e) {
      throw new OhError(e.toString() + NEWLINE + oPrintString("\t"));
    }

    _dx = _lengthOfStream / _numOfSepStream;
  }

  /**
   * Sets the initial state based on the data which are read in
   * from <b>aBr</b>.<p>
   *
   * <japanese>
   * ԗʂ̐ݒ <p>
   *
   * ̓Xg[ aBr Ŏw肳t@Cf[^ǂ݂Ƃ
   * ďԗʂݒ肵܂B
   * </japanese>
   *
   * @param aBr an <code>OhBufferedReader</code> value
   * @exception OhError if an error occurs
   */
  public void setInitialState(OhBufferedReader aBr) throws OhError {  //;(*3)
    /* set initial state */

    int  year = 0;
    long tsec = 0L;
    double frac = 0.0;

    // System.out.println("Kinematic#setInitialState#1");
    
    try {
      aBr.skipWhites();
      year = Integer.parseInt(aBr.readWord());
      tsec = Long.parseLong(aBr.readWord());
      frac = Double.parseDouble(aBr.readWord());
    }
    catch (IOException e) {
      // System.out.println("Kinematic#setInitialState#2");
      throw new OhError("Error in setInitialState()." + NEWLINE +
       oPrintString("\t"));
    }
    catch (NumberFormatException e) {
      // System.out.println("Kinematic#setInitialState#3");
      throw new OhError("Error in setInitialState().",
        oPrintString("\t"));
    }

    if (year > 0) {
      _currentTime = new OhTimePoint(year, tsec, frac);
    }

    int dataType;  // 0:ONLY_DOWN_STREAM_END 1:EVERY_SECTION

    try {
      aBr.skipWhites();
      dataType = Integer.parseInt(aBr.readWord());
    }
    catch (IOException e) {
      throw new OhError(oPrintString("\t"));
    }
    catch (NumberFormatException e) {
      throw new OhError(oPrintString("\t"));
    }

    if (dataType == 0)
      _isdt = InitialStateDataType.ONLY_DOWN_STREAM_END;
    else
      _isdt = InitialStateDataType.EVERY_SECTION;

    try {
      if (_isdt == InitialStateDataType.ONLY_DOWN_STREAM_END) {
        aBr.skipWhites();
        _qCrnt[_numOfSepStream] = Double.parseDouble(aBr.readWord());
      }
      else {
        for (int i = 1; i <= _numOfSepStream; i++) {
          _qCrnt[_numOfSepStream] = Double.parseDouble(aBr.readWord());
        }
      }
    }
    catch (IOException e) {
      throw new OhError(oPrintString("\t"));
    }
    catch (NumberFormatException e) {
      throw new OhError(oPrintString("\t"));
    }
  }

/*
 * <japanese>
 * 7. o̓\bh initialOutput() `B
 * </japanese>
 */

  /**
   * Outputs when the calculation starts. <p>
   *
   * @return a <code>boolean</code> value
   * @exception OhError if an error occurs
   */
  public boolean initialOutput() throws OhError {
    // initial output 

    if (_numOfUpIn == 0) {
      _qCrnt[0] = 0.;
      if (_isdt == InitialStateDataType.ONLY_DOWN_STREAM_END) {
        int i = 0;
        for (i = 1; i < _numOfSepStream; i++) {
          _qCrnt[i] = _qCrnt[_numOfSepStream] * i / (double)_numOfSepStream;
        }
      }
    }
    else {
      double qIn = 0.;
      int j;
      for (j = 0; j < _numOfUpIn; j++) {
        if (_rpUpIn[j].canYouGetData(_currentTime) == false) {
          return false;
        }
        qIn += _cUpIn[j] * _rpUpIn[j].getData(_currentTime);
      }

      if (_isdt == InitialStateDataType.ONLY_DOWN_STREAM_END) {
        int i;
        for (i = 0; i < _numOfSepStream; i++) {
          _qCrnt[i] = ( _qCrnt[_numOfSepStream] * i + qIn * (_numOfSepStream - 1))
                      / (double)_numOfSepStream;
        }
      }
    }

    _sp.setCommentString(
      "(Year Tsec(sec) Frac(sec)) " +
      "discharge(m**3/sec)");

    _sp.sendData(_currentTime, _qCrnt[_numOfSepStream]);

    return true;
  }

/*
 * <japanese>
 * 8. ŏIԗʂރ\bh saveTerminalState() `
 *    B
 * </japanese>
 */

  /**
   * <japanese>
   * ŏIԂۑ郁\bh
   * </japanese>
   *
   * @param aPw a <code>PrintWriter</code> value
   */
  public void saveTerminalState(PrintWriter aPw) {    //; (*4)
    /* output terminal state into a file */

    aPw.print(_currentTime + " # _currentTime");
    aPw.print(NEWLINE);
    aPw.print("1 # EVERY_SECTION");
    aPw.print(NEWLINE);

    for (int i = 1; i <= _numOfSepStream; i++) {
      aPw.print(_qCrnt[i] + " # Section " + i);
      aPw.print(NEWLINE);
    }
  }

/*
 * <japanese>
 * 9. vZԊԊuvZ郁\bh calculateTimeStep() `
 *    B
 * </japanese>
 */

  /**
   * Returns a time step for next calculation. The current
   * implementation of the method returns a fixed time _dt.<p>
   *
   * <japanese>
   * vZԊԊuԂ܂B̎ł́AŒ肵 _dt 
   * ܂B
   * </japanese>
   *
   * @return a <code>long</code> value
   * @exception OhError if an error occurs
   */
  public OhTime calculateTimeStep() throws OhError {
    return _dt;    // "_dt" has been read in from pmt file.
  }

/*
 * <japanese>
 * 10. vZ邱Ƃł邩𔻒f郁\bh canYouCalculate() 
 *     `B
 * </japanese>
 */

  /**
   * Judges whether the next calculation is possible or not.<p>
   *
   * <japanese>
   * vZ\̔f <p>
   *
   * <a href="OhElement.html#field_summary">OhElement.html</a> NX
   * \bh <a href="OhElement.html#method_summary">OhElement#getNecessaryTimeFrom()</a>,
   * <a href="OhElement.html#method_summary">OhElement#getNecessaryTimeTo()</a> 
   * `鎞Ԃ̊Ԃ̃f[^擾łΌvZ\Ɣf܂B
   * <a href="OhElement.html#method_summary">OhElement#getNecessaryTimeFrom()</a> \bh
   * _currentTime ԂA
   * <a href="OhElement.html#method_summary">OhElement#getNecessaryTimeTo()</a> \bh
   * _currentTime + _timeStep ̒lԂ܂B<p>
   *
   * f[^擾s\łƕԎĂM[qɑ΂Čv
   * Zvo܂B̓Iɂ́AM[q calReq() bZ[
   * W𑗂܂B
   * </japanese>
   *
   * @return a <code>boolean</code> value
   * @exception OhCannotGetDataError if an error occurs
   * @exception OhError if an error occurs
   */
  public boolean canYouCalculate() throws OhCannotGetDataError, OhError {
    return canYouCalculate0();
  }

/*
 * <japanese>
 * 11. 1Xebv̌vZ郁\bh calculate() `B
 * </japanese>
 */

  /**
   * The solution of power law kimematic wave model by Keith
   * Beven's method.
   *
   * <japanese>
   * <p>
   * Beven @ɂw@ kinematic wave f̉@
   * </japanese>
   *
   * @return a <code>boolean</code> value
   * @exception OhError if an error occurs
   */
  public boolean calculate() throws OhError {
    double dta;
    dta = _timeStep.toDouble();    //; "_timeStep" equals "_dt";

    OhTimePoint nextTime =
      OhTimePoint.add(_currentTime, _timeStep);
    
    /* Lateral inflow */

    double rLin = 0.;
    int i;

    for (i = 0; i < _numOfLateralIn; i++) {
      rLin += _cLateralIn[i] * _rpLateralIn[i].getMeanData(_currentTime, nextTime) / _lengthOfStream;
    }

    /* upstream end */

    _qNext[0] = 0.;

    for (i = 0; i < _numOfUpIn; i++) {
      _qNext[0] += _cUpIn[i] * _rpUpIn[i].getData(nextTime);
    }

    double qest;
    double cold;  //; `dx
    double cnew;  //; XVꂽ`dx

    double m1 = (_m - 1.) / _m;

    for (i = 1; i <= _numOfSepStream; i++) {
      double qbase = (_qCrnt[i-1] + _qCrnt[i] + _qNext[i-1]) / 3.;
      qest = _qNext[i-1] + rLin * _dx;
      if (_qCrnt[i] > qest) {
        qest = _qCrnt[i];
      }
      double qmean = 0.5 * (_qCrnt[i-1] + _qCrnt[i]);
      cold = _a * _m * Math.pow(qmean / _a, m1);


      while (true) {
        double qmean1 = 0.5 * (_qNext[i-1] + qest);
        cnew = _a * _m * Math.pow(qmean1 / _a, m1);

        _qNext[i] = (_qCrnt[i] + _theta * cnew * dta * (_qNext[i-1] / _dx + rLin)
          - (1. - _theta) * cold * dta *
          ((_qCrnt[i] - _qCrnt[i-1]) / _dx - rLin)) / (1. + _theta * cnew * dta / _dx);


        if (_qNext[i] < 0.)
          _qNext[i] = 0.;


        if (Math.abs(_qNext[i] - qest) < (_epsilon * (_qLevel + qbase)))
          break;

        qest = _qNext[i];
      }

      _qNext[i] = qest;
    }

  /* output */

    _sp.sendData(nextTime, _qNext[_numOfSepStream]);

  /* update state variable */

    for (i = 0; i <= _numOfSepStream; i++) {
      _qCrnt[i] = _qNext[i];
    }

    return true;
  }

/*
 * <japanese>
 * 12. ̈ǍvZEo͍Ƃ郁\bh work() `
 *     B
 * </japanese>
 */

  /**
   * <japanese>
   * <p>
   * _timeStep 𓾂Ă calculate() s work0() ̗p
   * ܂B
   * </japanese>
   *
   * @return a <code>boolean</code> value
   * @exception OhCannotGetDataError if an error occurs
   * @exception OhError if an error occurs
   */
  public boolean work() throws OhCannotGetDataError, OhError {
    return work0();
  }

  public String basicInfo() {
    return
      "Powerlaw kinematic wave model"
      ;
  }
  
} // PowerLawKinematicWave
