/**
 * @(#)  SimpleDam.java
 *
 */

package ohymosj;

import ohymosj.*;
import java.util.*;
import java.io.*;


/**
 * Simple dam model.<p>
 *
 * <b>classname</b>: <br>
 * SimpleDam<p>
 *
 * <b>Receiveports</b>: <br>
 * OhReceiveDouble("rp", i), i = 0, ..., nrp-1. nrp is variable.
 * Inflow from upstream channels into the dam. The dimension of
 * cRp[i]*rpi must be m**3/sec where rpi is the value obtained
 * from the port ("rp", i). See: <a href="OhReceiveDouble.html">OhReceiveDouble</a> <p>
 *
 * <b>Sendports</b>: <br>
 * OhSendDouble("sp", 0) -- discharge (m**3/sec) from the dam,
 * brokenline. <br>
 * OhSendDouble("sp", 1) -- storage (m**3) of the dam, brokenline.
 * <br>
 * See: <a href="OhSendDouble.html">OhSendDouble</a> <p>
 *
 * <b>parameters</b>: 
 * <dl>
 *  <dt> qPass <dd>
 *  pass discharge (m^3/sec)
 *  <dt> ratio <dd>
 *  discharge rate to surpass from qPass (-)
 *  <dt> dt    <dd>
 *  time step (sec)
 *  <dt> cRp[i], i=0, ..., nrp <dd>
 *  multiplier to inflow dischage to port ("rp", i), i = 0, ...,
 *  nrp-1.
 * </dl>
 *
 * <b>state variables</b>:<br>
 * the OhTimePoint variable and the storage in the dam reservoir
 * (m**3)
 * <p>
 *
 * The <b>init(String, int, OhBufferedReader)</b> method looks like:
 *
 * <pre>
 *  public void init(String aObjName, int aObjNum,
 *    OhBufferedReader aBr) throws OhError {
 *    int nrp;
 *    try {
 *      nrp = Integer.parseInt(aBr.readWord());
 *    }
 *    catch (IOException e) { ...; }
 *    catch (NumberFormatException e) { ...; }
 *    ...
 *  }
 * </pre>
 * 
 * The <b>setParemeter(OhBufferedReader)</b> method used for
 * reading parameters from a file looks like:
 *
 * <pre>
 *  public void setParameter(OhBufferedReader aBr)
 *    throws OhError {
 *    double dt = 0.0;
 *    try {
 *      _qPass = Double.parseDouble(aBr.readWord());
 *      _ratio = Double.parseDouble(aBr.readWord());
 *      dt = Double.parseDouble(aBr.readWord());
 *      _dt = new OhTime(dt);
 *    }
 *    catch (IOException e) { ...; throw new OhError(str); }
 *    catch (NumberFormatException e) {
 *      ...; throw new OhError(str); }
 *    if (_qPass < 0.0) { ...; throw new OhError(str); }
 *    if (_ratio < 0.0 || _ratio > 1.0) { ...; throw new OhError(str); }
 *    if (_dt.toDouble() < OhTime.NEGLIGIBLY_SMALL_TIME) {
 *      ...; throw new OhError(str); }
 *    if (_numberOfRp > 0) {  //; Weights for inflow receive ports.
 *      for (int i = 0; i < _numberOfRp; i++) {
 *        try {
 *          _cRp[i] = Double.parseDouble(aBr.readWord());
 *        }
 *        catch (IOException e) { ...; throw new OhError(str); }
 *        catch (NumberFormatException e) {
 *          ...; throw new NumberFormatException(str); }
 *        }
 *      }
 *    }
 *  }
 * </pre>
 * Note: <a href="OhTime.html#field_summary">OhTime.NEGLIGIBLY_SMALL_TIME</a> <p>
 *
 * The <b>setInitialState(OhBufferedReader)</b> method used for
 * reading initial states from a file looks like:<p>
 *
 * <pre>
 *  public void setInitialState(OhBufferedReader aBr)
 *    throws OhError {
 *    int  year = 0; long tsec = 0L; frac = 0.0;
 *    try {
 *      year = Integer.parseInt(aBr.readWord());
 *      tsec = Long.parseLong(aBr.readWord());
 *      frac = Double.parseDouble(aBr.readWord());
 *    }
 *    catch (IOException e) { ...; }
 *    catch (NumberFormatException e) { ...; }
 *    if (year > 0) {
 *      _currentTime = new OhTimePoint(year, tsec, frac);
 *    }
 *    try {
 *      _storageCrnt = Double.parseDouble(aBr.readWord());
 *    }
 *    catch (IOException e) { ...; }
 *    catch (NumberFormatException e) { ...; }
 *  }
 * </pre>
 *
 *
 * <japanese>
 * SimpleDam NX́AȒPȃ_rfłB<p>
 *
 * <b>M[q</b>:
 * ̃_rf́A̗ʂ󂯂Ƃ邽߂̎M
 * [qĂ܂BM[q̌ nrp ́ACӂɕύXł܂B
 * M[q́A
 *
 * <pre>
 *   OhReceiveDouble("rp", i), i = 0, ..., nrp-1
 * </pre>
 *
 * łB(Q <a href="OhReceiveDouble.html">OhReceiveDouble</a>) M[qMf[
 * ^ɏ搔 cRp(i) |̂AʂƂȂ܂BcRp(i)*rpi 
 * ̒PʂAm^3/sec ł悤ɂ܂BArpi ́AiԖڂ
 * M[q擾lłB
 * <p>
 *
 * <b>M[q</b>:
 * M[q́A2A
 *
 * <pre>
 *   OhSendDouble("sp", i), i = 0, 1
 * </pre>
 *
 * łB(Q <a href="OhSendDouble.html">OhSendDouble</a>)BA
 * <a href="DataConnectionType.html">DataConnectionType</a> ́Abrokenline łBM[q 
 * ("sp", 0) 瑗Mf[^̓_̕ʂŁAPʂ́A
 * m**3/sec łBM[q ("sp", 1) 瑗Mf[^́A_
 * ̒ʂŁAPʂ́Am**3 łB
 * <p>
 *
 * <b>p^</b>:
 * <dl>
 *  <dt> qPass <dd>
 *  ʉߗ (m^3/sec)B_ւ̗ʂ̑aA̒ʉߗʂ
 *  Ƃ̂܂ܕ܂B_ւ̗ʂ̑aAqPass 
 *  傫ƁAqPass + ratio * (qsum - qPass) ܂B
 *  Aqsum ͗ʂ̑ałB
 *
 *  <dt> ratio <dd>
 *  0  1 ̊Ԃ̖̒萔B_̕ʂvẐɎg
 *  p܂B
 *
 *  <dt> dt    <dd>
 *  vZԃXebv (sec)
 *  <dt> cRp[i], i=0, ..., nrp-1 <dd>
 *  M[q ("rp", i) Mf[^ւ̏搔B
 * </dl>
 *
 * <b>ԗ</b>:
 * <a href="OhTimePoint.html">OhTimePoint</a> ^̎ƒr̒ (m**3)<p>
 *
 * nrp ́A
 * <a href="SimpleDam.html#method_summary">init(String, int, OhBufferedReader)</a> Őݒ
 * ܂Bp^́A<a href="SimpleDam.html#method_summary">setParameter()</a> œǂ
 * ܂܂Bԗʂ́A<a href="SimpleDam.html#method_summary">setInitialState()</a> 
 * ǂݍ܂܂B
 * </japanese>
 *
 * 
 * @author satoh, shiiba
 * @version $Revision: 1.26 $ $Date: 2004/01/14 19:24:54 $
 */
public class SimpleDam extends OhElement
{
  // @begin {instance variable or static variable} -----------------------

  // 1. Receive ports and send ports.
  
  OhReceiveDouble[] _rP;

  OhSendDouble _sP0;
  OhSendDouble _sP1;

  // 2. Structural parameters.
  
  int _numberOfRp;

  // 3. Parameters.
  
  double _qPass;

  double _ratio;

  OhTime _dt;

  double[] _cRp;

  // 4. State variables.

  double _storageCrnt;

  double _qCrnt;

  // 5. Working variable.

  double _storageNext;

  double _qNext;

  double _qInCrnt;

  double _qInNext;


  // @end {instance variable or static variable} -------------------------

  // @begin {initialization} ---------------------------------------------

  public void init(String aObjName, int aObjNum, String aDataStr, int aNrp) throws OhError {
    _numberOfRp = aNrp;

    if (_numberOfRp < 1)
      _numberOfRp = 1;

    _rP  = new OhReceiveDouble[_numberOfRp];
    _cRp = new double[_numberOfRp];

    for (int i = 0; i < _numberOfRp; i++) {
      _rP[i] = new OhReceiveDouble();     //; Added by Satoh. (2003.12.10)
      _rP[i].init("rp", i);
    }

    super.init(aObjName, aObjNum, 0, aDataStr);
  }

  
  public void init(String aObjName, int aObjNum,
    OhBufferedReader aBr) throws OhError {
    int nrp;
    try {
      nrp = Integer.parseInt(aBr.readWord());
    }
    catch (IOException e) {
      throw new OhError(oPrintString("\t"));
    }
    catch (NumberFormatException e) {
      throw new OhError(oPrintString("\t"));
    }

    init(aObjName, aObjNum, null, nrp);
  }
  
  // @end {initialization} -----------------------------------------------
  
  // @begin {constructors} -----------------------------------------------

  public SimpleDam() throws OhError {
    super();

    _rP  = null;
    _sP0 = new OhSendDouble("sp", 0);
    _sP1 = new OhSendDouble("sp", 1);
    _numberOfRp = 0;
    _cRp = null;

    _className = "SimpleDam";
  }


  public SimpleDam(String aObjName, int aObjNum, String aDataStr, int aNrp) throws OhError {
    super();

    _rP  = null;
    _sP0 = new OhSendDouble("sp", 0);
    _sP1 = new OhSendDouble("sp", 1);
    _cRp = null;

    _className = "SimpleDam";

    init(aObjName, aObjNum, aDataStr, aNrp);
  }
  
  // @end {constructors} -------------------------------------------------

  // @begin {methods} ----------------------------------------------------

  // 1. Methods about receive ports and send ports.

  public void registerReceivePorts() throws OhError {   //(*1)
    for (int i = 0; i < _numberOfRp; i++) {
      register(_rP[i]); // might throw an OhError.
    }
  }

  public void registerSendPorts() throws OhError {
    register(_sP0);    // might throw an OhError.
    register(_sP1);    // might throw an OhError.
  }


  // 2. Parameter setting and initialization of states.

  
  /**
   * Read the values of the parameters from the specified
   * stream.<p>
   *
   * <japanese>
   * w肳ꂽXg[Ap^̒lǂݎ܂B
   * ǂݎp^̒ĺA̒ʂłB<p>
   *
   * <dl>
   *  <dt> qPass <dd>
   *  ʉߗ (m^3/sec)B_ւ̗ʂ̑aA̒ʉߗʂ
   *  Ƃ̂܂ܕ܂B_ւ̗ʂ̑aAqPass 
   *  傫ƁAqPass + ratio * (qsum - qPass) ܂B
   *  Aqsum ͗ʂ̑ałB
   *
   *  <dt> ratio <dd>
   *  0  1 ̊Ԃ̖̒萔B_̕ʂvẐɎg
   *  p܂B
   *
   *  <dt> dt    <dd>
   *  vZԃXebv (sec)
   *  <dt> cRp[i], i=0, ..., nrp-1 <dd>
   *  M[q ("rp", i) Mf[^ւ̏搔B
   * </dl>
   *
   * </japanese>
   *
   * @param aBr an <code>OhBufferedReader</code> value
   * @exception OhError if an error occurs
   */
  public void setParameter(OhBufferedReader aBr) throws OhError {  //(*2)

    double dt = 0.0;
    try {
      _qPass = Double.parseDouble(aBr.readWord());
      _ratio = Double.parseDouble(aBr.readWord());
      dt = Double.parseDouble(aBr.readWord());
      _dt = new OhTime(dt);
    }
    catch (IOException e) {
      throw new OhError(oPrintString("\t"));
    }
    catch (NumberFormatException e) {
      throw new OhError(oPrintString("\t"));
    }

    if (_qPass < 0.0) {
      throw new OhError(
      "_qPass should be non-negative but is " + _qPass);
    }

    if (_ratio < 0.0 || _ratio > 1.0) {
      throw new OhError(
        "_ratio should be in [0.0, 1.0] but it is " + _ratio
      );
    }

    if (_dt.toDouble() <  OhTime.NEGLIGIBLY_SMALL_TIME) {
      throw new OhError(
        "_dt should be greater than or equal to " +
        OhTime.NEGLIGIBLY_SMALL_TIME +
        " but it is " + _dt
      );
    }

    if (_numberOfRp > 0) {  //; Weights for inflow receive ports.
      for (int i = 0; i < _numberOfRp; i++) {
        try {
          _cRp[i] = Double.parseDouble(aBr.readWord());
        }
        catch (IOException e) {
          StringWriter sw = new StringWriter();
          PrintWriter  pw = new PrintWriter(sw);
          oPrint(pw, "\t");
          String str = sw.toString();
          pw.close();
          throw new OhError(str);
        }
        catch (NumberFormatException e) {
          throw new NumberFormatException(oPrintString("\t"));
        }
      }
    }
  }


  public void saveParameter (PrintWriter aPw) {
    aPw.print(_qPass + " " + _ratio + " " + _dt + NEWLINE);
    for (int i = 0; i < _numberOfRp; i++) {
      aPw.print(_cRp[i] + " ");
    }
    aPw.print(NEWLINE);
  }

  public void setInitialState(OhBufferedReader aBr) throws OhError {  //(*3)
    int  year = 0;
    long tsec = 0L;
    double frac = 0.0;
    
    try {
      year = Integer.parseInt(aBr.readWord());
      tsec = Long.parseLong(aBr.readWord());
      frac = Double.parseDouble(aBr.readWord());

//    System.out.println("year = " + year);  ł́u2000v
    }
    catch (IOException e) {
      throw new OhError("Error in setInitialState()." + NEWLINE +
       oPrintString("\t"));
    }
    catch (NumberFormatException e) {
      throw new OhError("Error in setInitialState().",
        oPrintString("\t"));
    }
    finally {
      try { aBr.close(); } catch (IOException e) {}
    }
    
    if (year > 0) {
      _currentTime = new OhTimePoint(year, tsec, frac);
    }

    try {
      _storageCrnt = Double.parseDouble(aBr.readWord());
    }
    catch (IOException e) {
      throw new OhError(oPrintString("\t"));
    }
    catch (NumberFormatException e) {
      throw new OhError(oPrintString("\t"));
    }
    finally {
      try { aBr.close(); } catch (IOException e) {}
    }
  }
  

  public void saveTerminalState(PrintWriter aPw) {
    aPw.print(_currentTime + " # _currentTime");
    aPw.print(NEWLINE);
    aPw.print(_storageCrnt + " # _storageCrnt (m^3)");
    aPw.print(NEWLINE);
  }


  // 3. Calculation.

  public boolean initialOutput() throws OhError {
    _qInCrnt = 0.0;

    if (_numberOfRp > 0) {
      int j = 0;

      for(j = 0; j < _numberOfRp; j++) {
        if (_rP[j].canYouGetData(_currentTime) == false) {
          return false;
        }

        _qInCrnt += _cRp[j] * (_rP[j].getData(_currentTime));
      }
    }

    if (_qInCrnt <= _qPass) {
      _qCrnt = _qInCrnt;
    }
    else {
      _qCrnt = _qPass + _ratio * (_qInCrnt - _qPass);
    }


    _sP0.setCommentString(
      "OhTimePoint.Year OhTimePoint.Tsec(sec) " +
      "OhTimePoint.Frac(sec) discharge(m**3/sec)");
    _sP0.sendData(_currentTime, _qCrnt);

    _sP1.setCommentString(
      "OhTimePoint.Year OhTimePoint.Tsec(sec) " +
      "OhTimePoint.Frac(sec) storage(m**3)");
    _sP1.sendData(_currentTime, _storageCrnt);

    return true;
  }


  public OhTime calculateTimeStep() throws OhError {
    return _dt;  //; "_dt" is read from pmt file.
  }

  public boolean canYouCalculate() throws OhCannotGetDataError, OhError {
    return canYouCalculate0();
  }

  public boolean work() throws OhCannotGetDataError, OhError {
    return work0();
  }



  
  public boolean calculate() throws OhError {  //(*6)
    double dta;
    dta = _timeStep.toDouble();    //; "_timeStep" equals "_dt";

    _qInNext = 0.0;

    OhTimePoint nextTime =
      OhTimePoint.add(_currentTime, _timeStep);

    if (_numberOfRp > 0) {
      int j = 0;
      for (j = 0; j < _numberOfRp; j++) {
        if (_rP[j].canYouGetData(nextTime) == false) {
          return false;
        }
        _qInNext += _cRp[j] * (_rP[j].getData(nextTime));
      }
    }

    if (_qInNext <= _qPass) {
      _qNext = _qInNext;
    }
    else {
      _qNext = _qPass + _ratio * (_qInNext - _qPass);
    }
    
    _storageNext = _storageCrnt + dta * (_qInCrnt + _qInNext) * 0.5 - dta * (_qCrnt + _qNext) * 0.5;


    //; Output.
    //;
    _sP0.sendData(nextTime, _qNext);
    _sP1.sendData(nextTime, _storageNext);


    //; Standard output. (This is not especially required.)
    //;
    
    //; Update state variables.
    //;
    _storageCrnt = _storageNext;
    _qCrnt       = _qNext;
    _qInCrnt     = _qInNext;

    return true;
  }


  public String basicInfo() {
    return
     "Simple dam model"
    ;

  }
  
  // @begin {methods} ----------------------------------------------------

} // SimpleDam
