/**
 * @(#) Class to express DataConnectionType.
 */

package ohymosj;

import java.util.*;
import java.io.*;

/**
 * The DataConnectionType class
 * is a class to express data connection type. <p>
 *
 * The following data connection types are defined.
 *
 * <pre>
 *  BROKENLINE
 *    The data are connected with broken line.
 *  STEP
 *    The value is taken constant and equal to the current data
 *    value from the previous data time till the current time.
 *  PULSE
 *    The value between data times is taken to be zero.
 *  NOINTP
 *    No interpolation.
 * </pre>
 *
 * <japanese>
 * <p>
 * f[^Ԍ^\NXB<p>
 *
 * ^ɈSȒ萔̃NXłB<p>
 *
 * <ul>
 *  <li> BROKENLINE: ܂B
 *  <li> STEP: Xebv. Õf[^pbÑf[^
 
 *       Ỹf[^pbÑf[^̊Ԃł́AỸf[^
 *       pbÑf[^lf[^lłƂ܂BỸf[
 *       ^̃f[^l͂̃f[^pbÑf[^lƂ܂B
 *       ܂A̃f[^pbN (tp1, v1), (tp2, v2) A
 *       ĂƂA(tp1, tp2] Œl v2 Ƃ܂B
 *  <li> PULSE : Pulse.
 *  <li> NOINTP : ԂłȂB
 * </ul>
 *
 * DataConnectionType ^̃IuWFNg̔zƂ܂B
 *
 * <pre>
 *   Vector dctvec = new Vector();
 *   dctvec.addElement(DataConnectionType.BROKENLINE);
 *   dctvec.addElement(DataConnectionType.STEP);
 * </pre>
 *
 * ̗ł́Adctvec ̑傫 2 ŁA1 BROKENLINE, 2
 *  STEP ɂȂ܂B
 *
 * 0Ƃ肾ɂ́Adctvec.get(0) Ƃ܂B
 * </japanese>
 *
 * @author satoh, shiiba
 * @version $Revision: 1.14 $ $Date: 2004/01/19 07:51:54 $
 */
public final class DataConnectionType implements Serializable
{
  // @begin {instance variable or static variable} -------------
  
  private static int nextOrdinal = 0;
  private final  int ordinal     = nextOrdinal++;
  
  /**
   * BROKENLINE: <br>
   * The data are connected with broken line.
   *
   */
  public static final DataConnectionType BROKENLINE =
    new DataConnectionType();
  
  /**
   * STEP: <br>
   * The value is taken constant and equal to the current data
   * value from the previous data time till the current time.
   *
   */
  public static final DataConnectionType STEP       =
    new DataConnectionType();
  
  /**
   * PULSE: <br>
   * The value between data times is taken to be zero.
   *
   */
  public static final DataConnectionType PULSE =
    new DataConnectionType();
  
  /**
   * NOINTP: <br>
   * No interpolation.
   *
   */
  public static final DataConnectionType NOINTP =
    new DataConnectionType();
  
  private static final DataConnectionType[] PRIVATE_VALUES =
    {BROKENLINE, STEP, PULSE, NOINTP};
  
  //; added by Satoh (2003.11.18)
  private static final String DCTStr[] =
    {"BROKENLINE", "STEP", "PULSE", "NOINTP"};
  
  // @end {instance variable or static variable} ---------------
  
  
  // @begin {constructors} -------------------------------------
  
  private DataConnectionType() {}

  // @end {constructors} ---------------------------------------

  
  // @begin {methods} ------------------------------------------

  //; added by Satoh (2003.11.18)
  public static int getNumOfDCT () {
    return DCTStr.length;
  }
  
  //; added by Satoh (2003.11.18)
  public static DataConnectionType getDctType(int k) {
    return PRIVATE_VALUES[k];
  }
  
  //; added by Satoh (2003.11.18)
  public static String getDctTypeStr(int k) {
    return DCTStr[k];
  }

  /**
   * Returns the index corresponding to this DataConnectionType
   * object.
   *
   * <japanese>
   * Ỹf[^Ԍ^IuWFNg̃CfbNXԂ܂B
   * ΁ADataConnectionType.BROKENLINE ɑ΂ẴbZ[W
   * 𑗂ƁA0 Ԃ܂B
   * </japanese>
   *
   * @return an <code>int</code> value
   */
  public int toInt() {
    for (int k = 0; k < DCTStr.length; k++) {
      if (this == PRIVATE_VALUES[k]) {
        return k;
      }
    }
    return -1; // This should not happen.
  }
  
  private Object readResolve() throws ObjectStreamException {
    return PRIVATE_VALUES[ordinal];
  }

  /**
   * Converts the String argument <b>aStr</b> into the
   * (Ref "DataConnectionType") vector and returns it.
   *
   * For example, String " 2 1 0" is converted into Vector[2] =
   * { DataConnectionType.STEP, DataConnectionType.BROKENLINE }
   *
   * <japanese>
   * <p>
   *  aStr f[^Ԍ^zǂݎA߂lƂĕ
   * ܂B
   * 
   * </japanese>
   *
   * @param aStr a <code>String</code> value
   * @return a <code>Vector</code> value
   * @exception OhError if an error occurs
   */
  public static Vector parseDctVec(String aStr) {
    int n = 0;    // The size of the vector.

    Vector dctVec = new Vector();

    StringTokenizer st = new StringTokenizer(aStr);

    try {
      n = Integer.parseInt(st.nextToken());
    }
    catch (NumberFormatException e) {
      throw e;
    }

    if (n < 0)
      throw new NumberFormatException("negative vector size");

    for (int i = 0; i < n; i++) {
      int k = Integer.parseInt(st.nextToken());
      if (0 <= k && k < DataConnectionType.getNumOfDCT())
        dctVec.addElement(DataConnectionType.getDctType(k));
      else {
        throw new NumberFormatException(
          "invalid DataConnectionType number: " + k
        );
      }
    }

    return dctVec;
  }

  // @end {methods} --------------------------------------------
}
