/**
 * @(#) OhElement; base class of element models.
 */

package ohymosj;

import java.util.*;
import java.io.*;

/**
 * OhElement class is an abstract class for creating various
 * element models by derivation from this class.<p>
 *
 * An element model
 *
 * <ul>
 *  <li> has data members for storing parameters, state variables,
 *       receive ports, and send port;
 *  <li> calculates the time change of the state variables using
 *       input data obtained through the receive ports; and
 *  <li> sends output data through the send ports.
 * </ul>
 *
 * In hydrolgic simulations, we need various element models
 * because various processes are involved. <p>
 *
 * The Class Element expresses the common part of those
 * hydrologic element models.
 *
 * <japanese>
 * <p>
 * vff쐬邽߂̒ۃNX OhElement <p>
 *
 * ۃNXłB<p>
 *
 * vff́Ap^, ԗ, M[q, M[qf[^
 * oɋLAp^EԗʁEM[q̓̓f[^p
 * ĐvZsȂAԗʂXVAM[qo̓f[^
 * 𑗐M܂BvZƂ̂Aevffɋʂȍ
 * ƂƋʂłȂƂƂ𕪗Aʕ̂ݒoEKi
 * ̂{^vffłB̃NX́A{^vff\
 * NX`Ă܂B<p>
 *
 * vffLSM[q, SM[q́A_
 * receivePortArray, _sendPortArray ɓo^܂B{^vff
 * ł́A̓o^葱ɊւKisȂAvff̍
 * Ƃ̊ȑfi߂ƂƂɁAevffɓo^葱Ɋ
 * 铯̐M^Ă܂B܂Aevf̑M[q̐
 * ͕nfsȂ܂A̍ƂeՂɂ邽߁Av
 * ffL鑗M[qɊւ錟@\`܂B
 * <p>
 *
 * ʏ́Avff̏ԂXV邽߂ɕKvȃf[^͎
 * M[qoĎ擾悤ɐ݌v܂Aނ𓾂Ȃꍇ́A
 * uڒʐMvɂĕKvȃf[^擾dg݂[U
 * `ł悤ɁAڒʐMvf̔zL邽߂̃f[^
 * o _objectArray pӂĂ܂B
 *
 * </japanese>
 *
 * @author satoh, shiiba
 * @version $Revision: 1.126 $ $Date: 2004/01/16 03:54:54 $
 */
public abstract class OhElement extends OhObject
{
  // @begin {instance variable or static variable} -------------
  
  /**
   * Whether the initial work has been completed or not.
   *
   * <japanese>
   * <p>
   * initialWork() I (true) ۂ (false)B
   * </japanese>
   */
  protected boolean _initialWorkDoneP;
  
  /**
   * Whether the (Ref "#shareInfo()") has been completed or not.
   *
   * <japanese>
   * <p>
   * ShareInfo() I (true) ۂ (false)B
   * </japanese>
   */
  protected boolean _shareInfoDoneP;
  
  /**
   * A list for storing Object. It is used as an object for work
   * at the time of creation of <b>_receivePortArray</b> and
   * <b>_sendPortArray</b>.
   *
   * <japanese>
   * <p>
   * Object ̃XgB_receivePortArray, _sendPortArray ̍쐬
   * ɍƗpƂĎgp܂B
   *
   * (Ref "OhLinkedList")
   * </japanese>
   */
  protected LinkedList _objectList; 
  
  /**
   * Array of "OhReceivePort" of type OhPort[].
   *
   * <japanese>
   * <p>
   * M[q̔zB
   * </japanese>
   */
  protected OhPort[] _receivePortArray;
  
  /**
   * Number of receive ports.
   *
   * <japanese>
   * <p>
   * _receivePortArray ̗vfB
   * </japanese>
   */
  protected int _numberOfReceivePorts;
  
  /**
   * Array of "OhSendPort".
   *
   * <japanese>
   * <p>
   * M[qzB
   * </japanese>
   */
  protected Object[] _sendPortArray;
  
  /**
   * Number of send ports.
   *
   * <japanese>
   * <p>
   * _sendPortArray ̗vfB
   * </japanese>
   */
  protected int _numberOfSendPorts;
  
  /**
   * Array of "OhObject".
   *
   * <japanese>
   * <p>
   * IuWFNgzB<p>
   *
   * ڒʐMvf̔zB
   * </japanese>
   */
  protected OhObject[] _objectArray;
  
  /**
   * Number of objects.
   *
   * <japanese>
   * <p>
   * _objectArray ̗vfB
   * </japanese>
   */
  protected int _numberOfObjects;
  
  /**
   * Data string.
   *
   * <japanese>
   * <p>
   * f[^擾pB
   * </japanese>
   */
  protected String _dataString;
  
  /**
   * Current(Calculated time) time.
   * "OhTimePoint" is defined in "OhTimePoint.java".
   *
   * <japanese>
   * <p>
   * ݁ivZς݁jB
   * </japanese>
   */
  protected OhTimePoint _currentTime;
  
  /**
   * Time step which is expressed by an (Ref "OhTime") object.
   *
   * <japanese>
   * <p>
   * vZPXebv̎Ԃ\킷 (Ref "OhTime") ^̃IuWFN
   * gB(Ref "OhElement") ̃RXgN^ŁAnull ɐݒ肳
   * B
   * </japanese>
   */
  protected OhTime _timeStep;

  /**
   * Time which the state has decided in repeated calculation.
   *
   * <japanese>
   * <p>
   * vZɂďԂm肵Ă鎞B
   * </japanese>
   */
  protected OhTimePoint _fixedTime;
  
  /**
   * Repeated calculation target time. [sec] <p>
   * 
   * It is necessary to make "_fixedTime" and "_targetTime" in
   * agreement with "_currentTime" with the element which
   * carries out repeated calculation at the time of a
   * calculation start. <p>
   *
   * After initial setting of the "_currentTime" is carried out
   * by "setInitialState()", call "void cur2fix2tar()" which the
   * bottom defines or give as direct "_targetTime = _fixedTime
   * = _currentTime;".
   *
   * <japanese>
   * <p>
   * vZڕW [sec] <p>
   *
   * iӁ)
   * vZvfł́AvZJn _fixedTime, _targetTime 
   *  _currentTime ƈvKv܂B
   * OhElement.setInitialState(OhBufferedReader)  _currentTime 
   * ݒ肳ꂽAOhElement.setCur2Fix2Tar() ĂԂA
   *  _targetTime = _fixedTime = _currentTime; ƂĂB
   * </japanese>
   */
  protected OhTimePoint _targetTime;
  
  /**
   * Flag of request of calculation. Initial value is NO. When a
   * cause incalculable although the lower-stream-of-a-river
   * side tended to calculate is in this element, it is made to
   * be set as "_calReqP = true". And, "OhElement" tries to
   * calculate hard at the time of "_calReqP = true".
   *
   * <japanese>
   * <p>
   * vZ悤Ƃ̂ɌvZłȂ̗vfɂ
   * ƂɁA_calReqP = true ɐݒ肳悤ɂ܂BāA
   * OhElement  _calReqP = true ̂Ƃ͌vZ悤ɓw͂
   * B_calReqP ͂̃NXŎn߂ĒǉꂽoϐłB
   * ̃NX̃IuWFNgƂɂ _calReqP  false 
   * ɏ܂B
   * </japanese>
   */
  protected boolean _calReqP;
  
  // @end {instance variable or static variable} ---------------
  
  // @begin {abstract methods} ---------------------------------

  // 1. creation and reference of "_receivePortArray",
  // "_sendPortArray" and "_objectArray"
  
  /**
   * Declares the registerReceivePorts() method as an abstract
   * method. In classes derived from the OhElement class, its
   * definition must be given. <p>
   *
   * The registerReceivePorts() method registers all the
   * receive ports of an inherited class, and all send ports by
   * register(OhObject) register(OhObject, int). <p>
   *
   * The port which is not registered is not applicable to
   * sending and receiving, since it is not connectable. <br>
   * For example, when an inherited class defines the receiving
   * port "abc", if it is performed as follows, "abc" can be
   * registered as a receiving terminal.
   *
   * <pre>
   *   void registerReceivePort() {
   *     register(abc);
   *   }
   * </pre>
   *
   * <japanese>
   * <p>
   * hNX̑SM[q̓o^ <p>
   *
   * ̃\bh́AhNX̑SM[q
   * register(OhObject), register(OhObject, int)
   * ɂo^܂B<p>
   *
   * o^ĂȂ[q͐ڑłȂ̂őMɗpł܂B<p>
   *
   * Ⴆ΁AhNXŎM[q abc `ꍇA
   *
   * <pre>
   *  public void registerReceivePort() {
   *    register(abc);
   *  }
   * </pre>
   *
   * Ƃ΁Aabc M[qƂēo^ł܂B
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public abstract void registerReceivePorts() throws OhError;

  
  /**
   * Declares method "registerSendPorts() as abstract method. In
   * classes derived from class Element, its definition should
   * be given. <p>
   *
   * "registerSendPorts()" registers all the receiving ports of
   * an inherited class, and all sending ports by
   * register(OhObject), and
   * register(OhObject, int). <p>
   *
   * The port which is not registered is not applicable to
   * sending and receiving, since it is not connectable. For
   * example, when an inherited class defines the receiving port
   * "abc", if it is performed as follows, "abc" can be
   * registered as a receiving terminal.
   *
   * <pre>
   *   void registerSendPort() {
   *     register(abc);
   *   }
   * </pre>
   *
   * <japanese>
   * <p>
   * hNX̑SM[q̓o^ <p>
   *
   * ̃\bh́AhNX̑SM[q
   * register(Object), register(Object, int)
   * ɂo^܂B<p>
   *
   * o^ĂȂ[q͐ڑłȂ̂őMɗpł܂B<p>
   *
   * Ⴆ΁AhNXőM[q abc `ꍇA
   *
   * <pre>
   *  public void registerSendPort() {
   *    register(abc);
   *  }
   * </pre>
   *
   * Ƃ΁Aabc 𑗐M[qƂēo^ł܂B
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public abstract void registerSendPorts() throws OhError;

  // 2. connection function

  // 3. numerical initialization
  
  /**
   * Declares the setParameter() method as abstract method. In
   * classes derived from the Element class, its definition
   * should be given. <p>
   *
   * Parameters are read in from the input stream
   * <b>aBr</b>. Programmers who create element models
   * need to write this method according to the property of each
   * inherited class. Data is read in from the file specified by
   * the input stream <b>aBr</b>.<p>
   *
   * <japanese>
   * <p>
   * p^ݒ肷郁\bhAehNXɉĒ`
   * B̓Xg[ aBr Ŏw肳t@Cf[^ǂ
   * Ƃăp^ݒ肵܂B<p>
   * </japanese>
   *
   * @param aBr a <code>OhBufferedReader</code> value
   * @exception OhError if an error occurs
   */
  public abstract void setParameter(OhBufferedReader aBr)
    throws OhError;
  
  /**
   * Declares the setInitialState() method as abstract
   * method. In classes derived from the Element class, its
   * definition must be given. <p>
   *
   * <japanese>
   * <p>
   * ԗʂݒ肷郁\bhAehNXɉĒ`
   * ܂B̓Xg[ aBr Ŏw肳t@Cf[^
   * ݂Ƃďԗʂݒ肵܂B<p>
   *
   * ŏIԂt@CɕۑvOƂ܂悤
   * ɂĂB
   * </japanese>
   *
   * @param aBr a <code>OhBufferedReader</code> value
   * @exception OhError if an error occurs
   */
  public abstract void setInitialState(OhBufferedReader aBr)
    throws OhError;

  // 4. time about a simulation

  // 5. execution of calculation
  
  /**
   * An initial output. In classes derived from class Element,
   * its definition should be given. <p>
   *
   * The output in "_currentTime" is defined immediately after
   * deciding the amount of initial states. Time and an output
   * value are written in a sending port. "true" is returned
   * when the output in "_currentTime" is able to be
   * calculated. "false" is returned when the output of
   * "_currentTime" is not able to be calculated by required
   * input data not being obtained.
   *
   * <japanese>
   * <p>
   * ̏o <p>
   *
   * ԗʂ߂ɁA_currentTime ł̏o͂`܂B
   * M[qɎƏo͒l݂܂B
   * _currnet_time ł̏o͂vZłꍇ true Ԃ܂B
   * Kvȓ̓f[^ĂȂȂǂāA
   * _currentTime ̏o͂vZłȂꍇ false Ԃ
   * B<p>
   *
   * vfɂẮAsetInitialState() ̒
   * ̏o͂悤ɂs悢m܂B
   * Ƃ΁Alݒ̂߂ɗ^镶ɂāA
   * ̏o͂̂ς肷ꍇȂǂłB
   * ꍇ́AinitialOutput() ͉Ȃ悤ɒ`āA
   * setInitialState() ŏ̏o͂悤ɒ`Ă\
   * B
   * </japanese>
   *
   * @return a <code>boolean</code> value
   * @exception OhError if an error occurs
   */
  public abstract boolean initialOutput() throws OhError;
  //;
  //; "OhError" is for SimpleDam class
  //;

  
  /**
   * Calculation of a time step. In classes derived from class
   * Element, its definition should be given. <p>
   *
   * Please calculate "_timeStep" according to a model in each
   * inherited class, and define the function which returns it
   * as a return value. That is, it defines under what
   * conditions "_timeStep" is made into what value.
   *
   * <japanese>
   * <p>
   * ^CXebv̌vZ <p>
   *
   * ehNXŃfɉ _timeStep [sec] vZA
   * ߂lƂĕԂ\bh`ĉB<p>
   * 
   * ܂A_timeStep ǂ̂悤ȏŁAǂ̂悤Ȓlɂ
   * `܂B<p>
   *
   * Ⴆ΁A_timeStep l 10.3 [sec] ɐݒ肷ꍇA
   *
   * <pre>
   *  public OhTime calculateTimeStep() {
   *    return new OhTime(10, 0.3);
   *  }
   * </pre>
   *
   * ƒ`ĉB<p>
   * 
   * ܂AL̂悤Ƀoϐ _k ̒lɂ _timeStep ύX
   * 邱Ƃł܂B<p>
   *
   * <pre>
   *  public OhTime calculateTimeStep() {
   *    return (_k > 0) ? new OhTime(360, 0.) : new OhTime(60, 0.);
   *  }
   *
   *  public OhTIme calculateTimeStep() {
   *    double dt = 360 * pow(10, _k);
   *    return new OhTime(dt);
   *  }
   * </pre>
   *
   * _timeStep ́Awork() ̒Őݒ肳̂ŁA
   *
   * <pre>
   *  public OhTime calculateTimeStep() {
   *    _timeStep = new OhTime(1, 0.0);  // this is not good.
   *    return _timeStep;
   *  }
   * </pre>
   *
   * ̂悤ɁÃ\bh̒ŕύXȂłB<p>
   * 
   * ۃ\bhȂ̂ŁAhNX calculateTimeStep() ̒`
   * ^Kv܂B<p>
   *
   * ߂l: ̌vZ̃^CXebvB[sec]
   * </japanese>
   *
   * @return a <code>long</code> value
   * @exception OhError if an error occurs
   */
  public abstract OhTime calculateTimeStep() throws OhError;
  
  /**
   * Judgment whether calculation can be started or not.
   * In classes derived from class Element, its definition
   * should be given. <p>
   *
   * First, when "_currentTime" is more than "_terminalTime", it
   * is judged that calculation has finished, and "false" is
   * returned. <p>
   *
   * Next, it checks that all possession receiving ports can
   * acquire the data pack between "getNecessaryTimeFrom" and
   * "getNecessaryTimeTo". It is judged that calculation
   * execution is possible the case of being acquirable, "true"
   * is returned. It is judged that calculation execution is
   * impossible the case of being unacquirable, "false" is
   * returned. <p>
   *
   * A calculation request is advanced to the receiving port
   * which will have replied that data is unacquirable. In
   * particular, a calculation request message is sent to a
   * receiving port.
   * <pre>
   * [RETURN]
   *  Etrue --> When calculation execution is possible.
   *  Efalse -> When calculation execution is impossible.
   *              When calculation has finished.
   * </pre>
   *
   * <japanese>
   * <p>
   * vZJn̉s̔f <p>
   *
   * ܂A_currentTime  _terminalTime ȏ̏ꍇAvZς݂
   * f false Ԃ܂B<p>
   *
   * ɁASĂ̏LM[q getNecessaryTimeFrom(),
   * getNecessaryTimeTo() Ԃ̃f[^pbN擾\ł邱Ƃ
   * mF܂B擾\ȏꍇAvZs\Ɣf true 
   * Ԃ܂B擾s\ȏꍇAvZssƔf false 
   * Ԃ܂B<p>
   *
   * f[^擾s\łƕԎĂM[qɑ΂Čv
   * Zvo܂B̓Iɂ́AM[q calReq() bZ[
   * W𑗂܂B
   * <pre>
   * [߂l]
   *  Etrue --> vZs\̏ꍇB
   *  Efalse -> vZss\̏ꍇA܂͌vZς݂̏ꍇB
   * </pre>
   * </japanese>
   *
   * @return a <code>boolean</code> value
   * @exception OhCannotGetDataError if an error occurs
   * @exception OhError if an error occurs
   */
  public abstract boolean canYouCalculate()
    throws OhCannotGetDataError, OhError;
  
  /**
   * Calculation procedure for 1 step. In classes derived from
   * class Element, its definition should be given.
   *
   * <ul>
   *  <li>At first, in each amount of states, the value at
   *      "_currentTime" is stored.
   *  <li>Here, the amount of states and output value in
   *      "_currentTime + _timeStep" are calculated.
   *  <li>After this function finishes, the amount of states is
   *      updated by the value in a certain time.
   * </ul>
   *
   * In a general element,
   * Please describe following (1) to (4) in order to fulfill
   * these 3 points.
   *
   * <pre>
   *  (1) the input from a receiving port.
   *  (2) numerical computation.
   *  (3) the output to a sending port.
   *  (4) updating of the amount of states.
   * </pre>
   *
   * At (2), by using parameters, the amount of states in
   * <b>_currentTime</b> , and the input from a receiving ports,
   * please define concretely one step of a numerical model
   * which calculate the amount of states and the output value
   * which should be transmitted in "_currentTime +
   * _timeStep". <p>
   *
   * In this function, please do not change <b>_currentTime</b> and
   * <b>_timeStep</b>. <p>
   *
   * All numerical values required for calculation are acquirable
   * as instance variable and input from a possession receiving port.
   * Therefore, arguments are unnecessary. <p>
   *
   * The case where it is calculable as planned,
   * <b>true</b> is returned.
   * When a calculation time step is calculated inside
   * <b>calculate()</b>, the data for the calculation time step
   * cannot be acquired and calculation is not able to be
   * advanced, <b>false</b> is returned.
   *
   * <japanese>
   * <p>
   * PXebv̌vZ葱 <p>
   *
   * PXebv̌vZ葱`ĉB
   *
   * ʓIȗvfł́A
   *
   * <ul>
   *  <li> ŏAeԗʂɂ _currentTime ɂli[
   *       ĂB
   *  <li> ł _currentTime + _timeStep ɂԗ, o
   *       ͒lvZB
   *  <li> ̃\bhIAԗʂ _currentTime + _timeStep
   *       ɂlɍXVĂB
   * </ul>
   *
   * ̂R_𖞑悤ɁAL (1)  (4) ɋLq
   * B
   *
   * <pre>
   *  (1) M[q̓́B
   *  (2) vZB
   *  (3) M[qւ̏óB
   *  (4) ԗʂ̍XVB
   * </pre>
   *
   * (2) ł́Ap[^A_currentTime ɂԗʁAM[q
   * ̓͂pA_currentTime + _timeStep ɂԗʁA
   * Mׂo͒lvZ鐔f̂PXebv̓I
   * ɒ`ĉB<p>
   *
   * ̃\bh̒ _currentTime, _timeStep ύXȂŉ
   * BKvȂ _timeStep Xɍׂԋ敪ɕČv
   * Zł܂A͍̎ƗpϐpA_currentTime,
   * _timeStep ύXȂ悤ɒӂĉB<p>
   *
   * vZɕKvȑSĂ̐ĺAf[^oyяLM[q
   * ̓͂ƂĎ擾ł܂B]Ĉ͕svłB<p>
   *
   * \ʂvZłꍇ́Atrue Ԃ܂B
   * calculate() ̓ŌvZԃXebvvZA
   * ̌vZԃXebṽf[^擾łȂȂǂāA
   * vZi߂邱ƂłȂꍇ́Afalse Ԃ܂B<p>
   *
   * ۃ\bhȂ̂ŁAhNXŕK`ĉB
   *
   * </japanese>
   *
   * @return a <code>boolean</code> value
   * @exception OhCannotGetDataError if an error occurs
   * @exception OhError if an error occurs
   */
  public abstract boolean calculate()
    throws OhCannotGetDataError, OhError;
  
  /**
   * Calculate between <b>_currentTime</b> and <b>_terminalTime</b>.
   * In classes derived from the OhElement class, its definition
   * must be given. <p>
   *
   * Calculation by <b>_terminalTime</b> is performed using
   * <b>calculate()</b>, <b>calculateTimeStep()</b>, etc. <p>
   *
   * First, if return value of <b>canYouCalculate()</b> becomes
   * <b>false</b>, it will be judged that calculation cannot be 
   * started and <b>false</b> will be returned. <p>
   *
   * When the return value is <b>true</b>,
   * <b>calculate()</b> and <b>_currentTime += _timeStep</b>
   * are repeated
   * and performed, as long as the return value of
   * <b>canYouCalculate</b> is "true".  <p>
   *
   * Finally, <b>setDataPackArray()</b> and
   * <b>reportToSender()</b> are performed and <b>true</b> is returned.
   *
   * <japanese>
   * <p>
   * _currentime, _terminalTime Ԃ̌vZ <p>
   *
   * calculate(), calculateTimeStep() ȂǂpāA
   * _terminalTime
   * ܂ł̌vZs܂B<p>
   *
   * ܂A
   * canYouCalculate() ̖߂l false Ȃ΁AvZJnsƔ
   * fAfalse Ԃ܂B<p>
   *
   * ߂l true ̏ꍇ́A
   * true łA
   * calculate()
   * _currentTime.add(_timeStep) JԂs܂B<p>
   *
   * ŌɁA
   * setDataPackArray()
   * sAtrue Ԃ܂B
   *
   * <pre> 
   * [߂l]
   *  Etrue --> vZsꍇB
   *  Efalse -> vZs\̏ꍇA܂͌vZς݂̏ꍇB
   * </pre>
   * </japanese>
   *
   * @return a <code>boolean</code> value
   * @exception OhCannotGetDataError if an error occurs
   * @exception OhError if an error occurs
   */
  public abstract boolean work() throws OhCannotGetDataError, OhError;

  // 6. information output function

  // 7. the function to save the last state at a file
  
  /**
   * This is the method which does the work which writes the
   * last state of an element in the file specified by the
   * output stream "aBr". <p>
   *
   * In classes derived from class Element, its definition
   * should be given.
   *
   * <japanese>
   * <p>
   * vf̍ŏIԂo̓Xg[ os Ŏw肳t@Cɏ
   * ލƂ郁\bhłB[U`vff
   * ́Ã\bhK`Ȃ΂Ȃ܂B<p>
   *
   * ŏIԂގdl͎̂悤ɂĂB
   * <pre>
   * (data_string)
   * </pre>
   * </japanese>
   *
   * @param aPr a <code>PrintWriter</code> value
   */
  public abstract void saveTerminalState(PrintWriter aPr);
  
  // @end {abstract methods} -----------------------------------


  // @begin {initialization} -----------------------------------
  
  /**
   * Initialization with the information which is given through
   * a file.
   *
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @param aBr an <code>OhBufferedReader</code> value
   * @return a <code>boolean</code> value
   * @exception OhError if an error occurs
   */
  public void init(
    String aObjName, int aObjNum, OhBufferedReader aBr)
    throws OhError {
    if (true)
      throw new OhError("This init() method is not defined.");
  }
  
  public void init(
    String aObjName, int aObjNum, int aNumOfObjs,
    String aDataStr) throws OhError {
    super.init(aObjName, aObjNum);
    setObjectArray(aNumOfObjs);
    setDataString(aDataStr);
  }
  
  public void init(
    String aObjName, int aObjNum, int aNumOfObjs)
    throws OhError {
    super.init(aObjName, aObjNum);
    setObjectArray(aNumOfObjs);
    setDataString(null);
  }
  
  public void init(
    String aObjName, int aObjNum, String aDataStr)
    throws OhError {
    
    super.init(aObjName, aObjNum);
    setObjectArray(0);
    setDataString(aDataStr);
  }
  
  public void init(String aObjName, int aObjNum)
    throws OhError {
    super.init(aObjName, aObjNum);
    setObjectArray(0);
    setDataString(null);
  }
  
  // @end {initialization} -------------------------------------
  
  
  // @begin {constructors} -------------------------------------
  
  /**
   * Creates a new default <code>OhElement</code> instance.
   */
  public OhElement() {
    super();
    _initialWorkDoneP     = false;
    _shareInfoDoneP       = false;
    _objectList           = null;
    _receivePortArray     = null;
    _numberOfReceivePorts = -1;
    _sendPortArray        = null;
    _numberOfSendPorts    = -1;
    _objectArray          = null;
    _numberOfObjects      = -1;
    _dataString           = null;
    _currentTime          = new OhTimePoint();
    _timeStep             = null;
    _fixedTime            = new OhTimePoint();
    _targetTime           = new OhTimePoint();
    _calReqP              = false;

    _className            = "OhElement";
  }

  
  public OhElement(String aObjName, int aObjNum, int aNumOfObjs,
    String aStr) {
    super();
    _initialWorkDoneP     = false;
    _shareInfoDoneP       = false;
    _objectList           = null;
    _receivePortArray     = null;
    _numberOfReceivePorts = -1;
    _sendPortArray        = null;
    _numberOfSendPorts    = -1;
    _objectArray          = null;
    _numberOfObjects      = -1;
    _dataString           = null;
    _currentTime          = new OhTimePoint();
    _timeStep             = null;
    _fixedTime            = new OhTimePoint();
    _targetTime           = new OhTimePoint();
    _calReqP              = false;

    _className            = "OhElement";

    try {
      init(aObjName, aObjNum, aNumOfObjs, aStr);
    }
    catch (OhError e) {
    }
  }
  
  public OhElement(String aObjName, int aObjNum, String aStr) {
    super();
    _initialWorkDoneP     = false;
    _shareInfoDoneP       = false;
    _objectList           = null;
    _receivePortArray     = null;
    _numberOfReceivePorts = -1;
    _sendPortArray        = null;
    _numberOfSendPorts    = -1;
    _objectArray          = null;
    _numberOfObjects      = -1;
    _dataString           = null;
    _currentTime          = new OhTimePoint();
    _timeStep             = null;
    _fixedTime            = new OhTimePoint();
    _targetTime           = new OhTimePoint();
    _calReqP              = false;

    _className            = "OhElement";

    try {
      init(aObjName, aObjNum, aStr); 
    }
    catch (OhError e) {
    }
  }
  
  // @end {constructors} ---------------------------------------

  
  // @begin {methods} ------------------------------------------

  // copy method

  // Do not define copy method for a super class

  // 1. creation and reference of "_receivePortArray",
  //    "_sendPortArray" and "_objectArray"
  
  /**
   * Finds an object with class name <b>aClassName</b>, object
   * name <b>aObjName</b>, and object number <b>aObjNum</b> from
   * <b>_objectList</b> and returns the object.
   *
   * When such an object could not be found from <b>_objectList</b>,
   * <b>null</b> will be returned.
   *
   * <japanese>
   * <p>
   * LinkedList _objectList  OhObject ^IuWFNg̎擾 <p>
   *
   * ̃\bh́Aw肵NX, IuWFNg, IuWF
   * NgԍOhObject ^IuWFNg _objectList 
   * ܂B<p>
   *
   * LinkedList ^̃IuWFNǵACӂ̎Qƌ^̕ϐ̃Xg
   * [ł܂A_objectList ɂ́AOhObject ^Ă
   * B<p>
   *
   * w肳ꂽIuWFNg_objectList ɓo^ĂȂꍇ
   *  null Ԃ܂B<p>
   *
   * _objectList  null ̏ꍇ́AOhError  throw ܂B
   *
   * </japanese>
   *
   * @param aClassName a <code>String</code> value
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @return an <code>OhObject</code> value
   * @exception OhError if an error occurs
   */
  public OhObject getFromList(
    String aClassName,
    String aObjName,
    int aObjNum)
    throws OhError {    //; (*1)
    if (_objectList == null) {
      // The list has already been made out.
      throw new OhError(
        "\tAsked > [" + aClassName + "] : '" +
        aObjName + "' : " + aObjNum + " :" + NEWLINE +
        oPrintString("\t")
      );
    }

    for (int i = 0; i < _objectList.size(); i++)
      if (((OhObject)_objectList.get(i)).
        objCompare(aClassName, aObjName, aObjNum) == true)
        return (OhObject) _objectList.get(i);
          // The specified object was found.
    
    return null; // It was found that the specified object
                 // had not been registered.
  }
  
  /**
   * Finds an object with object name <b>aObjName</b>, and object
   * number <b>aObjNum</b> from <b>_objectList</b> and returns the
   * object. When such an object could not be found from
   * <b>_objectList</b>, <b>null</b> will be returned.
   *
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @return an <code>OhObject</code> value
   * @exception OhError if an error occurs
   */
  public OhObject getFromList(String aObjName, int aObjNum)
    throws OhError {

    if (_objectList == null) { // The list was already arranged.
      String str = "\tAsked > " + aObjName + " : " +
        aObjNum + " :" + NEWLINE +
        oPrintString("\t");
      throw new OhError(str);
    }
    
    for (int i = 0; i < _objectList.size(); i++)
      if (((OhObject)_objectList.get(i)).
        objCompare(aObjName, aObjNum) == true)
        return (OhObject) _objectList.get(i);
          // The specified object was discovered.

    return null; // The specified object was not registered.
  }
  
  /**
   * Registers an object <b>aObj</b> to the data member
   * <b>_objectList</b>.
   *
   * <japanese>
   * <br>
   * OhObject ^IuWFNg̓o^ _owner ̐ݒ <p>
   *
   *  OhObject ^IuWFNg <b>aObj</b>  _objectList 
   * ɓo^܂BɁÃ\bhĂяoIuWFNg
   * ithisj <b>aObj</b>  _owner ƂȂ悤ɐݒ肳܂B
   * <p>
   *
   * o^ɂA̏ꍇ́AOhError IuWFNgX[
   * ܂B
   *
   * <ul>
   *  <li> aObj  null ̏ꍇB
   *  <li> object_list  NULL ̏ꍇB
   *  <li> NXAIuWFNgAIuWFNgԍS
   *       aObj ƈvIuWFNg
   *       _objectList ɓo^ĂꍇB
   *  <li>aObj  _owner ݒς݂̏ꍇB
   * </ul>
   *
   * </japanese>
   *
   * @param aObj an <code>Object</code> value
   * @exception OhError if an error occurs
   */
  public void register(OhObject aObj) throws OhError {
    if (aObj == null) { // The argument is inaccurate.
      throw new IllegalArgumentException(oPrintString("\t"));
    }

    if (_objectList == null) { // The list was already arranged.
      throw new OhError(oPrintString("\t"));
    }

    // Comparison of class name, object name, and object
    // number. When variables other than <b>Object</b> type are
    // stored in <b>aObj</b>, abnormalities occur.
    //
    String className = aObj.getClassName();
    String objName   = aObj.getObjectName();
    int    objNum    = aObj.getObjectNumber();

    if (objName == null || objNum == -1) {
      String str = aObj.oPrintString("    registered > ") +
        oPrintString("\t");
      throw new OhError(str);
    }

    if (getFromList(className, objName, objNum) != null) {
      String str = aObj.oPrintString("    registered > ") +
        oPrintString("\t");
      throw new OhError(str);
    }

    aObj.setOwner(this);  
    _objectList.add(aObj);

  }
  
  /**
   * Registers the OhObject <b>aObj</b> with which the current
   * object communicates directly to object_array[num].
   *
   * @param aObj an <code>OhObject</code> value
   * @param aNum an <code>int</code> value
   * @exception OhError if an error occurs
   */
  public void register(OhObject aObj, int aNum)
    throws OhError {    //; (*6)
    if ((aNum < 0) || (aNum >= _numberOfObjects)) {
      String str = "\tnum (arg) = " + aNum +
        ", _numberOfObjects = " + _numberOfObjects + NEWLINE +
        oPrintString("\t");
      throw new IllegalArgumentException(str);
    }

    if (aObj == null) {
      throw new IllegalArgumentException(oPrintString("\t"));
    }

    if (_objectArray[aNum] != null) {
      String str = "\tDoubly set _objectArray[" + aNum +
        "]." + NEWLINE +
        _objectArray[aNum].oPrintString("    first > ") +
        aObj.oPrintString("    next > ");
      throw new OhError(str);
    }

    _objectArray[aNum] = aObj;
  }
  
  /**
   * Creates an array of objects which are contained in
   * <b>_objectList</b> and stores it in <b>_aObjArray</b> and the number
   * of elements in <b>aNumOfObjs</b>.
   *
   * <japanese>
   * <p>
   * z񉻂 _objectList Ԃ܂BɁA aNumOfObjs 
   * izĵPԖڂɗvf _numberOfObjects i[ĕԂ
   * B<p>
   *
   * _objectList  null wĂꍇAG[bZ[W\
   * ܂B<p>
   *
   * ܂A aNumOfObjs[0]  0 ȏ̏ꍇ́Aɔz쐬
   * ς݂ƔfAG[bZ[W\܂Bz񉻌A_
   * numOfObjs[0]  0 ȏɐݒ肳܂B]āAzݒO
   *  _numOfObjs[0]  0 ɏĂĉB<p>
   *
   * _numberOfReceivePort, _numberOfSendPort ́ARXgN^
   *  -1 ɏ܂B<p>
   *
   * <pre>
   * [in] aNumOfObjs[0] - _objectList ̗vfi_numberOfObjectsj
   *
   * [RETURN] _objectList z񉻂 Object ^z
   * </pre>
   * </japanese>
   *
   * @param aNumOfObjs an <code>int[]</code> value
   * @return an <code>Object[]</code> value
   * @exception OhError if an error occurs
   */
  public Object[] makeObjectArray(int[] aNumOfObjs)
    throws OhError {
    if (_objectList == null) {
      throw new OhError(oPrintString("\t"));
    }

    if (aNumOfObjs[0] >= 0) {
      throw new OhError(oPrintString("\t"));
    }

    Object[] aObjArray;
    aObjArray =
      _objectList.toArray(new Object[_objectList.size()]);
    
    aNumOfObjs[0] = _objectList.size();

    return aObjArray;
  }
  
  public OhPort[] makePortArray(int[] aNumOfObjs)
    throws OhError {
    if (_objectList == null) {
      throw new OhError(oPrintString("\t"));
    }

    if (aNumOfObjs[0] >= 0) {
      throw new OhError(oPrintString("\t"));
    }

    Object[] aObjArray;
    aObjArray =
      _objectList.toArray(new Object[_objectList.size()]);
    
    aNumOfObjs[0] = _objectList.size();
    OhPort[] portArray = new OhPort[_objectList.size()];
    for (int i = 0; i < portArray.length; i++) {
      portArray[i] = (OhPort)aObjArray[i];
    }

    return portArray;
  }
  
  /**
   * Acquire the memory to store <b>aNumOfObjs</b> objects and
   * makes
   * <b>_objectArray</b> point to the memory.
   *
   * <japanese>
   * <p>
   * _objectArray ̍쐬 <p>
   *
   * aNumOfObjs (arg) ̃IuWFNgL邽߂̗̈q[
   * vɊmۂA_objectArray ̐擪AhXw悤ɃZb
   * g܂B<p>
   *
   * _objectArray ͒ڒʐM鑊L邽߂Ɏg܂B
   * <p>
   *
   * ܂A
   *
   * <pre>
   *  EaNumOfObjs (arg) ̏ꍇB
   *  E_numberOfObjects  0 ȏ̏ꍇBi쐬ς݁j
   * </pre>
   *
   * ̏ꍇ OhError IuWFNg𑗏o܂B
   *
   * <pre>
   * [ARGUMENT]
   *  EaNumOfObjs -> _objectArray ɋLIuWFNgB
   * [RETURN]
   *  Etrue --> _objectArray mۂłꍇB
   *  Efalse -> aNumOfObjs ̏ꍇA܂ _numberOfObjects  0 ȏ̏ꍇBi쐬ς݁j
   * </pre>
   * </japanese>
   *
   * @param aNumOfObjs an <code>int</code> value
   * @exception OhError if an error occurs
   */
  public void setObjectArray(int aNumOfObjs) throws OhError {
    if (aNumOfObjs < 0) {
      throw new IllegalArgumentException(oPrintString("    "));
    }

    if ((_numberOfObjects >= 0) || (_objectArray != null)) {
      throw new OhError(oPrintString("    "));
    }

    try {
      _objectArray = new OhObject[aNumOfObjs];
    }
    catch (ArrayIndexOutOfBoundsException e) {
      throw new ArrayIndexOutOfBoundsException(oPrintString("\t"));
    }

    for (int i = 0; i < aNumOfObjs; i++)
      _objectArray[i] = null;

    _numberOfObjects = aNumOfObjs;
  }
  
  /**
   * Makes <b>_receivePortArray</b> and <b>_sendPortArray</b>.
   *
   * <japanese>
   * <p>
   * Object ^IuWFNgLpz̍쐬 <p>
   *
   * Object ^IuWFNgLpz쐬܂B܂A_
   * receivePortArray, _sendPortArray, _objectArray 쐬܂B
   * <p>
   *
   * ܂A_objectList  null ł邱ƂmF܂Bnull ł
   * ꍇA_objectList gpƔfăG[bZ[W\
   * Areturn ܂B<p>
   *
   * ɁAȉ̍Ƃs܂B
   *
   * <ol>
   *  <li> OhLinkedList IuWFNgq[vɍ쐬
   *       _objectList ɐݒ肷B
   *  <li> OhElement.registerReceivePort(),
   *       OhElenment.makeObjectArray() p
   *       _receivePortArray 쐬B
   *  <li> OhLinkedList.clear() ɂ _objectList B
   *  <li> OhElement.registerSendPort(),
   *       OhElement.makeObjectArray() p
   *       _sendPortArray 쐬B
   * </ol>
   *
   * Ō _objectList  null ɐݒ肵܂B<p>
   *
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public void construct() throws OhError {
    if (_objectList != null) {
      throw new OhError(oPrintString("    "));
    }

    _objectList = new OhLinkedList();

    // --- Makes _receivePortArray. ---

    registerReceivePorts(); // Abstract method call.
                            // It might throw an OhError.

    int[] numberOfReceivePortsArray = {_numberOfReceivePorts};
    
    _receivePortArray     =
      makePortArray(numberOfReceivePortsArray);
    _numberOfReceivePorts = numberOfReceivePortsArray[0];

    _objectList.clear();

    // --- Makes _sendPortArray. ---

    registerSendPorts();    // Abstract method call.

    int[] numberOfSendPortsArray = {_numberOfSendPorts};
    
    _sendPortArray     = makeObjectArray(numberOfSendPortsArray);    
    _numberOfSendPorts = numberOfSendPortsArray[0];

    _objectList = null;
  }
  
  /**
   * Finds an object with class name <b>aClassName</b>, object name
   * <b>aObjName</b>, and object number <b>aObjNum</b>
   * from <b>aObjArray</b>
   * and returns the object.
   *
   * <japanese>
   * <p>
   * Object ^IuWFNg̎擾 <p>
   *
   * getReceivePort(), getSendPort(), getObject() ŗp
   * \bh̖SB
   * </japanese>
   *
   * @param aClassName a <code>String</code> value
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @param aObjArray an <code>Object[]</code> value
   * @param aNumOfObjs an <code>int</code> value
   * @return an <code>Object</code> value
   */
  public OhObject getFromArray(String aClassName, String aObjName,
    int aObjNum, Object[] aObjArray, int aNumOfObjs) { 
    if (aNumOfObjs < 0) { // The list is not arranged.
      throw new IllegalArgumentException(oPrintString("\t"));
    }
    
    for (int k = 0; k < aNumOfObjs; k++)
      if (((OhObject)aObjArray[k]).
        objCompare(aClassName, aObjName, aObjNum) == true)
        return (OhObject)aObjArray[k];
          // The specified object was discovered.
    
    return null;
  }
  

  
  /**
   * Finds an object with object name <b>aObjName</b>, and
   * object number <b>aObjNum</b> from <b>aObjArray</b> and
   * returns the object.
   *
   * <japanese>
   * <p>
   * Object ^IuWFNg̎擾 <p>
   *
   * getReceivePort(), getSendPort(), getObject() ŗp
   * \bh̖SB
   * </japanese>
   *
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @param aObjArray an <code>Object[]</code> value
   * @param aNumOfObjs an <code>int</code> value
   * @return an <code>Object</code> value
   */
  public Object getFromArray(String aObjName, int aObjNum,
    Object[] aObjArray, int aNumOfObjs) {
    if (aNumOfObjs < 0) { // The list is not arranged.
      throw new IllegalArgumentException(oPrintString("\t"));
    }
    
    for (int k = 0; k < aNumOfObjs; k++)
      if (((OhObject)aObjArray[k]).
        objCompare(aObjName, aObjNum) == true)
        return aObjArray[k];
          // The specified object was found.

    return null;
  }
  
  /**
   * By specifying class name, object name, and object number,
   * acquires a receiving port from
   * <b>_receivePortArray</b>. <p>
   *
   * When the specified object is not registered,
   * <b>null</b> is returned.
   *
   * <japanese>
   * <p>
   * _receivePortArray M[q擾 <p>
   *
   * NXAIuWFNgAIuWFNgԍw肷邱Ƃ
   * A_receivePortArray M[q擾܂Bw肳
   * IuWFNgo^ĂȂꍇ null Ԃ܂B
   * </japanese>
   *
   * @param aClassName a <code>String</code> value
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @return an <code>Object</code> value
   * @exception OhError if an error occurs
   */
  public Object getReceivePort(String aClassName,
    String aObjName, int aObjNum) throws OhError {
    Object p =
      getFromArray(aClassName, aObjName,
        aObjNum, _receivePortArray, _numberOfReceivePorts);

    if (p == null) { // The specified object was not registered.
      String str = "\tAsked > \'" + aObjName + "\' : " +
        aObjNum + " :" + NEWLINE +
        oPrintString("\t");
      throw new OhError(str);
    }
    return p;
  }

  /**
   * By specifying class name and object name, acquires a
   * receiving port from <b>_receivePortArray</b>. At this time,
   * the object number is set to 0. <p>
   *
   * When the specified object is not registered,
   * <b>null</b> is returned.
   *
   * <japanese>
   * <p>
   * _receivePortArray M[q擾 <p>
   *
   * NXAIuWFNgw肷邱ƂɂA_
   * receivePortArray M[q擾܂Bw肳ꂽIuWF
   * Ngo^ĂȂꍇ null Ԃ܂B
   * </japanese>
   *
   * @param aClassName a <code>String</code> value
   * @param aObjName a <code>String</code> value
   * @return an <code>Object</code> value
   * @exception OhError if an error occurs
   */
  public Object getReceivePort(String aClassName, String aObjName)
    throws OhError {

    Object p = getFromArray(aClassName, aObjName, 0,
      _receivePortArray, _numberOfReceivePorts);

    if (p == null) { // The specified object was not registered.
      throw new OhError(
        "\tAsked > \'" + aObjName + "\' : 0 :" + NEWLINE +
        oPrintString("\t")
      );
    }
    return p;
  }
  
  /**
   * By specifying class name, object name, and object number,
   * acquires a sending port from <b>_sendPortArray</b>. <p>
   *
   * When the specified object is not registered,
   * <b>null</b> is returned.
   *
   * <japanese>
   * <p>
   * _sendPortArray 瑗M[q擾 <p>
   *
   * NXAIuWFNgAIuWFNgԍw肷邱Ƃ
   * A_sendPortArray 瑗M[q擾܂Bw肳ꂽI
   * uWFNgo^ĂȂꍇ null Ԃ܂B
   * </japanese>
   *
   * @param aClassName a <code>String</code> value
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @return an <code>Object</code> value
   * @exception OhError if an error occurs
   */
  public Object getSendPort(String aClassName, String aObjName,
    int aObjNum) throws OhError {
    Object p = getFromArray(aClassName, aObjName, aObjNum,
      _sendPortArray, _numberOfSendPorts);

    if (p == null) { // The specified object was not registered.
      throw new OhError(
        "\tAsked > \'" + aObjName + "\' : " + aObjNum +
        " :" + NEWLINE +
        oPrintString("\t")
      );
    }
    return p;
  }

  
  /**
   * By specifying class name and object name, gets a
   * sending port from <b>_sendPortArray</b>. The
   * object number of the port is assumed to be 0. <p>
   *
   * When the specified object has not been registered,
   * <b>null</b> is returned.
   *
   * <japanese>
   * <p>
   * _sendPortArray 瑗M[q擾 <p>
   *
   * NXAIuWFNgw肷邱ƂɂA_
   * sendPortArray 瑗M[q擾܂Bw肳ꂽIuWF
   * Ngo^ĂȂꍇ null Ԃ܂B
   * </japanese>
   *
   * @param aClassName a <code>String</code> value
   * @param aObjName a <code>String</code> value
   * @return an <code>Object</code> value
   * @exception OhError if an error occurs
   */
  public Object getSendPort(String aClassName, String aObjName)
    throws OhError {

    Object p = getFromArray(aClassName, aObjName, 0,
      _sendPortArray, _numberOfSendPorts);

    if (p == null) {  //; The specified object was not registered.
      throw new OhError(
        "\tAsked > \'" + aObjName + "\' : 0 :" + NEWLINE +
        oPrintString("\t")
      );
    }
    return p;
  }
  
  /**
   * By specifying object name and object number,
   * get the receive port from <b>_receivePortArray</b>. <p>
   *
   * When the specified object was not found,
   * <b>null</b> is returned.
   *
   * <japanese>
   * <p>
   * _receivePortArray M[q擾 <p>
   *
   * IuWFNgAIuWFNgԍw肷邱ƂɂA
   * _receivePortArray M[q擾܂Bw肳ꂽIu
   * WFNgo^ĂȂꍇ null Ԃ܂B
   * </japanese>
   *
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @return an <code>Object</code> value
   * @exception OhError if an error occurs
   */
  public Object getReceivePort(String aObjName, int aObjNum)
    throws OhError {

    Object p = getFromArray(aObjName, aObjNum,
      _receivePortArray, _numberOfReceivePorts);

    if (p == null) { // The specified object was not registered.
      throw new OhError(
        "\tAsked > \'" + aObjName + "\' : " + aObjNum + " :"
        + NEWLINE + oPrintString("\t")
      );
    }
    return p;
  }

  /**
   * Gets a receive port from <b>_receivePortArray</b> by
   * specifying the object name of the receive port. The object
   * number of the port is assumed to be <p>
   *
   * When the specifyed object is not registered,
   * <b>null</b> is returned.
   *
   * <japanese>
   * <p>
   * _receivePortArray M[q擾 <p>
   *
   * IuWFNgw肷邱ƂɂA
   * _receivePortArray M[q擾܂Bw肳ꂽIu
   * WFNgo^ĂȂꍇ null Ԃ܂B
   * </japanese>
   *
   * @param aObjName a <code>String</code> value
   * @return an <code>Object</code> value
   * @exception OhError if an error occurs
   */
  public Object getReceivePort(String aObjName)
    throws OhError {

    Object p = getFromArray(aObjName, 0,
      _receivePortArray, _numberOfReceivePorts);

    if (p == null) { // The specified object was not registered.
      throw new OhError(
        "\tAsked > \'" + aObjName + "\' : 0 :" + NEWLINE +
        oPrintString("\t")
      );
    }
    return p;
  }

  
  /**
   * By specifying object name and object number,
   * acquires a sending port from <b>_sendPortArray</b>. <p>
   *
   * When the specified object is not registered,
   * <b>null</b> is returned.
   *
   * <japanese>
   * <p>
   * _sendPortArray 瑗M[q擾 <p>
   *
   * IuWFNgAIuWFNgԍw肷邱ƂɂA_
   * sendPortArray 瑗M[q擾܂Bw肳ꂽIuWF
   * Ngo^ĂȂꍇ null Ԃ܂B
   * </japanese>
   *
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @return an <code>Object</code> value
   * @exception OhError if an error occurs
   */
  public Object getSendPort(String aObjName, int aObjNum)
    throws OhError {
    Object p = getFromArray(aObjName, aObjNum,
      _sendPortArray, _numberOfSendPorts);

    if (p == null) { // The specified object was not registered.
      throw new OhError(
        "\tAsked > \'" + aObjName + "\' : " + aObjNum + " :" +
        NEWLINE + oPrintString("\t")
      );
    }
    return p;
  }

  /**
   * By specifying object name, gets a sending port from
   * <b>_sendPortArray</b>. The object number is assumed to be
   * 0. <p>
   *
   * When the specified object was not found, <b>null</b> is
   * returned.
   *
   * <japanese>
   * <p>
   * _sendPortArray 瑗M[q擾 <p>
   *
   * IuWFNgw肷邱ƂɂA
   * _sendPortArray 瑗M[q擾܂Bw肳ꂽIuWF
   * Ngo^ĂȂꍇ null Ԃ܂B
   * </japanese>
   *
   * @param aObjName a <code>String</code> value
   * @return an <code>Object</code> value
   * @exception OhError if an error occurs
   */
  public Object getSendPort(String aObjName)
    throws OhError {
    Object p = getFromArray(aObjName, 0, _sendPortArray,
      _numberOfSendPorts);

    if (p == null) { // The specified object was not registered.
      throw new OhError(
        "\tAsked > \'" + aObjName + "\' : 0 :" + NEWLINE +
        oPrintString("\t")
      );
    }
    return p;
  }
  
  /**
   * Gets an object from the _objectArray which has the classname,
   * objectName, and objectNumber specified by the arguments. <p>
   *
   * When the specified object is not found, <b>null</b> is
   * returned.
   *
   * <japanese>
   * <p>
   * IuWFNg̎擾 <p>
   *
   * w肳ꂽIuWFNgo^ĂȂꍇ null Ԃ܂B
   * </japanese>
   *
   * @param aClassName a <code>String</code> value
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @return an <code>Object</code> value
   * @exception OhError if an error occurs
   */
  public Object getObject(String aClassName, String aObjName,
    int aObjNum) throws OhError {

    Object obj = getFromArray(aClassName, aObjName, aObjNum,
      _objectArray, _numberOfObjects);

    if (obj == null) { // The specified object was not registered.
      throw new OhError(
        "\tAsked > \'" + aObjName + "\' : " + aObjNum + " :" +
        NEWLINE + oPrintString("\t")
      );
    }
    return obj;
  }

  
  /**
   * Gets an object from the _objectArray which has the classname
   * and objectName specified by the arguments. The objectNumber
   * is assumed to be 0.<p>
   *
   * When the specified object is not found, <b>null</b> is
   * returned.
   *
   * <japanese>
   * <p>
   * IuWFNg̎擾 <p>
   *
   * w肳ꂽIuWFNgo^ĂȂꍇ null Ԃ܂B
   * </japanese>
   *
   * @param aClassName a <code>String</code> value
   * @param aObjName a <code>String</code> value
   * @return an <code>Object</code> value
   * @exception OhError if an error occurs
   */
  public Object getObject(String aClassName, String aObjName)
    throws OhError {
    Object obj = getFromArray(aClassName, aObjName, 0,
      _objectArray, _numberOfObjects);

    if (obj == null) { // The specified object was not registered.
      throw new OhError(
        "\tAsked > \'" + aObjName + "\' : 0 :" + NEWLINE +
        oPrintString("\t")
      );
    }
    
    return obj;
  }
  
  /**
   * Gets a component object.
   *
   * <japanese>
   * <br>
   * \IuWFNgi\vfj̎擾 <p>
   *
   * NXAIuWFNgAIuWFNgԍw肵āA_
   * objectArray \vf擾܂BOhElement NX͍
   * PʂłvfȂ̂ŁA\vf͎Ă܂Bł
   * KG[sȂ悤ɂĂ܂BhNXł
   * OhSubSystem, OhTotalSystem ŃI[o[Ch܂B 
   * OhElement NXpĐVɃNX郆[U
   * ȂĂ悤ɁAۃ\bhɂĂ܂B
   * </japanese>
   *
   * @param aClassName a <code>String</code> value
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @return an <code>Object</code> value
   * @exception OhError if an error occurs
   * @exception OhError if an error occurs
   */
  public OhObject getComponent(String aClassName, String aObjName,
    int aObjNum) throws OhError {

    if (true) {
      throw new OhError(
        "\taClassName (arg) = " + aClassName +
        "aObjName (arg) = " + aObjName +
        "aObjNum (arg) = " + aObjNum + NEWLINE +
        oPrintString("\t")
      );
    }
    
    return null;
  }

  
  /**
   * Gets a component OhElement which has the specified
   * className and objectName from within
   * <b>_objectArray</b>. The objectNumber of the OhElement is
   * assumed to be 0.<p>
   *
   * <japanese>
   * <br>
   * \IuWFNgi\vfj̎擾 <p>
   *
   * NXAIuWFNgw肵āA_objectArray \v
   * f擾܂BIuWFNgԍ 0 ɂȂ܂BOhElement 
   * NX͍ŏPʂłvfȂ̂ŁA\vf͎Ă܂B
   * ł͕KG[sȂ悤ɂĂ܂BhNX
   * łOhSubSystem, OhTotalSystem ŃI[o[Ch܂B
   *  OhElement NXpĐVɃNX郆[U
   * ȂĂ悤ɁAۃ\bhɂĂ܂B
   * </japanese>
   *
   * @param aClassName a <code>String</code> value
   * @param aObjName a <code>String</code> value
   * @return an <code>Object</code> value
   * @exception OhError if an error occurs
   */
  public Object getComponent(String aClassName, String aObjName)
    throws OhError {

    if (true) {
      throw new OhError(
        "\taClassName (arg) = " + aClassName +
        "aObjName (arg) = " + aObjName +
        "aObjNum (arg) = 0" + NEWLINE +
        oPrintString("\t")
      );
    }
    
    return null;
  }

  
  /**
   * Gets a component OhElement which has the specified
   * objectName and ObjectNumber from
   * <b>_objectArray</b>. 
   *
   * <japanese>
   * <br>
   * \IuWFNgi\vfj̎擾 <p>
   *
   * IuWFNgAIuWFNgԍw肵āA_objectArray 
   * \vf擾܂BOhElement NX͍ŏPʂłv
   * fȂ̂ŁA\vf͎Ă܂Bł͕KG[
   * sȂ悤ɂĂ܂BhNXłOhSubSystem,
   * OhTotalSystem ŃI[o[Ch܂B OhElement N
   * XpĐVɃNX郆[UȂĂ
   * ɁAۃ\bhɂĂ܂B
   * </japanese>
   *
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @return an <code>Object</code> value
   * @exception OhError if an error occurs
   */
  public Object getComponent(String aObjName, int aObjNum)
    throws OhError{
    if (true) {
      throw new Error(
        "aObjName (arg) = " + aObjName + "aObjNum (arg) = " +
        aObjNum + NEWLINE + oPrintString("\t")
      );
    }
    return null;
  }
  
  /**
   * Gets a component OhElement which has the specified
   * objectName from <b>_objectArray</b>. 
   *
   * <japanese>
   * <br>
   * \IuWFNgi\vfj̎擾 <p>
   *
   * IuWFNgw肵āA_objectArray \vf擾܂B
   * OhElement NX͍ŏPʂłvfȂ̂ŁA\vf͎Ă܂B
   * ł͕KG[sȂ悤ɂĂ܂B
   * hNXł
   * OhSubSystem,
   * OhTotalSystem
   * ŃI[o[Ch܂B
   *  OhElement NXpĐVɃNX郆[U
   * ȂĂ悤ɁAۃ\bhɂĂ܂B
   * </japanese>
   *
   * @param aObjName a <code>String</code> value
   * @return an <code>Object</code> value
   * @exception OhError if an error occurs
   */
  public Object getComponent(String aObjName) throws OhError{
    if (true) {
      throw new OhError(
        "aObjName (arg) = " + aObjName + "aObjNum (arg) = 0" +
        NEWLINE + oPrintString("\t")
      );
    }

    return null;
  }

  // 2. connection function
  
  /**
   * Orders the send ports of this element to execute doSwitch.
   *
   * <japanese>
   * <p>
   * SĂ̏LM[q doSwitch() s܂B<p>
   *
   * ̓Iɂ́Aȉs܂B
   *
   * <ol>
   *  <li> MM[qz _dataPackArray 쐬B
   *  <li> MM[q̐ _numberOfReceivers ɐݒ肷B
   *  <li> _dataPackArray ̃mۂB
   *  <li> _numberOfDataPacks ̒l -1  0 ɕύXB
   * <ol>
   *
   * ̃\bh̎s͏LM[qڑANł܂B
   * </japanese>
   *
   * @return a <code>boolean</code> value
   * @exception OhError if an error occurs
   */
  public void doSwitch() throws OhError {
    int k = 0;

    for (k = 0; k < _numberOfReceivePorts; k++)
      ((OhPort)_receivePortArray[k]).doSwitch();

    for (k = 0; k < _numberOfSendPorts; k++)
      ((OhPort)_sendPortArray[k]).doSwitch();
  }
  
  /**
   * Execute checks before calculation work. <p>
   *
   * Checks to see sending ports and receiving ports are in the
   * state which can send and receive data packs.
   *
   * <japanese>
   * <p>
   * vZƑO̊mF <p>
   *
   * M[qAM[qf[^pbN𑗎M\ȏԂł邱
   * ƂmF܂B<p>
   *
   * SM[q OhReceivePort.areYouReady ̖߂lAyёSM
   * [q OhSendPort.areYouReady ̖߂l true ł邱Ƃ
   * mF܂B߂l false ̒[q݂ꍇ false 
   * ԂA݂Ȃꍇ true Ԃ܂B<p>
   *
   * ̃\bh̖߂l true ̏ꍇAOhElement ͌vZs
   * ł܂B<p>
   *
   * false ̏ꍇ́Af[^pbNMłȂM[qA
   * ܂̓f[^pbN𑗐MłȂM[q݂A
   *  OhElement, M Element ŌvZsł܂B<p>
   *
   * ̏ꍇAڑƂɌ肪邩AOhSendPort  canSwitch() 
   * sĂȂ\܂B
   *
   * <pre>
   * [RETURN]
   *  Etrue --> S[q areYouReady ̖߂l true ̏ꍇB
   *  Efalse -> ߂l false ̒[q݂ꍇB
   * </pre>
   * </japanese>
   *
   * @return a <code>boolean</code> value
   * @exception OhError if an error occurs
   */
  public boolean areYouReady() throws OhError {
    boolean flag = true;    // For judgment.
    int k = 0;

    // --- Check "_receivePort". ---

    boolean bufFlag = true;  //; For judgment.
    for (k = 0; k < _numberOfReceivePorts; k++)
      if (((OhPort)_receivePortArray[k]).areYouReady() != true) {
        ((OhPort)_receivePortArray[k]).
          oPrint(new PrintWriter(System.err, true), "\t");
        bufFlag = false;
      }

    if (bufFlag != true) {
      throw new OhError("Receive ports or relay " +
        "ports are not connected." + oPrintString("\t"));
    }
    
    // --- Check "_sendPortArray". ---

    bufFlag = true;    //; For judgment.

    for (k = 0; k < _numberOfSendPorts; k++)
      if (((OhPort)_sendPortArray[k]).areYouReady() != true) {
        ((OhPort)_sendPortArray[k]).
          oPrint(new PrintWriter(System.err, true), "\t");
        bufFlag = false;
      }

    if (bufFlag != true) {
      throw new OhError(oPrintString("\t"));
    }

    // --- Check "_objectArray". ---

    bufFlag = true;    // For judgment.

    for (k = 0; k < _numberOfObjects; k++)
      if (_objectArray[k]== null) {
        throw new OhError(
          "\t_objectArray[" + k + "] has not set." + NEWLINE +
          oPrintString("\t")
        );
      }

    if (bufFlag != true) {
      throw new OhError(oPrintString("\t"));
    }

    return flag;
  }

    


  // 3. numerical initialization
  
  /**
   * Sets the data string.
   *
   * @param aDataStr a <code>String</code> value
   * @exception OhError if an error occurs
   */
  public void setDataString(String aDataStr) throws OhError { 
    if (_dataString != null) {
      throw new OhError(oPrintString("\t"));
    }
    _dataString = aDataStr;
  }

  
  /**
   * Clear the data string.
   *
   */
  public void clearDataString() {
    _dataString = null;
  }

  
  /**
   * Read a line from the input stream <b>aBr</b> and store the
   * line into <b>aStr</b>. The lines which start with "#" or
   * contain only white characters are regarded as comment lines
   * and skipped.
   *
   * @param aBr a <code>OhBufferedReader</code> value
   * @param aStr a <code>String</code> value
   * @param aStockP a <code>boolean</code> value
   * @return a <code>String</code> value
   * @exception OhError if an error occurs
   */
  public String getLine(OhBufferedReader aBr, String aStr,
    boolean[] aStockP, int[] aIntFlag) throws OhError {
    try {
      String str = null;

      if (aStockP[0] == false) {
        while (true) {
          if ((aStr = aBr.readLine()) == null)
            return null;
          
          if (aStr.length() == 0 ||
            aStr.substring(0, 1).equals("#"))
            continue;
          
          str = aStr.trim();
          
          break;
        }
        
        aIntFlag[0] = str.charAt(0);

      }
      else {
        aStockP[0] = false;
        str = aStr.trim();

        aIntFlag[0] = str.charAt(0);

      }
    }
    catch (Exception e) {
      throw new OhError("Error in getLine()." + NEWLINE +
        oPrintString("\t"));
    }
    
    return aStr;
  }
  
  /**
   * Gets the parameter.
   *
   * <japanese>
   * <p>
   * p^ǂݍރ\bhB
   * `ς݂ŁA[U͐VKɒ`Kv͂܂B
   * </japanese>
   *
   * @param aBr an <code>OhBufferedReader</code> value
   * @exception OhError if an error occurs
   */
  public void getParameter(OhBufferedReader aBr) throws OhError {

    try {
      if (aBr.skipWhites() == -1) {
        OhError e = new OhError("Parameter file error.");
        throw e;
      }
      
      if ((char)aBr.read() != '{') {
        throw new OhError("{ should come first" + NEWLINE +
          "Now reading parameters for " + NEWLINE +
          oPrintString(" > ")
        );
      }
      
      aBr.readLine();

      setParameter(aBr);    // Abstract method call.

      
      if (aBr.skipWhites() == -1) {
        OhError e = new OhError("Parameter file error.");
        throw e;
      }
      
      if ((char)aBr.read() != '}') {
        throw new OhError(
          "Should end with {" + NEWLINE +
          "Now reading parameters for " + NEWLINE +
          oPrintString(" > ")
        );
      }

      aBr.readLine();

    }
    catch (IOException e) {
      throw new OhError("Error in parameter line." +
        NEWLINE + oPrintString("\t"));
    }
  }


  
  /**
   * Gets the initial state.
   *
   * <japanese>
   * <p>
   * ԗʂǂݍރ\bh
   * </japanese>
   *
   * @param aBr an <code>OhBufferedReader</code> value
   * @exception OhError if an error occurs
   */
  public void getInitialState(OhBufferedReader aBr) throws OhError {
    try {
      if (aBr.skipWhites() == -1) {
        OhError e = new OhError("Initial state file error.");
        throw e;
      }
      
      if ((char)aBr.read() != '{') {
        throw new OhError("{ should come first" + NEWLINE +
          "Now reading initial state for " + NEWLINE +
          oPrintString(" > ")
        );
      }
      
      aBr.readLine();
      setInitialState(aBr);    // Abstract method call.
      
      if (aBr.skipWhites() == -1) {
        OhError e = new OhError("Initial state file error.");
        throw e;
      }
      
      if ((char)aBr.read() != '}') {
        throw new OhError("Should end with {" + NEWLINE +
          "Now reading initial state for " + NEWLINE +
          oPrintString(" > ")
        );
      }

      aBr.readLine();

    }
    catch (IOException e) {
      throw new OhError("Error in initial state line."
        + NEWLINE + oPrintString("\t"));
    }
  }      
 
  
  /**
   * This is a function for sharing information, such as a
   * parameter, using direct communication.
   *
   * <japanese>
   * <p>
   * ̃\bh́AڒʐM𗘗păp^Ȃǂ̏L
   * 邽߂̃\bhłBlXȗp@l܂A
   * ΂IuWFNg̃p^𑼂̃IuWFNgłp
   * ꍇȂǂɗpł܂Bꂼꂪp^ 
   * setParameter() Őݒ肷̂ł͂ȂAݒ肵p^
   * lARs[Ă炤A邢͂̃p^
   * Ă炤AȂǂƂgl܂BL
   * ƂsȂꍇ true ԂĉBƂsȂȂ
   * ꍇ false ԂĉBftHgł͒P true 
   * 悤ɒ`Ă܂B
   * </japanese>
   *
   * @return a <code>boolean</code> value
   */
  public boolean shareInfo() {
    return true;
  }

  
  /**
   * This is the method which calls
   * (Ref "OhElement.shareInfo()").
   *
   * <japanese>
   * <p>
   * ̃\bh (Ref "OhElement#shareInfo()") Ăяo\b
   * hłB߂l (Ref "OperationFlag")  ^łB<p>
   *
   * doShareInfo() ́A<p>
   *
   * (Ref "OhElement#shareInfo()") ܂ȂĂȂꍇA
   *
   * <ol>
   *  <li> (Ref "OhElement#shareInfo()") ĂсA߂l
   *       true Ȃ
   *       _shareInfoDoneP  true ɐݒ肵A
   *       OperationFlag.E_DONE Ԃ܂B
   *  <li> (Ref "OhElement#shareInfo()") ̖߂l false
   *       Ȃ 񋤗LƂ
   *       ȂƂāAOperationFlag.E_UNDONE Ԃ܂B
   * </ol>
   *
   * (Ref "OhElement#shareInfo()") łɂłꍇ
   * OperationFlag.E_ALREADY Ԃ܂B
   * </japanese>
   *
   * @return an <code>OperationFlag</code> value
   * @exception OhError if an error occurs
   */
  public OperationFlag doShareInfo() throws OhError { 
    if (_shareInfoDoneP == true)
      return OperationFlag.E_ALREADY;
    if (shareInfo() == true)
      return OperationFlag.E_UNDONE;

    _shareInfoDoneP = true;
    return OperationFlag.E_DONE;
  }
  
  /**
   * The data pack recorded on the data of all the sending ports
   * connected to the element is written out to the file
   * specified by stream <b>aPw</b>. Work of inclusion is performed
   * using method <b>saveSendPortStock</b> of the same name of a
   * sending port. <p>
   *
   * This function is re-defined by (Ref "OhSubSystem").
   *
   * <japanese>
   * <p>
   * vfɐڑĂ邷ׂĂ̑M[q̃f[^Ɏ^Ă
   * f[^pbNAXg[ <b>aPw</b> Ŏw肳Ăt@
   * Cɏo܂B^̎d́AM[q̓̃\bh
   * (Ref "OhSendPort#saveSendPortStock()") gčsȂ܂B<p>
   *
   * ̃\bh́AOhSubSystem ł͍Ē`܂B
   * </japanese>
   *
   * @param aPw a <code>PrintWriter</code> value
   */
  public void saveSendPortStock(PrintWriter aPw) {
    if (_numberOfSendPorts <= 0)
      return;

    aPw.print(_className + " " + _objectName + " " + _objectNumber + " (" + NEWLINE);

    for (int k = 0; k < _numberOfSendPorts; k++) {
      aPw.print("(" + k + ") # sending ports number." + NEWLINE);
      ((OhSendPort)_sendPortArray[k]).saveSendPortStock(aPw);
    }

    aPw.print(")" + NEWLINE);
  }

  
  /**
   * The data character sequence <b>_dataString</b> is read and
   * data is recorded on the sending port connected to the
   * element.
   *
   * <japanese>
   * <p>
   * f[^ _dataString ǂ݂ƂAvfɐڑĂ鑗
   * M[qɃf[^^܂B<p>
   * </japanese>  
   *
   * @exception OhError if an error occurs
   */
  public void restoreSendPortStock() throws OhError {
    int k;
    int m;
    String dStr = _dataString;

    try {
      StringReader     sr = new StringReader(dStr);
      OhBufferedReader br = new OhBufferedReader(sr);

      for (int i = 0; i < _numberOfSendPorts; i++) {

        k = Integer.parseInt(br.readWord());
      
        if (k != i) {
          throw new OhError(oPrintString("\t"));
        }

        m = Integer.parseInt(br.readWord());

        for (int j = 0; j < m; j++) {
          String dataPackStr =
            br.readWord() + " " + br.readWord() +
            " " + br.readWord();
          ((OhSendPort)_sendPortArray[i]).
            sendByString(dataPackStr);
        }
      }
      
    }
    catch (IOException e) {
      e.printStackTrace();
    }
    catch (NumberFormatException e) {
      e.printStackTrace();
    }
  }

  // 4. time about a simulation
  
  /**
   * Returns <b>_currentTime</b>.
   *
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePoint getCurrentTime() {
    return _currentTime;
  }
  
  /**
   * Returns <b>_timeStep</b>.
   *
   * @return a <code>long</code> value
   */
  public OhTimeImmutable getTimeStep() {
    return _timeStep;
  }
  
  /**
   * Returns the nexttime.
   *
   * <japanese>
   * <p>
   * Xebv _currentTime ԂB
   * </japanese>
   *
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePointImmutable getNextTime() {
    return OhTimePoint.add(_currentTime, _timeStep);
  }
  
  /**
   * <japanese>
   * <p>
   * vZIԂB
   * </japanese>
   *
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePoint getTerminalTime() {
    return ((OhElement)_owner).getTerminalTime();
  }
  
  /**
   * Returns the medium term target time from '_owner'.
   *
   * <japanese>
   * <p>
   * ԖڕW_owner 擾܂B
   * </japanese>
   *
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePoint getMediumTermTargetTime() {
    return ((OhElement)_owner).getMediumTermTargetTime();
  }
  
  /**
   * Returns '_fixedTime'.
   *
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePoint getFixedTime() {
    return _fixedTime;
  }
  
  /**
   * Returns '_targetTime'.
   *
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePoint getTargetTime() {
    return _targetTime;
  }
  
  /**
   * <japanese>
   * <p>
   * _targetTime, _fixedTime  _currentTime Ɉv郁\b
   * hłBvfԔvZsȂvffł́A
   * setInitialState()  _currentTime ݒ肳ꂽɂ
   * \bhĂԕKv܂B
   * </japanese>
   */
  public void setCur2Fix2Tar() {
    OhTimePoint crnTime = new OhTimePoint();
    crnTime.copy(_currentTime);
    _targetTime = crnTime;

    OhTimePoint crnTp = new OhTimePoint();
    crnTp.copy(_currentTime);
    _fixedTime = crnTp;
  }

  
  /**
   * <japanese>
   * <p>
   * _currentTime ̍XV
   * </japanese>
   */
  public void renewCurrentTime() throws OhError {
    _currentTime.add(_timeStep);
  }
  
  /**
   * <japanese>
   * <p>
   * _currentTime  _fixedTime ɖ߂܂B
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public void rewindCurrentTime() throws OhError {
    if (_currentTime.compareTo(_fixedTime) <= 0) {
      throw new OhError(
        "'_currentTime' not exceeds '_fixedTime.'" +
        NEWLINE +
        oPrintString(" > ")
      );
    }
    _currentTime.copy(_fixedTime);
  }

  
  /**
   * Updates '_fixedTime'.
   *
   * <japanese>
   * <p>
   * _fixedTime  _currentTime ɐi߂܂B
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public void renewFixedTime() throws OhError {
    if (_currentTime.compareTo(_fixedTime) <= 0) {
      throw new OhError(
        "'_fixedTime' exceeds '_currentTime.'" + NEWLINE +
        oPrintString(" > ")
      );
    }

    _fixedTime.copy(_currentTime);
  }

  
  /**
   * This function is not used by OhElement. It uses re-defining
   * by OhSubSystem. The reason for having defined this function
   * is that it is necessary to call this function on principle
   * also to the OhElement type composition element.
   *
   * <japanese>
   * <p>
   * ̃\bh  OhElement ł͎g܂BOhSubSystem 
   * Ē`Ďg܂BȂ̃NXŒ`ƂƁA
   * OhElement ł\vfɑ΂ĂAOÃ\bh
   * ĂԕKv邩łBڍׂ́A
   * OhSubSystem.initializeTargetTime() QƂĉB
   * </japanese>
   */
  public void initializeTargetTime() throws OhError {}

  
  /**
   * Updates '_targetTime'.
   *
   * <japanese>
   * <p>
   * _targetTime XV܂B
   * ɔvZԊԊuƂ܂B
   * </japanese>
   *
   * @param aT a <code>long</code> value. Repeated calculation time interval.
   */
  public void renewTargetTime(OhTimeImmutable aT) {
    OhTime time = new OhTime(aT.getSec(), aT.getFrac());
    OhTimePoint tarTime = OhTimePoint.add(_targetTime, aT);
    if (tarTime.compareTo(getTerminalTime())  > 0) {
      if ((OhOption.getDefault().getTimeControlOption() ==
            TimeControl.DO_NOT_GO_BEYOND_TERMINAL_TIME)
        && (_targetTime.compareTo(getTerminalTime()) < 0)) {
        time = OhTimePoint.sub(getTerminalTime(), _targetTime);
      }
    }
    _targetTime.add(time);
  }

  
  /**
   * Calculates repeated calculation target time.
   *
   * <japanese>
   * <p>
   * vZڕWvZ܂B
   * </japanese>
   *
   * @return a <code>OhTime</code> value
   */
  public OhTime calculateIterationStep() throws OhError {
    throw new OhError(
      "'_fixedTime' exceeds '_currentTime.'" + NEWLINE +
      oPrintString("\t")
    );
  }

  
  /**
   * Returns the data record time of the data which
   * <b>OhElement</b> should receive from <b>aRp</b> out of data
   * required for next calculation. <p>
   *
   * <b>OhElement</b> cannot start calculation if the data
   * pack from
   * <b>getNecessaryTimeFrom</b> to <b>getNecessaryTimeTo</b>
   * is not
   * received from the receiving terminal <b>aRp</b>,
   * <b>OhElement</b>
   * cannot start calculation.
   *
   * <japanese>
   * <p>
   * vf̌vZɕKvȃf[^̂AaRp (arg) 󂯎
   * ׂf[^̃f[^L^Ԃ܂B<p>
   *
   * getNecessaryTimeFrom()  getNecessaryTimeTo() ܂ł̃f[
   * ^pbNM[q aRp MĂȂ΁Avf͌vZ
   * Jn邱Ƃł܂B
   * </japanese>
   *
   * @param aRp an <code>OhReceivePort</code> value
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePointImmutable getNecessaryTimeFrom(
    OhReceivePort aRp) {
    return _currentTime;
  }

  
  /**
   * Returns the data record time of the data which
   * <b>OhElement</b> should receive from <b>aRp</b> out of data
   * required for next calculation. <p>
   *
   * <b>OhElement</b> cannot start calculation if the
   * data pack from
   * <b>getNecessaryTimeFrom</b> to <b>getNecessaryTimeTo</b>
   * is not
   * received from the receiving terminal <b>aRp</b>,
   * <b>OhElement</b>
   * cannot start calculation.
   *
   * <japanese>
   * <p>
   * vf̌vZɕKvȃf[^̂AaRp (arg) 󂯎
   * ׂf[^̃f[^L^Ԃ܂B<p>
   *
   * getNecessaryTimeFrom()  getNecessaryTimeTo() ܂ł̃f[
   * ^pbNM[q aRp MĂȂ΁Avf͌vZ
   * Jn邱Ƃł܂B
   * </japanese>
   *
   * @param aRp an <code>OhReceivePort</code> value
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePointImmutable getNecessaryTimeTo(OhReceivePort aRp) {    //; (*32)
    return getNextTime();
  }

  
  /**
   * <b>_currentTime</b> and <b>_timeStep</b> of
   * <b>OhElement</b> are
   * initialized to <b>aTp</b>.
   *
   * <japanese>
   * <p>
   * _currentTime, _timeStep  aTp , 0 ɏ܂B<p>
   *
   * aTp  0 ȂA_currentTime ύX܂B<p>
   *
   * (Ref "OhSubSystem") g _currentTime 
   * ƂɁÅe\vf _curentTime ݒ肵܂B̂ƂA
   * \bh (Ref "OhSubSystem#initialTime()") gp܂B
   * <p>
   *
   * TuVXe OhSubSystem 悤Ƃ鎞 tps 
   * ƂƁA\vf _currentTime ́Aŏ tps ɐݒ肳
   * ܂B̌A\vfgɑ΂āA _currentTime 
   * 邽߂ɁAԃt@C^ꂽ 
   * tpe łƂƁA OhElement::initialTime(tpe) 
   * ΂܂B<p>
   *
   * ۂɂ́Atps  tpe ͓ł邱ƂƎv܂B
   * ŁAOhElement::initialTime(OhTime) ̈ 
   * OhTimePoint.ZERO łƂɂ́A_currentTime ̕ύXȗ
   * 悤ɂĂ܂B̂߁Aԃt@C̒ŁA
   * tpe w肷Ƃł́AOhTimePoint.ZERO Ӗf[^
   * ƂłAt@C쐬ƂȕւɂȂ܂B
   * </japanese>
   *
   * @param aTp an <code>OhTimePoint</code> value
   */
  public void initialTime(OhTimePoint aTp) {
    if (aTp.compareTo(OhTimePoint.ZERO) > 0) {
      //
      // Note that _currentTime has been set to new
      // OhTimePoint() in the constructors of the OhElement
      // class. So _currentTime is not null.
      //
      _currentTime.copy(aTp);
    }
    _timeStep = new OhTime(); // same as OhTime(0, 0.0);
  }

  // 5. execution of calculation
  
  /**
   * <japanese>
   * <p>
   *  <p>
   *
   * ɂāASf[^擾܂B
   * </japanese>
   *
   * @return a <code>boolean</code> value
   * @exception OhError if an error occurs
   */
  public boolean initialWork() throws OhError {
    if (initialOutput() == false)    //; Throws "OhError".
      return false;

    setDataPackArray();
    
    return true;
  }
  
  /**
   * <japanese>
   * <p>
   * yƂ܂ȂĂȂꍇz
   *
   * <ol>
   *  <li> łׂĂ̎M[q̃f[^擾ł΁A
   *       Ƃ\łƔfāAinitialWork() ĂсA
   *       _initialWorkdoneP  true ɐݒ肵
   *       OperationFlag.E_DONE
   *       Ԃ܂B
   *  <li> ŏȂƂЂƂ̎M[q̃f[^擾łȂƂ́A
   *       Ƃs\łƔf
   *       OperationFlag.E_UNDONE
   *       Ԃ܂B
   * </ol>
   *
   * yƂłɂłꍇz
   *
   * OperationFlag.E_ALREADY
   * Ԃ܂B
   * </japanese>
   *
   * @return an <code>OperationFlag</code> value
   * @exception OhError if an error occurs
   */
  public OperationFlag doInitialWork() throws OhError {
    if (_initialWorkDoneP == true)
      return OperationFlag.E_ALREADY;
    if (initialWork() != true)       //; Throws "OhError".
      return OperationFlag.E_UNDONE;

    _initialWorkDoneP = true;
    return OperationFlag.E_DONE;
  }
  
  /**
   * <japanese>
   * <p>
   * canYouCalculate() ̌ <p>
   *
   * getNecessaryTimeFrom(), getNecessaryTimeTo() Œ`鎞Ԃ̊Ԃ
   * f[^擾łΌvZ\ƔfB
   * IԂ̂RłB
   * getNecessaryTimeFrom(), getNecessaryTimeTo() \bh́A
   * [URɍĒ`ł܂AftHgł
   * getNecessaryTimeFrom()  _currentTime ԂA
   * getNecessaryTimeTo  _currentTime + _timeStep ̒lԂ܂B<p>
   *
   * f[^擾s\łƕԎĂM[qɑ΂ČvZvo܂B
   * ̓Iɂ́AM[q calReq() bZ[W𑗂܂B
   *
   * Ƃ΁A
   *
   * <pre>
   *  public boolean canYouCalculate() {
   *    return canYouCalculate0();
   *  }
   * </pre>
   *
   *ƒ`΂悢łB
   * </japanese>
   *
   * @return a <code>boolean</code> value
   * @exception OhCannotGetDataError if an error occurs
   * @exception OhError if an error occurs
   */
  public boolean canYouCalculate0() throws OhCannotGetDataError, OhError {    //; (*36)
    boolean reply = true;
    int k;

//; System.out.println(_numberOfReceivePorts);

    for (k = 0; k < _numberOfReceivePorts; k++) { // for loop start.

      OhTimePointImmutable tFrom;
      OhTimePointImmutable tTo;

      Object rp = _receivePortArray[k];
      
      tFrom = getNecessaryTimeFrom((OhReceivePort)rp);    //;  _currentTime ԂB
      tTo   = getNecessaryTimeTo((OhReceivePort)rp);

      if (((OhPort)rp).canYouGetData(tFrom, tTo) != true) {
        ((OhPort)rp).calReq();
        reply = false;
      }
      
    }    // -------------------------------------------------------- for loop end.

    return reply;
  }

  
  /**
   * Sample code for canYouCalculate().
   *
   * <japanese>
   * <p>
   * canYouCalculate() ̌ <p>
   *
   * getNecessaryTimeFrom() Œ`f[^擾łΌvZ
   * \Ɣf܂Bȉꍇ getNecessaryTimeTo() \b
   * h`Ӗ͂܂Bj<p>
   *
   * f[^擾s\łƕԎĂM[qɑ΂Čv
   * Zvo܂B̓Iɂ́AM[q calReq() bZ[
   * W𑗂܂B<p>
   * </japanese>
   *
   * @return a <code>boolean</code> value
   * @exception OhCannotGetDataError if an error occurs
   * @exception OhError if an error occurs
   */
  public boolean canYouCalculate1()
    throws OhCannotGetDataError, OhError {  //; (*37)
    boolean reply = true;
    int k = 0;

    for (k = 0; k < _numberOfReceivePorts; k++) {    // ------------ for loop start.
      OhTimePointImmutable tFrom = new OhTimePoint();

      Object rp = _receivePortArray[k];

      tFrom = getNecessaryTimeFrom((OhReceivePort)rp);


      if (((OhPort)rp).canYouGetData(tFrom, tFrom) != true) {
        ((OhPort)rp).calReq();
        reply = false;
      }

    }    // -------------------------------------------------------- for loop end.

    return reply;
  }


  
  /**
   * Sends data packs. <p>
   *
   * 'OhSendPort setDataPackArray' is used for all possession sending ports,
   * and sending data pack arrangement is updated.
   * Thereby, old data pack arrangement is eliminated and new data pack arrangement is created. <p>
   *
   * But, nothing is carried out in fact.
   *
   * <japanese>
   * <p>
   * f[^pbN̑M <p>
   *
   * SĂ̏LM[q
   * OhSendPort.setDataPackArray()
   * pA
   * Mf[^pbNzXV܂B<p>
   * 
   * ɂAÂf[^pbNz͏AVf[^pb
   * Nz񂪍쐬܂B
   * </japanese>
   */
  public void setDataPackArray() {               //; (*38)
    for (int k = 0; k < _numberOfSendPorts; k++) {
      ((OhPort)_sendPortArray[k]).setDataPackArray();
    }
  }
  
  /**
   * Deletion of an unnecessary datapacks.
   *
   * <japanese>
   * <p>
   * svf[^pbN̏ <p>
   *
   * SĂ̏LM[q
   * OhReceivePort.reduceDataPackStock()
   * pA
   * MM[qɌvZI񍐂sȂ܂B<p>
   * 
   * ɂAMM[qŕsvf[^pbNAf[
   * ^pbNz񂪍XV܂B
   * </japanese>
   */
  public void reduceDataPackStock() {    //; (*39)
    int k;

    for (k = 0; k < _numberOfReceivePorts; k++)
      ((OhPort)_receivePortArray[k]).reduceDataPackStock();    //; "(OhSenderPort)_receivePortArray[k]" is NG.

    for (k = 0; k < _numberOfSendPorts; k++)
      ((OhPort)_sendPortArray[k]).reduceDataPackStock();
  }
    
  /**
   * Returns true.
   *
   * @return a <code>boolean</code> value
   */
  public boolean printStatus() {
    return true;
  }
  
  /**
   * Records status.
   *
   * @return a <code>boolean</code> value
   * @exception OhError if an error occurs
   */
  public boolean recordStatus() throws OhError {    //; (*41)
    if (_dataString == null) {
      throw new OhError(oPrintString("\t"));
    }
    printStatus();
    return true;
  }
  
  /**
   * 'Work' which performs 'calculate()' after obtaining
   * '_timeStep'.
   *
   * <japanese>
   * <p>
   * work() ̗ <p>
   *
   * _timeStep 𓾂ĂAcalculate() sB
   * </japanese>
   *
   * @return a <code>boolean</code> value
   * @exception OhCannotGetDataError if an error occurs
   * @exception OhError if an error occurs
   */
  public boolean work0() throws OhCannotGetDataError, OhError {             //; (*42)
    //; vZvȂA_currenetTime ԖڕWzĂ
    //; ΌvZȂB
    //;
    if (_calReqP == false &&
      _currentTime.compareTo(getMediumTermTargetTime()) >= 0)    //; (*27)̏
      return false;

    boolean calc = false;

    while(true) {
      // Setup of a time step.
      //
      _timeStep = calculateTimeStep();
      if (_timeStep.toDouble() <= 0.0) {    
        if (_timeStep.toDouble() == 0.0)
          return false;

        throw new IllegalArgumentException(
          "\t_currentTime = " + _currentTime + ", _timeStep = " +
          _timeStep + NEWLINE + oPrintString("\t")
        );
      }

      OhTimePoint crnTime =
        OhTimePoint.add(_currentTime, _timeStep);
      
      if (crnTime.compareTo(getTerminalTime()) > 0) {
        if (OhOption.getDefault().getTimeControlOption() ==
          TimeControl.DO_NOT_GO_BEYOND_TERMINAL_TIME) {
          _timeStep = OhTimePoint.
            sub(getTerminalTime(), _currentTime);
        }
      }

      // Check before calculation.
      //
      if (canYouCalculate() != true) {
        // Throw "OhCannotGetDataError".
        if (calc == false)
          return false;
        break;          // Finish without calculation
      }

      // Calculation of 1 step.
      //
      if (calculate() == false) {
        if (calc == false)
          return false;
        break;          // Finish without calculation
      }

      calc = true;
      _calReqP = false;

      // Update _currentTime.
      //
//;   if ((_currentTime = _currentTime.add(_timeStep)).compareTo(getMediumTermTargetTime()) >= 0)  //; 2003.12.14
      _currentTime.add(_timeStep);
      if (_currentTime.compareTo(getMediumTermTargetTime()) >= 0)
        break;
    }
    
    // Process after calculation.
    //
    setDataPackArray();
    reduceDataPackStock();
    return true;
  }
  
  /**
   * '_timeStep' is calculated inside 'calculate()'.
   *
   * <japanese>
   * <p>
   * work() ̗ <p>
   * _timeStep ́Acalculate() ̓ŌvZB
   * </japanese>
   *
   * @return a <code>boolean</code> value
   * @exception OhCannotGetDataError if an error occurs
   * @exception OhError if an error occurs
   */
  public boolean work1() throws OhCannotGetDataError, OhError {              //; (*43)
    //;  vZvȂAcurrenet_time ԖڕWzĂ
    //;  ΌvZȂB
    //;
    if (_calReqP == false &&
      _currentTime.compareTo(getMediumTermTargetTime()) >= 0)
      return false;

    boolean calc = false;

    while(true) {
      boolean flgDbg;

      // Check before calculation.
      //
      if ((flgDbg = canYouCalculate()) != true) {
        //; Throw "OhCannotGetDataError".
        if (calc == false)
          return false;
        break;          // Finish without calculating.
      }

      // Calculation of 1 step.
      //
      if (calculate() != false) {
        if (calc == false)
          return false;
        break;          // Finish without calculating.
      }

      calc = true;
      _calReqP = false;

      // Update _currentTime.
      //
      if (_timeStep.toDouble() <= 0.0)
        break;

//;   _currentTime = _currentTime.add(_timeStep);    //; 2003.12.14
      _currentTime.add(_timeStep);

      if (_currentTime.compareTo(getMediumTermTargetTime()) >= 0)
        break;
    }

    // Process after calculation.
    //
    setDataPackArray();
    reduceDataPackStock();
    
    return true;
  }
  
  
  /**
   * For the element which carries out repeated calculation.
   *
   * <japanese>
   * <p>
   * vZvfp work() ̗
   * </japanese>
   *
   * @return a <code>boolean</code> value
   * @exception OhCannotGetDataError if an error occurs
   * @exception OhError if an error occurs
   */
  public boolean work2() throws OhCannotGetDataError, OhError {                  //; (*44)
    if (_fixedTime.compareTo(getTerminalTime()) >= 0)     // Finished calculation.
      return false;

    if (_currentTime.compareTo(getTerminalTime()) >= 0)   // Waiting other element.
      return false;

    boolean calc = false;

    
    while(true) {

      // Setup of a time step.
      //
      _timeStep = calculateTimeStep();
      if (_timeStep.toDouble() <= 0.0) {    // Abstract methos call.
        if (_timeStep.toDouble() == 0.0)
          return false;

        throw new IllegalArgumentException(
          "\t_currentTime = " + _currentTime +
          ", _timeStep = " + _timeStep + NEWLINE +
          oPrintString("\t")
        );
      }

      // Make not to exceed '_targetTime'.
      //
      OhTimePoint crnTime =
        OhTimePoint.add(_currentTime, _timeStep);
      if (crnTime.compareTo(_targetTime) > 0) {
        _timeStep = OhTimePoint.sub(_targetTime, _currentTime);
      }

      // Check before calculation.
      //
      if (canYouCalculate() != true) {
        // throw "OhCannotGetDataError".
        if (calc == false)
          return false;
        break;          // Finish without calculating.
      }

      // Calculation of 1 step.
      //
      if (calculate() != false) {
        if (calc == false)
          return false;
        break;          // Finish without calculating.
      }

      calc = true;

      // Update _currentTime.
      //
      _currentTime.add(_timeStep);
      if (_currentTime.compareTo(_targetTime) >= 0)
        break;
    }      

    // Process after calculation.
    //
    return true;
  }

  
  /**
   * This function judges whether repeated calculation reached
   * <b>_targetTime</b>.
   *
   * <japanese>
   * <p>
   * vZ <b>_targetTime</b> ɓBǂ𔻒f
   * BBĂ trueABĂȂ false Ԃ
   * B
   * </japanese>
   *
   * @return a <code>boolean</code> value
   */
  public boolean hasReachedTargetTime() {       //; (*46)
    if (_currentTime.compareTo(_targetTime) >= 0)
      return true;
    return false;
  }

  
  /**
   * Convergence of the repeated calculation between elements is
   * judged. Please be sure to re-define by the element which
   * performs repeated calculation. In the function, please
   * return 'true' if it converges. When not giving a
   * re-definition, an error message is sent and it exits.
   *
   * <japanese>
   * <p>
   * vfԂ̔vZ̎𔻒肵܂BvZsȂvf
   * KĒ`ĉB trueAĂȂ 
   * false ԂĉBĒ`Ȃꍇ̓G[bZ[W
   * o exit ܂B
   * </japanese>  
   *
   * @return a <code>boolean</code> value
   */
  public boolean hasConverged() throws OhError {
    throw new OhError(oPrintString("\t"));
  }

  /**
   * It is the function which carries out output of the data by
   * repetitive target time, and processing of a data pack.
   *
   * <japanese>
   * <p>
   * ڕW܂ł̃f[^̏o͂ƁAf[^pbN̏
   * \bhłB
   * </japanese>
   */
  public void flushData() throws OhError {  //; (*48)
    sendStockedData();
    
    setDataPackArray();

    reduceDataPackStock();
  }

  
  /**
   * Outputs the data accumulated untill the repetitive target
   * time. For example, please use 'OhSendDouble
   * sendData(OhTimePoint aT, double aValue)'. <p>
   *
   * Please be sure to re-define by the element which performs
   * repeated calculation. When not giving a re-definition, an
   * error message is sent and it exits.
   *
   * <japanese>
   * <p>
   * ڕW܂łɂ܂f[^o͂郁\bhłB
   * o͒[q́AႦ
   * OhSendDouble.sendData(OhTimePoint, double)
   * Ȃǂ𗘗pĉB
   * vZsȂvfŕKĒ`ĉB
   * Ē`Ȃꍇ̓G[bZ[Wo exit ܂B
   * </japanese>
   *
   */
  public void sendStockedData() throws OhError {
    throw new OhError(oPrintString("\t"));
  }

  
  /**
   * Response to the calculation request from s send port of
   * this element.
   * <p>
   *
   * <japanese>
   * <p>
   * ̗vfvZ邱ƂɂȂAKvƂȂf[^M
   * 邩`FbN郁\bh canYouCalculate() s邱
   * ŁA̗vfˑĂ鑼̗vf֌vZv`B܂B
   * <p>
   *
   * \bh canYouCalculate() sŁA̗vfւ̌vZ
   * vƂL^܂B̗vf work() ł̗v
   * l܂B<p>
   *
   * ̃\bh́A̗vf̑M[qĂ΂܂Bn
   * ́Ã\bhĂ΂邱Ƃ͂܂Bp[q́Av
   * Zv𑗐M[qɓ]邾ŁAp[q̏L҂ł镔
   * nɂ͌vZv𑗂ȂłB<p>
   *
   * ̒`sK؂łꍇ́ApNXōĒ`
   * B
   * </japanese>
   *
   * @exception OhCannotGetDataError if an error occurs
   * @exception OhError if an error occurs
   */
  public void calReq() throws OhCannotGetDataError, OhError {
    OhTime timeStepBackup = _timeStep;
    _timeStep = calculateTimeStep();    // Abstract method call.
                                        //; throws "OhError".
    canYouCalculate();                  //; Throw "OhCannotGetDataError" and "OhError".
    _timeStep = timeStepBackup;
    _calReqP = true;
  }


  
  // 6. information output function
  
  /**
   * Information output on possession objects.
   *
   * <japanese>
   * <p>
   * LIuWFNg̏̏o
   * </japanese>
   *
   * @param aPw a <code>PrintWriter</code> value
   * @param aHead a <code>String</code> value
   * @param aTitle a <code>String</code> value
   * @param aPortArray an <code>Object[]</code> value
   * @param aNumOfObjs an <code>int</code> value
   */
  public void printObjects(PrintWriter aPw, String aHead,
    String aTitle, Object[] aPortArray, int aNumOfObjs) {
    if (aTitle != null) {
      aPw.print(aHead + aTitle);
      oPrint(aPw, "of : ");
    }
    else
      oPrint(aPw, aHead);

    for (int k = 0; k < aNumOfObjs; k++) {
      aPw.print(aHead);
      ((OhObject)aPortArray[k]).cPrint(aPw, "    ");
    }
  }

  
  /**
   * Information output on possession receiving ports.
   *
   * <japanese>
   * <p>
   * LM[q̏̏o
   * </japanese>
   *
   */
  public void printReceivePorts(PrintWriter aPw, String aHead) {
    printObjects(aPw, aHead, "receiving ports :",
      _receivePortArray, _numberOfReceivePorts);
  }

  
  /**
   * Information output on possession receiving ports.
   *
   * <japanese>
   * <p>
   * LM[q̏̏o
   * </japanese>
   */
  public void printReceivePorts(PrintWriter aPw) {
    printObjects(aPw, "    ", "receiving ports :",
      _receivePortArray, _numberOfReceivePorts);
  }

  
  /**
   * Information output on possession receiving ports.
   *
   * <japanese>
   * <p>
   * LM[q̏̏o
   * </japanese>
   */
  public void printReceivePorts(String aHead) {
    PrintWriter pw = new PrintWriter(System.err, true);
    printObjects(pw, aHead, "receiving ports :",
      _receivePortArray, _numberOfReceivePorts);
  }

  
  /**
   * Information output on possession receiving ports.
   *
   * <japanese>
   * <p>
   * LM[q̏̏o
   * </japanese>
   */
  public void printReceivePorts() {
    PrintWriter pw = new PrintWriter(System.err, true);
    printObjects(pw, "    ", "receiving ports :",
      _receivePortArray, _numberOfReceivePorts);
  }

  
  /**
   * Information output on possession sending ports.
   *
   * <japanese>
   * <p>
   * LM[q̏̏o
   * </japanese>
   */
  public void printSendPorts(PrintWriter aPw, String aHead) {
    printObjects(aPw, aHead, "sending ports :",
      _sendPortArray, _numberOfSendPorts);
  }

  
  /**
   * Information output on possession sending ports.
   *
   * <japanese>
   * <p>
   * LM[q̏̏o
   * </japanese>
   */
  public void printSendPorts(PrintWriter aPw) {
    printObjects(aPw, "    ", "sending ports :",
      _sendPortArray, _numberOfSendPorts);
  }

  
  /**
   * Information output on possession sending ports.
   *
   * <japanese>
   * <p>
   * LM[q̏
   * </japanese>
   */
  public void printSendPorts(String aHead) {
    PrintWriter pw = new PrintWriter(System.err, true);
    printObjects(pw, aHead, "sending ports :",
      _sendPortArray, _numberOfSendPorts);
  }

  
  /**
   * Information output on possession sending ports.
   *
   * <japanese>
   * <p>
   * LM[q̏
   * </japanese>
   */
  public void printSendPorts() {
    PrintWriter pw = new PrintWriter(System.err, true);
    printObjects(pw, "    ", "sending ports :",
      _sendPortArray, _numberOfSendPorts);
  }

  
  /**
   * Information output on connection.
   *
   * <japanese>
   * <p>
   * ڑ̏o
   * </japanese>
   */
  public void printConnection(PrintWriter aPw, String aHead) {
    String str  = aHead + "    ";
    int k;

    for (k = 0; k < _numberOfReceivePorts; k++) {
      ((OhSenderPort)_receivePortArray[k]).cPrint(aPw, aHead);
      ((OhSenderPort)_receivePortArray[k]).
        printConnection(aPw, str);
    }

    for (k = 0; k < _numberOfReceivePorts; k++) {
      ((OhSenderPort)_sendPortArray[k]).cPrint(aPw, aHead);
      ((OhSenderPort)_sendPortArray[k]).printConnection(aPw, str);
    }
  }

  
  /**
   * Information output on connection.
   * At this time, the header is "    ".
   *
   * <japanese>
   * <p>
   * ڑ̏o
   * </japanese>
   */
  public void printConnection(PrintWriter aPw) {
    String head = "    ";
    String str  = "        ";
    int k;

    for (k = 0; k < _numberOfReceivePorts; k++) {
      ((OhSenderPort)_receivePortArray[k]).cPrint(aPw, head);
      ((OhSenderPort)_receivePortArray[k]).
        printConnection(aPw, str);
    }

    for (k = 0; k < _numberOfReceivePorts; k++) {
      ((OhSenderPort)_sendPortArray[k]).cPrint(aPw, head);
      ((OhSenderPort)_sendPortArray[k]).printConnection(aPw, str);
    }
  }

  
  /**
   * Information output on connection.
   *
   */
  public void printConnection(String aHead) {
    PrintWriter pw = new PrintWriter(System.err, true);
    String str  = aHead + "    ";
    int k;

    for (k = 0; k < _numberOfReceivePorts; k++) {
      ((OhSenderPort)_receivePortArray[k]).cPrint(pw, aHead);
      ((OhSenderPort)_receivePortArray[k]).printConnection(pw, str);
    }

    for (k = 0; k < _numberOfSendPorts; k++) {
      ((OhSenderPort)_sendPortArray[k]).cPrint(pw, aHead);
      ((OhSenderPort)_sendPortArray[k]).printConnection(pw, str);
    }
  }

  
  /**
   * Information output on connection.
   *
   */
  public void printConnection() {
    PrintWriter pw = new PrintWriter(System.err, true);
    String head = "    ";
    String str  = "        ";
    int k;

    for (k = 0; k < _numberOfReceivePorts; k++) {
      ((OhSenderPort)_receivePortArray[k]).cPrint(pw, head);
      ((OhSenderPort)_receivePortArray[k]).printConnection(pw, str);
    }

    for (k = 0; k < _numberOfReceivePorts; k++) {
      ((OhSenderPort)_sendPortArray[k]).cPrint(pw, head);
      ((OhSenderPort)_sendPortArray[k]).printConnection(pw, str);
    }
  }

  // 7. the function to save the last state at a file
  
  /**
   * <japanese>
   * <p>
   * ŏIԂ̕ۑ
   * </japanese>
   *
   * @param aPw a <code>PrintWriter</code> value
   */
  public void terminalState(PrintWriter aPw) {
    aPw.print(_className + " " + _objectName + " "
      + _objectNumber + NEWLINE);
    aPw.print("{" + NEWLINE);

    saveTerminalState(aPw);

    aPw.print(NEWLINE);
    aPw.print("}" + NEWLINE);
  }

  /**
   * Creates an Element which has the class name specified by
   * the argument.<p>
   *
   * <japanese>
   * Ŏw肳ꂽNX OhElement IuWFNg쐬
   * ĕԂ܂Bw肳ꂽNX́ApbP[W ohymosj ɑ
   * ̂Ƃ܂B<p>
   *
   * . OhElement elm = OhElement.create("Storage");
   * </japanese>
   *
   * @param aClassName a <code>String</code> value
   * @return an <code>OhElement</code> value
   */
  public static OhElement create(String aClassName)
    throws OhError {
    OhElement elm = null;
    try {
      elm = (OhElement)Class.forName("ohymosj." + aClassName).newInstance();
    }
    catch (Exception e) {
      throw new OhError("Cannot create " + aClassName + NEWLINE +
        OhError.getStackTraceString(e));
      // System.out.println(OhError.getStackTraceString(e));
    }
    return elm;
  }
  
  // @end {methods} --------------------------------------------

} // OhElement
