/**
 * @(#) OhInputFile; abstract class to express port models of input files.
 *
 * (Ref "OhObject")
 * <-- (Ref "OhPort")
 * <-- (Ref "OhSenderPort")
 * <-- (Ref "OhSendPort")
 * <-- (Ref "OhInputPort")
 * <-- (Ref "OhInputFile")
 */

package ohymosj;

import java.util.*;
import java.io.*;

import java.util.regex.Pattern;
import java.util.regex.Matcher;

/**
 * The OhInputFile class is an abstract class of port models of
 * input files. It extends (Ref "OhInputPort").<p>
 *
 * An object of "OhInputFile" class
 *
 * <dl><dd>
 * <dl><dt> (Object) _objectName
 *       <dd> input file name
 *     <dt> _fileTimeAtInput
 *       <dd> final update time of the input file
 * </dl></dl>
 *
 * and has
 *
 * <dl><dd>
 * function to read in data and send them as datapacks to the
 * send port.
 * </dl>
 *
 * <japanese>
 * <P>
 * ̓t@CpM[qf <p>
 *
 * ̓t@CpM[qf́AM[q݂̂vff
 * łƉ߂邱Ƃł܂BȂ킿Ãf́At@
 * Cf[^ǂݍ݁Af[^pbNɕϊđM[qɃf[
 * ^pbNXgbN܂B<p>
 *
 * ۂɂ́A̓t@CpM[qf́Avff̂P
 * ăfꂸɁAM[qf̔hfƂĎ
 * Ă܂BȂ킿AM[qfɃt@Cǂݎ@\
 * ǉāA̓t@CpM[qf쐬Ă܂B
 * ̊֌Ŵ͎悤łB<p>
 *
 * NX OhInputFile ł́A<p>
 *
 *  E̓t@C             : (Ref "OhObject#_objectName")<br>
 *  E̓t@C̍ŏIXV : _fileTimeAtInput <p>
 *
 * LA<p>
 *
 *  E̓t@C̃f[^pbNM@\ <p>
 *
 * `܂B<p>
 *
 * p^t@C⏉lf[^t@CȂǂ́ANX
 * (Ref "OhTotalSystem") ̃IuWFNg
 * (Ref "OhTotalSystem#readParameterFile()") 
 * (Ref "OhTotalSystem#readInitialStateFile()") Ȃǂ
 * pēǂݎƂ܂B
 * āAp^⏉lȂǂ́A[qʂċ
 * ł͂܂BOhInputFile ̓t@CƂ́AԓI
 * ɕωf[^^Ăt@CłB<p>
 *
 * NX OhInputFile ̃\bh
 * (Ref "OhInputFile#inputData()") ́A
 * f[^Ԍ^zw肷si#& dctVec Ŏn܂sj
 * t@Cǂݎ@\Ă܂B<p>
 *
 * NX OhInputFile ̃\bh 
 * (Ref "OhInputFile#inputData()") ǂݎf[^̗܂B
 *
 * <pre>
 *  # rainfall.ia
 *  # ----------------------------------------------------------
 *  # Year(int) Sec(long) Frac(double) R(double, mm/hour)
 *  # DataConnectionType = 0:BROKENLINE 1:STEP 2:PULSE 3:NOINTP
 *  #& dctvec 1 1
 *  #
 *
 *  2000     0 0.0 0.0
 *  2000  7200 0.0 5.0
 *  2000 14400 0.0 0.0
 *  2000 50000 0.0 0.0
 * </pre>
 *
 * s # Ŏn܂s̓RgsłB<p>
 *
 * As # ̒ & s͈ӖĂāA
 * ̉pPɂāÄӖقȂ܂B̗ŁA#& dctvec 
 * ƏĂs܂ÁAf[^Ԍ^zw
 * ĂsłBf[^Ԍ^ɂẮANX
 * (Ref "DataConnectionType") QƂĂB̗́Af[
 * ^1̃f[^ŁÃf[^Ԍ^́ASTEP Ȃ킿AKi
 * łƎw肵Ă܂B<p>
 *
 * f[^̍ŏ3̃f[^́A(Ref "OhTimePoint") ^̃f[
 * ^Ńf[^\킵Ă܂B̌̃f[^́Af[^lŁA
 * ̗ł́AOhTimePoint(2000, 0, 0.0) ̎̃f[^ĺA0.0 
 * łB<p>
 *
 * OHyMoSJ ł́A(Ref "OhInputFile") A(Ref "OhInputFileDouble"),
 * (Ref "OhInputFileFloat"), (Ref "OhInputFileLong"), (Ref "OhInputFileInt")
 * Ȃǂpč쐬Ă܂BvO[UA
 * (Ref "OhInputFile") pāA[UƎ̓͒[qNX
 * 쐬邱Ƃł܂B<p>
 *
 * [UƎ̓͒[qf̒ۃNXhč쐬
 * Ƃ́AΉf[^pbN^`ƂƂɁA̒
 * NX<p>
 *
 * <pre>
 *   public abstract String dataType();
 *   public abstract OhDataPack makeDataPack(String aLineData);
 * </pre>
 *
 * ̓̒ۃ\bhÃf[^pbN^ɑΉĒ`
 * B
 *
 * </japanese>
 *
 * @author satoh, shiiba
 * @version $Revision: 1.32 $ $Date: 2004/01/16 03:35:41 $
 */
public abstract class OhInputFile extends OhInputPort
{
  // @begin {instance variable or static variable} -------------

  /**
   * Final update time of the input file.
   *
   * <japanese>
   * <p>
   * ̓t@C̍ŏIXV <p>
   *
   * Vf[^擾ꂽƂɁAO̓̓t@Ct@
   * Cɒǉ邩A㏑邩Iׂ悤ɑz肵Ă
   * ܂B̂߂ɁAf[^ǂݍ񂾂ƂɁAt@C̍ŏI
   * XVL^ĂAɃt@CǂޕKvƂ
   * ɁAt@C̍ŏIXVāAt@CXVꂽ
   * 悤ɂ܂B
   *
   * </japanese>
   *
   */
  long _fileTimeAtInput;

  // @end {instance variable or static variable} ---------------


  // @begin {abstract methods} ---------------------------------

  // 1. connection function

  /**
   * Returns the string showing the type of the sending datapacks.
   *
   * <japanese>
   * <p>
   * Mf|^pbN^\Ԃ
   * </japanese>
   *
   * @return a <code>String</code> value
   */
  public abstract String dataType();


  // 2. sending function of the input file data
  
  /**
   * Makes datapacks from string data.
   *
   * <japanese>
   * <p>
   * f[^̃f[^pbN̍쐬 <p>
   *
   * f[^f[^pbN쐬܂B
   * ۃ\bhłB
   * hNXŎ邱ƂvĂ܂B<p>
   *
   * Psf[^P̃f[^pbNɕϊ郁\bh
   * hNXŒ`ĉB<p>
   *
   * WIȔhNXł́A
   * Mf[^pbN^ɉĎ̎葱`ĉB<p>
   *
   * <pre>
   * (1) Mpf[^pbN̋L̈q[vɐB
   * (2)  aLineData f[^L^Af[^lǂݎB
   * (3) (2) œǂݎl (1) ̃f[^pbÑoɑB
   * (4) (1) ̃f[^pbN Object ^ɃLXgĕԂB
   * </pre>
   *
   * ۃ\bhȂ̂ŔhNXŕK`ĉB
   *
   * <pre>
   * [ARGUMENT]
   *  EaLineData -> ̓f[^Ps̕
   * [RETURN]
   *  EObject : Mf[^pbN Object ^ɃLXǵB
   * </pre>
   * </japanese>
   *
   * @param aLineData a <code>String</code> value
   * @return an <code>Object</code> value
   */
  public abstract OhDataPack makeDataPack(String aLineData);

  // @end {abstract methods} -----------------------------------


  // @begin{initialization} ------------------------------------

  /**
   * Initialization.
   *
   * <japanese>
   * ۂɃt@CI[v̂́Ainit ̒ł͂ȂāA
   * (Ref "#inputData()") ̒ɂĂłB
   * </japanese>
   *
   * @param aFileName a <code>String</code> value
   * @return a <code>boolean</code> value
   * @exception OhError if an error occurs
   */
  public void init(String aFileName) throws OhError {
    super.init(aFileName, 0);
  }

  // @end{initialization} --------------------------------------

  
  // @begin {constructors} -------------------------------------
  
  /**
   * Creates a new default <code>OhInputFile</code> instance.
   *
   * @exception OhError if an error occurs
   */
  public OhInputFile() {
    super();
    _fileTimeAtInput = Long.MIN_VALUE;
    _className = "OhInputFile";
  }

  /**
   * Creates a new <code>OhInputFile</code> instance.
   *
   * @param aFileName a <code>String</code> value
   * @exception OhError if an error occurs
   */
  public OhInputFile(String aFileName) {
    super();
    _fileTimeAtInput = Long.MIN_VALUE;
    _className = "OhInputFile";
    _objectName = aFileName;
    _objectNumber = 0;
  }

  // @end {constructors} ---------------------------------------

  // @begin {methods} ------------------------------------------
    
  /**
   * When the input port receives a calculation request, throw
   * (Ref "OhCannotGetDataError"). This will be caught in the catch
   * block in OhTotalSystem.work().
   *
   * @exception OhCannotGetDataError if an error occurs
   * @exception OhError if an error occurs
   */
  public void calReq() throws OhCannotGetDataError, OhError {
    String pauseComment =
      "Cannot get data from " + _objectName;
    
    OhInterface.pause(pauseComment);
    throw new OhCannotGetDataError(pauseComment);
  }
  
  /**
   * The information on this input file is printed to the
   * specified PrintWriter <b>aPw</b> with the preceding String
   * <b>aHead</b>
   *
   * <japanese>
   * <p>
   * ̓̓t@C̏Xg[ <b>aPw</b> ɏo͂܂B
   * o͂ƂɁAes̍sɁA <b>aHead</b> t
   * ܂B
   * <p>
   * </japanese>
   *
   * @param aPw a <code>PrintWriter</code> value
   * @param aHead a <code>String</code> value
   */
  public void cPrint(PrintWriter aPw, String aHead) {
    aPw.print(aHead + _className + 
      " " + _objectName);
    aPw.print(System.getProperty("line.separator"));
  }
    
  /**
   *
   * <japanese>
   * <p>
   * ̓t@CǂݎA
   * (Ref "OhInputFile#makeDataPack(String)") ɂ
   * f[^pbN쐬A쐬f[^pbN
   * (Ref "OhSendPort#dataPackStock(Object)") 
   * pđM (~) ܂BiOhInputFile  OhSendPort ̈
   * łBj
   * <p>
   *
   * t@CǂݎƂAs # ̏ꍇA
   * ̍sRgsƌȂ܂B<p>
   *
   * As # łǍ & ƁAȂ킿 #& 
   * n܂śARgł͂ȂAȃf[^LڂĂ
   * ̂Ƃ܂B<p>
   *
   * ݁i2004/01/16ĵƂA
   * (Ref "DataConnectionType") w肷ꍇ邾łB
   *
   * <pre>
   * # DataConnectionType = {0:BROKENLINE, 1:STEP, 2:PULSE, 3:NOINTP}
   * #& dctvec 2 0 1
   * </pre>
   *
   * ̂悤ɏĂƁA(Ref "DataConnectionType") Vector 
   * w肷̂ƌȂ܂B̗́Af[^QāA
   * P BROKENLINE AQ STEP ł邱ƂӖ
   * ܂BwȂʂ̏ꍇ́A
   *
   * <pre>
   * #& dctvec 1 0
   * </pre>
   *
   * Ƃ̂ƓłB<p>
   *
   * t@C܂ŏIA
   * (Ref "OhSendPort#dataPackStock(Object)") 
   * p _dataPackArray 쐬܂B<p>
   *
   * A̓t@C̍ŏIύX _fileTimeAtInput ȑO
   * ̏ꍇ͉ return ܂BRXgN^ 
   * _fileTimeAtInput  Long.MIN_VALUE ɐݒ肳̂ŏ̓
   * ͕͂Ks܂B
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public void inputData() throws OhError {

    setDataPackArray();

    reduceDataPackStock();  // Delete the unnecessary datapacks.

    File f = new File(_objectName);
      // Checks work directory by "System.getProperty("user.dir")".

    if (f.exists() != true) {
      throw new OhError("Cannot find the file " + _objectName);
    }

    if (f.lastModified() <= _fileTimeAtInput)
      return;

    _fileTimeAtInput = f.lastModified();


    FileReader       fr       = null;
    BufferedReader br       = null;
    String           lineData = null;
    Pattern p = Pattern.compile("(\\w+) ");
    // pattern to match a keyword

    try {
      fr = new FileReader(_objectName);
      br = new BufferedReader(fr);
      //; Do not use OhBufferedReader(fr, true).

      while ((lineData = br.readLine()) != null) {
        
        if (lineData.length() <= 0)
          continue;

        if (lineData.substring(0, 1).equals(" ")
          || lineData.substring(0,1).equals("\t")) {
          continue;
        }
        
        if (lineData.substring(0, 1).equals("#")) { //  start "if"
          
          if (lineData.length() <= 1)
            continue;

          if (lineData.substring(1, 2).equals("&")) {
            parseSpecLine(lineData);
          }
          
          continue;    // Skip comment lines.
          
        }    // --------------------------------------- end "if"

        dataPackStock(makeDataPack(lineData));

      } // ----------------------------------------- end "while"

    }
    catch (IOException e) {
      throw new OhError(
        "Cannot open " + _objectName + "." + NEWLINE +
        oPrintString("\t")
      );
    }
    finally {
      if (fr != null) {
        try { fr.close(); } catch (IOException e) {}
      }
    }
    
    setDataPackArray();    //; Set "_numberOfDataPacks = _dataPackArray.size();"

    if (getNumberOfDataPacks() > 0) {
      OhTimePointImmutable tp = getNewestDataTime();

      StringWriter buffer = new StringWriter();
      PrintWriter  out    = new PrintWriter(buffer);
      out.print("File: " + _objectName +
        "  newest data time : " + tp.getYear() + " "
        + tp.getTsec() + " " + tp.getFrac());
      out.print(NEWLINE);
      String str = buffer.toString();
      out.close();  
      OhObject top = (OhObject)getTopOwner();
      if (top != null && top instanceof OhTotalSystem) {
        // (Ref "OhTotalSystem#addNewestDataTime()")
        ((OhTotalSystem) top).addNewestDataTime(str);
      }
    }
  }

  /**
   * The information on this input file is printed to the
   * specified PrintWriter <b>aPw</b> with preceding String
   * <b>aHead</b>
   *
   * <japanese>
   * <p>
   * ̓̓t@C̏Xg[ <b>aPw</b> ɏo͂܂B
   * o͂ƂɁAes̍sɁA <b>aHead</b> t
   * ܂B
   * <p>
   * </japanese>
   *
   * @param aPw a <code>PrintWriter</code> value
   * @param aHead a <code>String</code> value
   */
  public void oPrint(PrintWriter aPw, String aHead) {
    cPrint(aPw, aHead);
  }

  /**
   * The parseSpecLine method analyzes the line which starts
   * with "#& " and executes the specified operation by the
   * line. What operation will be done is specified in this
   * method. You can redifine this method in subclasses derived
   * from this OhInputFile class.<p>
   *
   * In this default implementation of the parseSpecLine()
   * method, only the dctvec line as <p>
   *
   * <pre>
   *    #& dctvec 2 1 0
   * </pre>
   *
   * are analyzed. This exmaple line specifies the
   * DataConnectionType vector Vector[2] =
   * { DataConnectionType.STEP, DataConnectionType.BROKENLINE }
   * <p>
   *
   * <japanese>
   * ŗ^ꂽA"#& " Ŏn܂s͂āAΉĒ`
   * ꂽ OhInputFile IuWFNgɑ΂Ďs܂B
   * ǂ̂悤ȑ삪s邩́Ã\bh̎Ɉˑ
   * ܂BOHyMosJ ̃vO~O[ÚAOhInputFile p
   * TuNXŁÃ\bhĒ`āÃTuNX
   * Ǝ̑`邱Ƃł܂B<p>
   *
   * (Ref "OhInputFile#parseSpecLine()") ł́Af[^Ԍ^z
   * ǂݍޑ삾`Ă܂B
   * </japanese>
   *
   * @param aSpecLine a <code>String</code> value
   * @exception OhError if an error occurs
   */
  public void parseSpecLine(String aSpecLine) throws OhError {
    // aSpecLine sample
    // #& dctvec 2 1 0
    Pattern p = Pattern.compile("(\\w+) ");
    Matcher matcher =
      p.matcher(aSpecLine.substring(3, aSpecLine.length()));
    if (matcher.find()) {
      String keyword = matcher.group(1);
      String info =
        aSpecLine.substring(keyword.length() + 3);
      if (keyword.toLowerCase().equals("dctvec")) {
        setDctVec(DataConnectionType.parseDctVec(info));
      }
      else {
        throw new OhError(
          "OhError wad detected." + NEWLINE +
          " File: " + _objectName + "." + NEWLINE +
          " in:   " + aSpecLine + NEWLINE +
          " Unrecognized option: " + keyword + NEWLINE
        );
      }
    }
    else {
      throw new OhError(
        "OhError wad detected." + NEWLINE +
        " File: " + _objectName + "." + NEWLINE +
        " Illegal option line: " + aSpecLine + NEWLINE
      );
    }
  }
  
  // @end {methods} --------------------------------------------

} // OhInputFile
