/**
 * @(#) OhInputFileDouble: Derived from OhInputFile.
 *
 * (Ref "OhObject")
 * <-- (Ref "OhPort")
 * <-- (Ref "OhSenderPort")
 * <-- (Ref "OhSendPort")
 * <-- (Ref "OhInputPort")
 * <-- (Ref "OhInputFile")
 */

package ohymosj;

import java.util.*;
import java.io.*;
import java.util.regex.Pattern;
import java.util.regex.Matcher;

/**
 * The OhInputFileDouble class is a class of input port which
 * reads in datapacks of type double from an input file and
 * sends them. The OhInputFileDouble class extends
 * (Ref "OhInputFile") .<p>
 *
 * OhInputFileDouble can be connected with
 * (Ref "OhReceiveDouble"), (Ref "OhRelayDouble"),
 * (Ref "OhOutputFileDouble").
 *
 * OhInputFileDouble reads an input file using "simple data
 * format for input and output." <p>
 *
 * "Simple data format for input and output" is a format in
 * which
 *
 * <pre>
 *  .....
 *    all the data items for a datapack should be written in one
 *    line, and data items should be separated with a number of
 *    spaces or tabs.
 *
 *    Example: 2003 120 0.3 1.2 2.3 4.5
 * ....
 * </pre>
 *
 * If you need different format, please make a derived class of
 * OhInputFileDouble and redefine "makeDataPack()".
 *
 * <japanese>
 * <p>
 * t@C double ^̃f[^pbNǂݎ͒[q <p>
 * </japanese>
 *
 * @author satoh, shiiba
 * @version $Revision: 1.20 $ $Date: 2004/01/17 07:57:44 $
 */
public class OhInputFileDouble extends OhInputFile
{
  
  /**
   * Multiplication factor to the value which is read in from
   * the input file. For example, assume that _mulfactor is
   * equal to 2.0 and a numerical value 3.2 is read in from the
   * input file, then the data value is set to 6.4.
   *
   */
  protected double _mulfactor = 1.0; //; {
  
  // @begin{initialization} ------------------------------------

  /**
   * Initializ the object. This method assigns the argument
   * aFileName to the instance variable
   * (Ref "OhObject._className").<p>
   *
   * @param aFileName a <code>String</code> value
   * @return a <code>boolean</code> value
   * @exception OhError if an error occurs
   */
  public void init(String aFileName) throws OhError {
    super.init(aFileName);
  }

  // @end{initialization} --------------------------------------

  
  // @begin {constructors} -------------------------------------
  
  /**
   * Creates a new default <code>OhInputFileDouble</code>
   * instance.
   *
   * @exception OhError if an error occurs
   */
  public OhInputFileDouble() {
    super();
    _className = "OhInputFileDouble";
  }


  /**
   * Creates a new <code>OhInputFileDouble</code> instance.
   *
   * @param aFileName a <code>String</code> value
   * @exception OhError if an error occurs
   */
  public OhInputFileDouble(String aFileName) {
    super(aFileName);
    _className = "OhInputFileDouble";
  }

  // @end {constructors} ---------------------------------------

  
  // @begin {methods} ------------------------------------------

  // 1. connection function.

  /**
   * Returns the data pack type which expresses the type of data
   * to be received.
   *
   * @return a <code>String</code> value
   */
  public String dataType() {
    return "double";
  }


  // 2. data input function

  /**
   * Reads string datapacks of type double, creates datapacks of
   * type double, and returns it.
   *
   * <japanese>
   * <p>
   *  aLineData Ɋi[Ăf[^pbN񂩂f[^
   * pbN쐬AObject ^ɃLXgĖ߂܂B<p>
   *
   * ygpz<p>
   * Object dataPack = makeDataPack("2003 12 0.3 3.5");<p>
   *
   * ̗ႾƁAf[^ OhTimePoint(2003, 12, 0.3)AɃf[^
   * l 3.5 f[^pbNǂݎ܂B
   * </japanese>
   *
   * @param aDataStr a <code>String</code> value
   * @return an <code>Object</code> value
   */
  public OhDataPack makeDataPack(String aDataStr) {
    OhDataPackDouble dp = new OhDataPackDouble(aDataStr);
    dp.multiply(_mulfactor);
    return dp;
  }
    
  /**
   * The parseSpecLine method analyzes the line which starts
   * with "#& " and executes the specified operation by the
   * line. This method is implemented in the super class
   * (See: (Ref "OhInputFile#parseSpecLine()") ), but is
   * overrided in this OhInputFileDouble class.<p>
   *
   * In this implementation of the parseSpecLine()
   * method, the following multiplication specification<p>
   *
   * <pre>
   *    #& multiply 2.0
   * </pre>
   *
   * is alse analyzed. This exmaple line specifies the
   * multiplication factor to 2.0, that is, the data which are
   * read in fromt this file are multiplied with 2.0.
   * <p>
   *
   * <japanese>
   * ŗ^ꂽA"#& " Ŏn܂s͂āAΉĒ`
   * ꂽs܂B̃\bh̒`́AX[p[N
   * XŁA^Ă܂ÃNXŃI[o[Ch
   * ܂B(Q: (Ref "OhInputFile#partsSpecLine()") ).
   * <p>
   *
   * (Ref "OhInputFile#parseSpecLine()") Œ`Ăf[^
   * Ԍ^zǂݍޑɉāAf[^l萔{鑀
   * `Ă܂B"#& " ̌ multiply ƏśA
   * ĂāAt@Cǎf[^lɂ̏搔
   * |̂f[^lƂč̗p܂B
   * </japanese>
   *
   * @param aSpecLine a <code>String</code> value
   * @exception OhError if an error occurs
   */
  public void parseSpecLine(String aSpecLine) throws OhError {
    // aSpecLine samples
    // #& dctvec 2 1 0
    // #& multiply 2.0
    Pattern p = Pattern.compile("(\\w+) ");
    Matcher matcher =
      p.matcher(aSpecLine.substring(3, aSpecLine.length()));
    if (matcher.find()) {
      String keyword = matcher.group(1);
      String info =
        aSpecLine.substring(keyword.length() + 3);
      if (keyword.toLowerCase().equals("dctvec")) {
        setDctVec(DataConnectionType.parseDctVec(info));
      }
      else if (keyword.toLowerCase().equals("multiply")) {
        _mulfactor = Double.parseDouble(info);
      }
      else {
        throw new OhError(
          "OhError wad detected." + NEWLINE +
          " File: " + _objectName + "." + NEWLINE +
          " in:   " + aSpecLine + NEWLINE +
          " Unrecognized option: " + keyword + NEWLINE
        );
      }
    }
    else {
      throw new OhError(
        "OhError wad detected." + NEWLINE +
        " File: " + _objectName + "." + NEWLINE +
        " Illegal option line: " + aSpecLine + NEWLINE
      );
    }
  }
  
  /**
   * Stocks the datapack of type double which is stored in the
   * form of string.
   *
   * <japanese>
   * <p>
   *  aStr Ɋi[Ăf[^pbN񂩂f[^pb
   * N쐬A͒[qɎ^܂B<p>
   *
   * \bh̒` (Ref "OhSendPort#dataPackStock()") gp
   * Ă܂B<p>
   * </japanese>
   * @param aDataStr a <code>String</code> value
   */
  public void sendByString(String aDataStr) {
    OhDataPackDouble dp = new OhDataPackDouble(aDataStr);
    dp.multiply(_mulfactor);
    dataPackStock(dp);
  }

  // @end {methods} --------------------------------------------

} // OhInputFileDouble
