/**
 * @(#) OhInputPort; abstract class to express input ports.
 *
 */

package ohymosj;

import java.util.*;
import java.io.*;

/**
 * The OhInputPort class is an abstract class for creating input
 * ports for a total system. It extends (Ref "OhSendPort").<p>
 *
 * This is the base class from which the user creates his own
 * input port models. <p>
 *
 * Do not confuse "OhInputPort" with "OhReceivePort".
 *
 * <japanese>
 * <p>
 * {^͒[qf <p>
 * 
 * Šn̓͒[q쐬邽߂̊NXłB <p>
 * ۃNXłB<p>
 *
 * {^͒[q́AŠn̓͒[q쐬邽߂̒ۃNX
 * łB[ÚÃNXpđŠn̓͒[q̃f
 * 쐬ĂB{^͒[q̃NX錾ł́A[U
 * ׂ͒[q̊{dl߂Ă܂B܂A̎
 * ͎Oɗpӂł@\`Ă܂B<p>
 *
 * Šnf́Å{^͒[q̃NXpă[U
 * 쐬͒[qʂăf[^肷̂ƂĐ݌v
 * ܂B<p>
 *
 * Šn̓ɑ΂ē͒[q̓f[^󂯂Ƃ
 * ƂӖł́A͒[q́A{^vff̎M[qɎ
 * ܂ASʂ̂̂ƍl悢łB<p>
 *
 * Šnf́At@CóAlbg[NɂóAL[
 * {[h́ACRT o͂ȂǁAXGȃf[^̓o͂ɑΉ
 * ΂Ȃ܂BŠnf́AGȃf[^n
 * Ɏ荞ގ́AS (Ref "OhDataPack") ^̃f[^ɕϊ
 * Ȃ΂Ȃ܂B܂AV~[VʂV~[
 * Vs҂ɒ񋟂鎞́A(Ref "OhDataPack") ^Œ~ς
 * f[^V~[Vs҂̖]ރf[^`ɕϊ
 * ΂Ȃ܂B̂悤ȃf[^̕ϊɊւ̓ÍA
 * [U쐬͒[qEo͒[qōsȂƂɂAŠn
 * f̕ł́A͒[qEo͒[q̎葱ő삷邱
 * ɂ܂BȂ킿A͒[qEo͒[q́A̒^Iȏ
 * ƊȌƂ̒邱ƂɂȂ܂B
 *
 * </japanese>
 *
 * @author satoh, shiiba
 * @version $Revision: 1.24 $ $Date: 2004/01/16 00:02:31 $
 */
public abstract class OhInputPort extends OhSendPort
{
  // @begin {abstract methods} ---------------------------------

  // 1. Connection function.
  
  /**
   * Returns the string which shows data type treated by the
   * current input-port. Because this is pure virtual, the
   * function must be defined in derived classes from this
   * class.
   *
   * <dl> <dt>Example
   * <dd> public String dataType() { return "double"; }
   * </dl>
   *
   * We call the string returned by this funcion "dataType".
   *
   * <japanese>
   * <p>
   * ̓f[^pbN^Ԃ
   * <p>
   *
   * @return a <code>String</code> value
   */
  public abstract String dataType();

  // 2. Input data sending function.

  /**
   * Sending input data. <br>
   *
   * With each input terminal, Please create a data pack from
   * input data, and define the function which sends a data
   * pack.
   *
   * <japanese>
   * <p>
   * ̓f[^͒[qɑM <p>
   *
   * e͒[qŁA̓f[^f[^pbN쐬A
   * f[^pbN𑗐M郁\bh`ĉB
   * </japanese>
   *
   */
  public abstract void inputData() throws OhError;

  // @end {abstract methods} -----------------------------------


  // @begin{initialization} ------------------------------------

  /**
   * Initialization.
   *
   * @param aPortName a <code>String</code> value
   * @param aPortNum an <code>int</code> value
   * @exception OhError if an error occurs
   */
  public void init(String aPortName, int aPortNum)
    throws OhError {
    super.init(aPortName, aPortNum);
  }

  // @end{initialization} --------------------------------------

  
  // @begin {constructors} -------------------------------------
  
  /**
   * Creates a new default <code>OhInputPort</code> instance.
   *
   * @exception OhError if an error occurs
   */
  public OhInputPort() {
    super();
    _className = "OhInputPort";
  }


  /**
   * Creates a new <code>OhInputPort</code> instance.
   *
   * @param aPortName a <code>String</code> value
   * @param aPortNum an <code>int</code> value
   */
  public OhInputPort(String aPortName, int aPortNum) {
    super();
    _className = "OhInputPort";
    _objectName = aPortName;
    _objectNumber = aPortNum;
  }

  // @end {constructors} ---------------------------------------


  // @begin {methods} ------------------------------------------

  // input data sending function
  
  /**
   * Calculation request message.  Since there is no calculation
   * function in the OhInputPort class, even if a calculation
   * request is received, it cannot be carried out.  So the
   * OhInputPort object throws the OhCannotGetDataError object
   * when it receives the calReq() message.
   *
   * <japanese>
   * <p>
   * t@C͌vZ@\Ȃ̂ŁAvZvĂǂ傤
   * ܂B
   * </japanese>
   *
   * @exception OhCannotGetDataError if an error occurs
   * @exception OhError if an error occurs
   */
  public void calReq() throws OhCannotGetDataError, OhError{
    throw new OhCannotGetDataError("");
  }
  
  // @end {methods} --------------------------------------------

} // OhInputPort

