/**
 * @(#) OhObject; class to express an Ohymos object
 */

package ohymosj;

import java.io.*;
import java.util.regex.*;

/**
 * Base class of all ohymos classes. <p>
 *
 * An element of type Object has four basic data members:
 * _owner, _className, _objectName, _objectNumber.
 *
 * <dl>
 * <li> Definition of <b>_owner</b>.
 * <dd> When an object A has an object B as its data member,
 *      we call A is the owner of B. <br>
 *      The data member <b>owner</b>
 *      is the referrence to the owner of this object.
 *
 * <li> Definition of <b>_className</b>
 * <dd> _className stores the name of the class. <br>
 *      For example, <b>_className</b> of an object of the class
 *      Object is <b>"OhObject"</b>.
 *
 * <li> Definition of <b>_objectName</b>
 * <dd> _objectName stores the name of the objcet. <br>
 *      User descides the name of each object.
 *
 * <li> Definition of <b>_objectNumber</b>
 * <dd> _objectNumber stores the identification number.
 *
 * </dl>
 *
 * _objectName and _objectNumber are used to identify ohymos
 * elements owned by an ohymos element. Suppose an ohymos
 * element A owns ohymos elements e1, e2, ..., en. Then elements
 * e1, e2, ..., en should be able to be identified by the pair
 * _objectName and _objectNumber.
 *
 * <japanese>
 * <p> {^f̃NX <p>
 *
 * NX OhObject ł́A
 *
 * <pre>
 *  EIuWFNg̖O    : _className, _objectName, _objectNumber
 *  EIuWFNg̏L  : _owner
 * </pre>
 *
 * LA
 *
 * <pre>
 *  EIuWFNg̔r@\
 *  EL҂̐ݒ@\
 *  EIuWFNg̏o͋@\
 * </pre>
 *
 * `܂B<p>
 *
 * OHyMoSJ ̃NX͑SĊ{^f̃NX OhObject p
 * č쐬܂B<p>
 *
 * OHyMoSJ ̃[ÚAʂ̓NX OhObject pč쐬
 * 
 * OhElement  OhSendPort ȂǁA
 * NX OhObject ̉ʂ̃NXpă[UKvƂN
 * X쐬Ώ\łāANX OhObject 𒼐ڌp
 * NX쐬邱Ƃ͂Ȃł傤B<p>
 *
 * NX OhObject ́Aoϐ
 *
 * <pre>
 *	String _className;   //; NX
 *	String _objectName;  //; IuWFNg
 *	int _objectNumber;   //; IuWFNgԍ
 * 	OhObject _owner;     //; OhObject ^ɃLXgꂽL
 * </pre>
 *
 * Ă܂B<p>
 *
 * OHyMoSJ ̃NX͑SĊ{^ OhObject pč쐬̂ŁA
 * OHyMoSJ ̑SẴCX^X̃oϐĂ
 * ƂɂȂ܂B<p>
 *
 * _className ́AehNXƂɋʂ̕ɐݒ肵ĉB
 * Ⴆ΁AhNX ABC ł̓RXgN^
 *
 * <pre>
 *  _className = "ABC";
 * </pre>
 *
 * Ɛݒ肵܂B<p>
 *
 * _objectName, _objectNumber ɂ́ÃIuWFNgg̖OƔԍ
 * (Ref "OhObject#OhObject(String, int)")
 * ܂ (Ref "OhObject#init(String, int)") Őݒ肵܂B
 * dݒ͂ł܂B<p>
 *
 * vO}I _objectName, _objectNumber ݒ肵ĂȂƂ́A
 * _objectName  nullA_objectNumber  -1 ɐݒ肳Ă܂B<p>
 *
 * yȒPȗz
 *
 * <pre>
 *  public class Yahoo extends OhObject
 *  {
 *    public Yahoo() {
 *      super();
 *      _className = "Yahoo";
 *    }
 *
 *    public Yahoo(String aObjName, int aObjNum) {
 *      super();
 *      _className = "Yahoo";
 *      init(aObjName, aObjNum);
 *    }
 *    .
 *    .
 *    .
 *  }
 * </pre>
 *
 * _owner ɂ́AL҂ (Ref "OhObject#setOwner(Object)") 
 * ݒ肵܂B
 * IuWFNg`IuWFNgãoϐłꍇAua
 * `̏LҁvƌĂт܂BႦ΁A
 *
 * <pre>
 *  EM[q`vffãoϐłꍇB
 *  Evff`nfãoϐłꍇB
 * </pre>
 *
 * ̃oϐpāA
 *
 * <ul>
 *  <li> IuWFNg̔r@\
 *  <li> L҂̐ݒ@\
 *  <li> IuWFNg̏o͋@\
 * </ul>
 *
 * Ȃǂ`Ă܂B
 *
 * </japanese>
 *
 * @author satoh, shiiba
 * @version $Revision: 1.28 $ $Date: 2004/01/18 14:00:20 $
 */
public class OhObject implements OhConst
{
  // @begin{field} ---------------------------------------------

  /**
   * The class name.
   *
   * <japanese>
   * <p>
   * NXB
   * </japanese>
   *
   */
  protected String _className; //; final ɂ邱Ƃ͂ł܂B
  
  /**
   * The object name.
   *
   * <japanese>
   * <p>
   * IuWFNg
   * </japanese>
   *
   */
  protected String _objectName;
  
  /**
   * The object number.
   *
   * <japanese>
   * <p>
   * IuWFNgԍ
   * </japanese>
   *
   */
  protected int _objectNumber;
  
  /**
   * The owner of the current object.
   *
   * <japanese>
   * <p>
   * L
   * </japanese>
   *
   */
  protected OhObject _owner;

  // @end{field} -----------------------------------------------


  // @begin{initialization} ------------------------------------

  /**
   * Initialization.
   *
   * <japanese>
   * <p>
   * IuWFNgAIuWFNgԍ̏ <p>
   *
   * IuWFNgAIuWFNgԍ܂B̓Iɂ́A
   *  aObjName, aObjNum  _objectName, _objectNumber ɐݒ
   * ܂B<p>
   *
   * ܂A_objectName ݒς݂̎́AdݒƔfAG[
   * bZ[W\ false Ԃ܂B̏ꍇA_objectName,
   * _objectNumber ύX܂B<p>
   *
   * RXgN^̒`AyуftHgRXgN^Ő
   * IuWFNg̏ɗp܂B
   * </japanese>
   *
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @exception OhError if an error occurs
   */
  public void init(String aObjName, int aObjNum) throws OhError {
    // System.out.println(aObjName);
    if (aObjName == null) {
      throw new IllegalArgumentException(
        oPrintString("\t"));
    }

    if (_objectName != null) {
      throw new OhError(
        "Error in OhObject.init(): " + 
          "Tried to set objectName twice." + NEWLINE +
        "\tYou called init(\"" + aObjName +
                "\"), " + aObjNum + ")" + NEWLINE +
        "\tfor the OhObject instance:" + NEWLINE +
        oPrintString("\t\t")
      );
    }

    _objectName   = aObjName;
    _objectNumber = aObjNum;
  }

  // @end{initialization} --------------------------------------


  // @begin{constructors} --------------------------------------
  
  /**
   * Creates a new default <code>OhObject</code> instance.
   *
   */
  public OhObject() {
    _className    = "OhObject";
    _objectName   = null;
    _objectNumber = -1;
    _owner        = null;
  }


  /**
   * Creates a new <code>OhObject</code> instance.
   *
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   */
  public OhObject(String aObjName, int aObjNum) {
    _className    = "OhObject";
    _objectName   = aObjName;
    _objectNumber = aObjNum;
    _owner        = null;
  }

  // @end{constructors} ----------------------------------------


  // @begin{methods} -------------------------------------------
  
  /**
   * Gets the class name.
   *
   * @return a <code>String</code> value
   */
  public String getClassName() {
    return _className;
  }

  /**
   * Gets the object name.
   *
   * @return a <code>String</code> value
   */
  public String getObjectName() {
    return _objectName;
  }

  /**
   * Gets the object number.
   *
   * @return an <code>int</code> value
   */
  public int getObjectNumber() {
    return _objectNumber;
  }

  /**
   * Gets the owner of the current object casted to type
   * <code>OhObject</code>.
   *
   * @return an <code>OhObject</code> value
   */
  public OhObject getOwner() {
    return _owner;
  }

  /**
   * Gets the top owner of the receiver.
   *
   * @return an <code>OhObject</code> value
   */
  public OhObject getTopOwner() {  //; (*5)
    OhObject top    = this;
    OhObject parent = _owner;

    while (parent != null) {
      top = parent;
      parent = top.getOwner();
    }
    
    return top;
  }

  // 1. method about a name

  /**
   * If the current object has the same triplet as those given
   * by the arguments, this method returns true. Otherwise,
   * false.
   *
   * <japanese>
   * <p>
   * NXAIuWFNgAIuWFNgԍɊւIuWF
   * Ng̔r <p>
   *
   * ̃\bh́A_className, _objectName, _objectNumber 
   * aClassName, aObjName, aObjNum Ɉvꍇ true Ԃ܂B
   * ǂꂩPłvȂ̂ false Ԃ܂B
   * </japanese>
   *
   * @param aClassName a <code>String</code> value
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @return a <code>boolean</code> value
   */
  public boolean objCompare(String aClassName, String aObjName, int aObjNum) {
    if ((_objectName.equals(aObjName))
      && (_objectNumber == aObjNum)
      && (_className.equals(aClassName)))
      return true;
    
    return false;
  }

  
  /**
   * If the current object has the same object name and object
   * number as those given as the arguments, this method returns
   * true. Otherwise, false.
   *
   * <japanese>
   * <p>
   * IuWFNgAIuWFNgԍɊւIuWFNg̔r <p>
   *
   * ̃\bh́A_objectName, _objectNumber 
   * aObjName, aObjNum Ɉvꍇ true Ԃ܂B
   * ꂩPłvȂ̂ false Ԃ܂B
   * </japanese>   
   *
   * @param aObjName a <code>String</code> value
   * @param aObjNum an <code>int</code> value
   * @return a <code>boolean</code> value
   */
  public boolean objCompare(String aObjName, int aObjNum) {
    if ((_objectName.equals(aObjName))
      && (_objectNumber == aObjNum))
      return true;
    
    return false;
  }


  /**
   * If the current object has the same class name, the object name, 
   * the object number, and the owner as those which aObj has,
   * this method returns true. Otherwise, false.
   *
   * <japanese>
   * <p>
   * NXAIuWFNgAIuWFNgԍAL҂Ɋւ
   * IuWFNg̔r <p>
   *
   * ̃\bh́A_className, _objectName, _objectNumber,
   * _owner ׂ aObj ƈvꍇ true Ԃ܂Bǂꂩ
   * PłvȂ̂ false Ԃ܂B܂AaObj 
   *  null ̏ꍇ false Ԃ܂B
   * </japanese>
   *
   * @param aObj an <code>OhObject</code> value
   * @return a <code>boolean</code> value
   */
  public boolean objCompare(OhObject aObj) {
    if (aObj == null)
      return false;
    
    if ((aObj.getObjectName().equals(_objectName))
      && (aObj.getOwner() == _owner)
      && (aObj.getObjectNumber() == _objectNumber)
      && (aObj.getClassName().equals(_className)))
      return true;
    
    return false;
  }


  // 2. method to set up an owner.

  /**
   * Set the owner of this object to the object specified. <p>
   *
   * When already a different object has been registerd as the
   * owner, you cannot set the new owner. If this case happens,
   * this method throws OhError exception.
   *
   * <japanese>
   * <p>
   * L҂̐ݒ <p>
   *
   * aOwnerijL҂ɐݒ肵܂B<p>
   *
   * aOwner Ƃ͈قȂIuWFNgłɏL҂ƂĐݒ肳
   * ꍇ́AV aOwner L҂ɐݒ肹ɁAO 
   * OhError ܂B
   * </japanese>
   *
   * @param aOwner an <code>Object</code> value
   * @exception OhError if an error occurs
   */
  public void setOwner(OhObject aOwner) throws OhError {
    if (aOwner == null) {
      throw new IllegalArgumentException(
        oPrintString("\t")
      );
    }
    
    if (_owner != null && _owner != aOwner) {
      StringWriter buffer = new StringWriter();
      PrintWriter  out    = new PrintWriter(buffer);
      _owner.oPrint(out, "    first : ");
      aOwner.oPrint(out, "    next  : ");
      cPrint(out, "    ");
      String str = buffer.toString();
      out.close();
      throw new OhError(
        "\"owner\" has already been set." + NEWLINE +
        str);
    }

    _owner = aOwner;
  }


  // 3. Information output method
  
  /**
   * Returns a string which specifies the object uniquely. <p>
   *
   * The string which Absname() returns will be created by
   * concatenating
   *
   * <pre>
   *
   *   "absolute name of the owner"
   *   "/"
   *   _objectName
   *   "__"
   *   _objectNumber.
   *
   * </pre>
   *
   * When _owner == NULL, "absolute name of the owner" is taken
   * as "". <p>
   *
   * <EXAMPLE> <p>
   *
   * Assume that the absolute name of the owner is
   * "/total_sytetem__0", _objectName of the current object is
   * "elm_t", and _objectNumber is 0, then "absname" is
   *
   * <pre>
   *
   *   /total_system__0/elm_t__0
   *
   * </pre>
   *
   * @return a <code>String</code> value
   */
  public String getAbsName() {                 //; (*2)
    StringBuffer absName = new StringBuffer("");
    
    if (_owner == null) {
      absName = absName.append("");
    }
    else {
      absName = absName.append(_owner.getAbsName());
//;   ((OhObject)_owner).getAbsName();  owner ̏ʂɂȂB
    }

    absName = absName.append("/");
    absName = absName.append(_objectName);
    absName = absName.append("__");
    absName = absName.append((new Integer(_objectNumber)).toString());

    return absName.toString();    
  }

  /**
   * Print "aHead", _className, _objectName, _objectNumer to
   * aPw. <p>
   *
   * <EXAMPLE>
   *
   * <pre>
   * |  Yahoo a("yahoo", 0);
   * |  a.cPrint(aPw, "+");
   *
   * prints to aPw
   *
   *    +Yahoo yahoo 0
   * </pre>
   *
   * @param aPw a <code>PrintWriter</code> value
   * @param aHead a <code>String</code> value
   */
  public void cPrint(PrintWriter aPw, String aHead) {
    aPw.print(aHead + _className + " " +
      _objectName + " " + _objectNumber + NEWLINE);
  }

  /**
   * Returns the String aHead + _className + " " +
   * _objectName + " " + _objectNumer + NEWLINE<p>
   *
   * <EXAMPLE>
   *
   * <pre>
   * |  Yahoo a("yahoo", 0);
   * |  a.cPrintString(aPw, "+");
   *
   * returns 
   *
   *    +Yahoo yahoo 0
   * </pre>
   *
   * @param aHead a <code>String</code> value
   */
  public String cPrintString(String aHead) {
    String str =  aHead + _className + " " +
      _objectName + " " + _objectNumber + NEWLINE;
    return str;
  }

  /**
   * Trace back the object's ancestry, that is, prints all the
   * ancesters to aPw as far as an ancester exists. "aHead" is
   * printed before the ancestry. <p>
   *
   * <EXAMPLE>
   *
   * <pre>
   * |  Yahoo a("yahoo", 0);
   * |  Yahoo b("yahoobb", 1);
   * |  b.setOwner(a);
   * |  PrintWriter aPw;
   * |  b.Oprint(aPw, "+ ");
   * 
   * prints to aPw as follows
   * 
   * |  + Yahoo yahoobb 1
   * |  +     owner > Yahoo yahoo 0
   * </pre>
   *
   * <japanese>
   * Y OhObject ȀL֌Wɏ]ďL҂̕ɑk
   * āÃNXAIuWFNgAIuWFNgԍ̌n
   * w肳ꂽ PrintWriter ɏo͂܂Bes̍sɁAaHead 
   * w肳ꂽu܂BŌ̍sŉsĂ邱Ƃɒ
   * ӁB
   * </japanese>
   *
   * @param aPw a <code>PrintWriter</code> value
   * @param aHead a <code>String</code> value
   */
  public void oPrint(
    PrintWriter aPw,
    String aHead) {  //; (*4)
    aPw.println(aHead + "(" + NEWLINE + aHead + " " +
      objectInfo());
    OhObject obj;
    for(obj = _owner; obj != null; obj = obj.getOwner()) {
      aPw.println(aHead + " " + obj.objectInfo());
    }
    aPw.println(aHead + ")");
  }

  /**
   * Returns a message string which contains the information
   * about the OhObject object itself ant its owner sequence.<p>
   *
   * <japanese>
   * Y OhObject Ƃ̏Ln ( OhObject  owner k
   * ĂĂłn) ̏𕶎 (s) ɂĕԂ܂B
   * aHead ́A̕s̊̕es̑Oɕt镶łB
   * </japanese>
   *
   * @param aHead a <code>String</code> value
   * @return a <code>String</code> value
   */
  public String oPrintString(String aHead) {
     StringWriter buffer = new StringWriter();
     PrintWriter  out    = new PrintWriter(buffer);
     oPrint(out, aHead);
     String str = buffer.toString();
     out.close();
     return str;
  }

  /**
   * Returns a String which contains the information about of
   * _className, _objectName, and _objectNumber.
   *
   * @return a <code>String</code> value
   */
  public String objectInfo() {
    return _className + " " + _objectName + " " + _objectNumber;
  }
  
  /**
   * Returns a String representation of this object.
   *
   * <japanese>
   * <p>
   * YIuWFNgoA_owner Ɋi[ĂIuWF
   * NgHA_owner  null ɂȂ܂ł̃IuWFNǧn
   * āAeIuWFNg _ className, _objectName,
   * _objectNumber ̏sɏs̕Ԃ܂B
   * <p>
   *
   * (Ql) OhObject obj  toString() ɑ΂ĕԂs
   * ̊es̍sɃ^uɂ́A<br>
   *
   * <pre>
   *   Pattern p = Pattern.compile("^", Pattern.MULTILINE);
   *   Matcher matcher = p.matcher(obj.toString());
   *   String output = matcher.replaceAll("\t");
   * </pre>
   *
   * ̂悤ɂ܂Boutput ړI̕łB
   * </japanese>
   *
   * @return a <code>String</code> value
   */
  public String toString() {
    StringBuffer str = new StringBuffer();
    str.append("(" + NEWLINE + " " + objectInfo() + NEWLINE);
    OhObject obj;
    for (obj = _owner; obj != null; obj = obj.getOwner()) {
      str.append(" " + obj.objectInfo() + NEWLINE);
    }
    str.append(")" + NEWLINE);
    return str.toString();
  }

  
  /**
   * Returns the information string about this class.<p>
   *
   * For ports, this method returns the information about
   *   classname;
   *   datatype it carries.
   *
   * For element, this method returns the information about
   *   classname;
   *   list of port name, port numbers
   *   list of parameters
   *   formats which are used for reading parameter values
   *   list of state variables
   *   formats which are used for reading iniitial states.
   *
   * @return a <code>String</code> value
   */
  public String basicInfo() {
    return
      "No basic info is given for " + _className + ".";
  }

  // @end{methods} ---------------------------------------------

} // OhObject
