/**
 * @(#) total system created from a structure configuration file.
 */

package ohymosj;

import ohymosj.*;
import java.util.*;
import java.io.*;

/**
 * The OhScfTotalSystem class is a total system class which uses
 * a structure configuration file.<p>
 *
 * <japanese>
 * \`t@Cgpč\Šn\NXłB
 * <p>
 *
 * ̃NX́A[q\vf̍\t@C "Structure
 * Configuration File" ǂ݂Ƃč肠ŠñNX
 * B
 * </japanese>
 * <p>
 *
 * @author satoh, shiiba
 * @version $Revision: 1.12 $ $Date: 2004/02/14 08:32:57 $
 */
public class OhScfTotalSystem extends OhTotalSystem
{
  // @begin {instance variable or static variable} -------------

  /**
   * The array where relay ports for receiveing data are stored.
   *
   */
  protected OhPort[] _receivePorts;

  /**
   * Number of relay ports for receiving data.
   *
   */
  protected int _numOfReceivePorts;

  /**
   * The array where relay ports for sending data are stored.
   *
   */
  protected OhPort[] _sendPorts;

  /**
   * Number of relay ports for sending data.
   *
   */
  protected int _numOfSendPorts;

  /**
   * Usual(Non-iterate) Components.
   *
   */
  protected OhElement[] _elements;

  /**
   * Number of usual components.
   *
   */
  protected int _numOfElements;

  /**
   * Iterative Components.
   *
   */
  protected OhElement[] _iterators;

  /**
   * Number of iterative components.
   *
   */
  protected int _numOfIterators;

  /**
   * "Structure Configuration File" Name.
   *
   */
  protected String _scfFile;

  // @end {instance variable or static variable} ---------------


  // @begin {constructors} -------------------------------------

  /**
   * <japanese>
   * ftHgRXgN^łB
   * </japanese>
   *
   * @param aFileName a <code>String</code> value
   * @exception OhError if an error occurs
   */
  public OhScfTotalSystem(String aFileName) throws OhError {
    super();
    _receivePorts = null;
    _sendPorts    = null;
    _elements     = null;
    _iterators    = null;
    _scfFile      = null;
    _className    = "OhScfTotalSystem";

    if (_objectName != null) {
      throw new OhError(
        "\t\"OhScfTotalSystem\", aObjNum(arg) = 0" + NEWLINE +
        oPrintString("\t")
      );
    }

    _objectName   = "totalSystem";
    _objectNumber = 0;

    // Read SCF name from "aBr".
    //
    _scfFile = aFileName;

    // Open the SCF. Comment strings are removed.
    //
    FileReader       fr = null;
    OhBufferedReader br = null;

    try {
      fr = new FileReader(_scfFile);
      br = new OhBufferedReader(fr, true);

      // Read Numbers part.
      //
      
      br.skipWhites();
      _numOfReceivePorts = Integer.parseInt(br.readWord());
      if (_numOfReceivePorts > 0) {
        _receivePorts = new OhPort[_numOfReceivePorts];
      }
      
      br.skipWhites();
      _numOfElements = Integer.parseInt(br.readWord());
      if (_numOfElements > 0) {
        _elements = new OhElement[_numOfElements];
      }
      
      br.skipWhites();
      _numOfIterators = Integer.parseInt(br.readWord());
      if (_numOfIterators > 0) {
        _iterators = new OhElement[_numOfIterators];
      }

      br.skipWhites();
      _numberOfIteratorSets = Integer.parseInt(br.readWord());
      
      br.skipWhites();
      _numOfSendPorts = Integer.parseInt(br.readWord());
      if (_numOfSendPorts > 0) {
        _sendPorts = new OhPort[_numOfSendPorts];
      }

      // System.out.println("complete read numbers part.");
        
      // Read Memory allocation part.
      //
      
      String header;
      String objectName;
      String className;
      int    objectNumber;

      // Read [InputPorts] section.
      //
      
      br.skipWhites();
      header = br.readWord();
      
      if (header.equals("[InputPorts]")) {
        int id;
        for (int i = 0; i < _numOfReceivePorts; i++) {
          br.skipWhites();
          id = Integer.parseInt(br.readWord());
          br.skipWhites();
          objectName = br.readWord();
          br.skipWhites();
          className = br.readWord();

          if (id != i) {
            StringWriter buffer = new StringWriter();
            PrintWriter  out    = new PrintWriter(buffer);
            out.print("SCF Total data error in [InputPorts] section. The id number " + i
              + " was expected, but it was " + id + "." + NEWLINE);
            String str = buffer.toString();
            out.close();
            throw new OhError(str);
          }
          _receivePorts[i] = OhPort.create(className);
          _receivePorts[i].init(objectName, 0);
        }
      }
      else {
        String str =
          "SCF Total data error: no [InputPorts] section." +
          NEWLINE;
        throw new OhError(str);
      }
      
      // System.out.println("complete inputports part.");
      
      // Read [Components] section.
      //
      
      br.skipWhites();
      header = br.readWord();
      
      if (header.equals("[Components]")) {
        for (int i = 0; i < _numOfElements; i++) {
          int id;
          br.skipWhites();
          id = Integer.parseInt(br.readWord());
          br.skipWhites();
          objectName = br.readWord();
          br.skipWhites();
          objectNumber = Integer.parseInt(br.readWord());
          br.skipWhites();
          className = br.readWord();

          if (id != i) {
            StringWriter buffer = new StringWriter();
            PrintWriter  out    = new PrintWriter(buffer);
            out.print(
              "SCF Total data error in [Components] section. " +
              "The id number " + i +
              " was expected, but it was " + id + "." + NEWLINE);
            String str = buffer.toString();
            out.close();
            throw new OhError(str);
          }
          _elements[id] = OhElement.create(className);
          _elements[id].init(objectName, objectNumber, br);
        }
      }
      else {
        String str = "SCF Total data error: " +
          "no [Components] section." + NEWLINE;
        throw new OhError(str);
      }
      
      // System.out.println("complete components part.");
      
      // Read [Iterators] section.
      //
      
      br.skipWhites();
      header = br.readWord();
      
      if (header.equals("[Iterators]")) {
        for (int i = 0; i < _numOfIterators; i++) {
          int id;
          br.skipWhites();
          id = Integer.parseInt(br.readWord());
          br.skipWhites();
          objectName = br.readWord();
          br.skipWhites();
          objectNumber = Integer.parseInt(br.readWord());
          br.skipWhites();
          className = br.readWord();

          if (id != i) {
            StringWriter buffer = new StringWriter();
            PrintWriter  out    = new PrintWriter(buffer);
            out.print("SCF Total data error in [Iterators] " +
              "section. The id number " + i +
              " was expected, but it was " + id + "." + NEWLINE);
            String str = buffer.toString();
            out.close();
            throw new OhError(str);
          }
          _iterators[id] = OhElement.create(className);
          _iterators[id].init(objectName, objectNumber, br);
        }
      }
      else {
        String str = "SCF Total data error: " +
          "no [Iterators] section." + NEWLINE;

        throw new OhError(str);
      }
      
      // System.out.println("complete iterators part.");
      
      // Read [OutputPorts] section.
      //
      
      br.skipWhites();
      header = br.readWord();
      
      if (header.equals("[OutputPorts]")) {
        for (int i = 0; i < _numOfSendPorts; i++) {
          int id;
          int lineLength;
          br.skipWhites();
          id = Integer.parseInt(br.readWord());
          br.skipWhites();
          objectName = br.readWord();
          br.skipWhites();
          lineLength = Integer.parseInt(br.readWord());
          br.skipWhites();
          className = br.readWord();

          if (id != i) {
            StringWriter buffer = new StringWriter();
            PrintWriter  out    = new PrintWriter(buffer);
            out.print("SCF Total data error in [SendPorts] " +
              "section. The id number " + i +
              " was expected, but it was " + id + "." + NEWLINE);
            String str = buffer.toString();
            out.close();
            throw new OhError(str);
          }
          _sendPorts[i] = OhPort.create(className);
          // System.out.println("className: " + className);
          _sendPorts[i].init(objectName, lineLength);
          // System.out.println("end of init: " + objectName);
        }
      }
      else {
        String str = "SCF Total data error: " +
          "no [SendPorts] section." + NEWLINE;
        throw new OhError(str);
      }
      
      // System.out.println("complete outputports part.");
      
    }
    catch (FileNotFoundException e) {
      // System.out.println("File not found: " + _scfFile);
      // e.printStackTrace();
      throw new OhError(
        "Fine not found: " + _scfFile
      );
    }
    catch (OhError e) {
      throw new OhError(
        "SCF Total data error in " + aFileName + NEWLINE +
        e.what());
    }
    catch (Exception e) {
      throw new OhError(
        "Error when reading " + _scfFile + NEWLINE +
        OhError.getStackTraceString(e));
    }
    finally {
      if (br != null) {
        try { br.close(); } catch (Exception e) {}
      }
    }
  }
  
  // @end {constructors} ---------------------------------------

  // @begin {methods} ------------------------------------------

  /**
   * <japanese>
   * Mp[qo^܂B
   * (Ref "OhSubSystem.construct()")
   * Ă΂܂B
   * o^ĂȂMp[q͐ڑł܂B
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public void registerInputPorts() throws OhError {  //; (*7)
    for (int i = 0; i < _numOfReceivePorts; i++) {
      register(_receivePorts[i]);
    }
  }


  /**
   * <japanese>
   * OhScfTotalSystem.registerOutputPorts()
   * s܂B
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public void registerReceivePorts() throws OhError {  //; (*1)
    registerOutputPorts();
  }



  /**
   * <japanese>
   * Mp[qo^܂B
   * OhSubSystem.construct()
   * Ă΂܂B
   * o^ĂȂMp[q͐ڑł܂B
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public void registerOutputPorts() throws OhError {  //; (*8)
    for (int i = 0; i < _numOfSendPorts; i++) {
      register(_sendPorts[i]);
    }
  }


  /**
   * <japanese>
   * OhScfTotalSystem.registerInputPorts()
   * s܂B
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public void registerSendPorts() throws OhError {  //; (*2)
    registerInputPorts();
  }


  /**
   * <japanese>
   * S\vf̂AvZsȂȂ\IuWFNg̃AhX _componentArray ɓo^܂B<p>
   *
   * o^ĂȂ\IuWFNgɂ́AvZs߂𑗂邱Ƃł܂B<p>
   *
   * Ⴆ΁AhNXō\IuWFNg abc oϐƂĒ`ꍇA
   * ̂悤 abc o^܂B
   *
   * <pre>
   *   public void registerComponents() {
   *     register((Object)abc);
   *   }
   * </pre>
   *
   * ̎AK abc  Object ^ɃLXgĉB<p>
   *
   * ɂ abc ́AꎞI _objectList Ɏ^ꂽA
   * ŏII Object ^z _componentArray ɋL܂B
   * iOhSubSystem.construct() QƁj
   *
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public void registerComponents() throws OhError {  //; (*3)
    for (int i = 0; i < _numOfElements; i++) {
      register(_elements[i]);
    }
  }


  /**
   * <japanese>
   * ̃\bh́AS\vf̂AvZsȂvf̃AhX _iteratorArray ɓo^܂B
   * vZsȂnfVɍ쐬vO}Ã\bh`Ȃ΂Ȃ܂B<p>
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public void registerIterators() throws OhError {  //; (*4)
    if (_numberOfIteratorSets <= 0)
      return;

    // Open the SCF. Comment strings are removed.
    //
    FileReader       fr = null;
    OhBufferedReader br = null;

    try {
      fr = new FileReader(_scfFile);
      br = new OhBufferedReader(fr, true);
    }
    catch (FileNotFoundException e) {
      e.printStackTrace();
    }

    String  strLine = null;
    boolean foundP  = false;

    try {
      br.skipWhites();
      while ((strLine = br.readWord()) != null) {
        if (strLine.equals("[RegisterIterators]")) {
          foundP = true;
          break;
        }
      }
    }
    catch (IOException e) {
      e.printStackTrace();
    }

    if (foundP == false) {
      throw new OhError("");
    }


    try {
      for (int m = 0; m < _numberOfIteratorSets; m++) {
        int num;
        br.skipWhites();
        num = Integer.parseInt(br.readWord());
        
        //; _iteratorArray[i] == null ƂȂĂ i TB exit B
        //;
        int i;
        boolean flag = false;
        
        for (i = 0; i < _numberOfIteratorSets; i++) {
          if (_iteratorArray[i] == null) {
            flag = true;
            break;
          }
        }
        
        if (flag == false) {
          throw new OhError(oPrintString("    "));
        }
        
        for (int k = 0; k < num; k++) {
          int s;
          br.skipWhites();
          s = Integer.parseInt(br.readWord());
          register(_iterators[s]);
        }
        
        int[] numberOfIteratorsArray = {_numberOfIterators[i]};
        
        _iteratorArray[i]     = makeObjectArray(numberOfIteratorsArray);    //; OhElement (*4)
        _numberOfIterators[i] = numberOfIteratorsArray[0];
        
        _objectList.clear();
      }
    }
    catch (IOException e) {
      e.printStackTrace();
    }
  }


  /**
   * <japanese>
   * nf̒[q̐ڑ
   * OhSubSystem.connect()
   * pĒ`܂B
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public void connectPorts() throws OhError {  //; (*5)
    // Open the SCF. Comment strings are removed.
    //
    FileReader       fr = null;
    OhBufferedReader br = null;

    try {
      
      fr = new FileReader(_scfFile);
      br = new OhBufferedReader(fr, true);

      String  strLine = null;
      boolean foundP  = false;

      br.skipWhites();
      while ((strLine = br.readWord()) != null) {
        if (strLine.equals("[ConnectPorts]")) {
          foundP = true;
          break;
        }
      }

      if (foundP == false) {
        throw new OhError("No [ConnectPorts] section.");
      }

      String objNameS = null;
      String objNameR = null;
      int objNumS = 0;
      int objNumR = 0;

      while (true) {  // ----------- starts while(true) loop.
        String str;
        br.skipWhites();
        str = br.readWord();

        if (str.charAt(0) == 'z') {
          break;
        }

        // -- There are nine cases. ---

        if (str.charAt(0) == 'r') {    // The sender is a receive port.
          int i;
          br.skipWhites();
          i = Integer.parseInt(br.readWord());    // Receive port ID.

          if (i >= _numOfReceivePorts) {
            StringWriter buffer = new StringWriter();
            PrintWriter  out    = new PrintWriter(buffer);
            out.print(_scfFile + ": Illegal receive port number: r " + i);
            String errMsg = buffer.toString();
            out.close();
            throw new OhError(errMsg);
          }

          // The receiver.
          //
          String strR;
          br.skipWhites();
          strR = br.readWord();

          int j;

          char r = strR.charAt(0);

          switch (r) {

          case 'c':    // The receiver is a usual component.
            br.skipWhites();
            j = Integer.parseInt(br.readWord());
            br.skipWhites();
            objNameR = br.readWord();
            br.skipWhites();
            objNumR = Integer.parseInt(br.readWord());

            if (j >= _numOfElements) {
              StringWriter buffer = new StringWriter();
              PrintWriter  out    = new PrintWriter(buffer);
              out.print(_scfFile + ": Illegal element number: c " + j);
              String errMsg = buffer.toString();
              out.close();
              throw new OhError(errMsg);
            }

            connect(_receivePorts[i], _elements[j], objNameR, objNumR);

            break;

          case 'i':    // The receiver is an iterative component.
            br.skipWhites();
            j = Integer.parseInt(br.readWord());
            br.skipWhites();
            objNameR = br.readWord();
            br.skipWhites();
            objNumR = Integer.parseInt(br.readWord());

            if (j >= _numOfIterators) {
              StringWriter buffer = new StringWriter();
              PrintWriter  out    = new PrintWriter(buffer);
              out.print(_scfFile + ": Illegal iterator element number: i " + j);
              String errMsg = buffer.toString();
              out.close();
              throw new OhError(errMsg);
            }

            connect(_receivePorts[i], _iterators[j], objNameR, objNumR);

            break;

          case 's':    // The receiver is a send port.
            br.skipWhites();
            j = Integer.parseInt(br.readWord());
            
            if (j >= _numOfSendPorts) {
              StringWriter buffer = new StringWriter();
              PrintWriter  out    = new PrintWriter(buffer);
              out.print(_scfFile + ": Illegal output port number: s " + j);
              String errMsg = buffer.toString();
              out.close();
              throw new OhError(errMsg);
            }

            connect(_receivePorts[i], _sendPorts[j]);

            break;

          default:
            throw new OhError("");
          }
        }
        else if (str.charAt(0) == 'c') {    // The sender is a usual component.
          int i;
          br.skipWhites();
          i = Integer.parseInt(br.readWord());
          br.skipWhites();
          objNameS = br.readWord();
          br.skipWhites();
          objNumS = Integer.parseInt(br.readWord());

          if (i >= _numOfElements) {
            StringWriter buffer = new StringWriter();
            PrintWriter  out    = new PrintWriter(buffer);
            out.print(_scfFile + ": Illegal element number: c " + i);
            String errMsg = buffer.toString();
            out.close();
            throw new OhError(errMsg);
          }

          // The receiver.
          //
          String strR;
          br.skipWhites();
          strR = br.readWord();

          int j;

          char r = strR.charAt(0);

          switch (r) {

          case 'c':    // The receiver is a usual component.
            br.skipWhites();
            j = Integer.parseInt(br.readWord());
            br.skipWhites();
            objNameR = br.readWord();
            br.skipWhites();
            objNumR = Integer.parseInt(br.readWord());

            if (j >= _numOfElements) {
              StringWriter buffer = new StringWriter();
              PrintWriter  out    = new PrintWriter(buffer);
              out.print(_scfFile + ": Illegal element number: c " + i);    //; j ̊ԈႢł́HHH
              String errMsg = buffer.toString();
              out.close();
              throw new OhError(errMsg);
            }

            connect(_elements[i], objNameS, objNumS, _elements[j], objNameR, objNumR);

            break;

          case 'i':    // The receiver is an iterative component.
            br.skipWhites();
            j = Integer.parseInt(br.readWord());
            br.skipWhites();
            objNameR = br.readWord();
            br.skipWhites();
            objNumR = Integer.parseInt(br.readWord());

            if (j >= _numOfIterators) {
              StringWriter buffer = new StringWriter();
              PrintWriter  out    = new PrintWriter(buffer);
              out.print(_scfFile + ": Illegal iterator element number: i " + j);
              String errMsg = buffer.toString();
              out.close();
              throw new OhError(errMsg);
            }

            connect(_elements[i], objNameS, objNumS, _iterators[j], objNameR, objNumR);

            break;

          case 's':    // The receiver is a send port.
            br.skipWhites();
            j = Integer.parseInt(br.readWord());
            
            if (j >= _numOfSendPorts) {
              StringWriter buffer = new StringWriter();
              PrintWriter  out    = new PrintWriter(buffer);
              out.print(_scfFile + ": Illegal output port number: s " + j);
              String errMsg = buffer.toString();
              out.close();
              throw new OhError(errMsg);
            }

            connect(_elements[i], objNameS, objNumS, _sendPorts[j]);

            break;

          default:
            throw new OhError("");
          }
        }
        else if (str.charAt(0) == 'i') {    // The sender is an iterative component.
          int i;
          br.skipWhites();
          i = Integer.parseInt(br.readWord());
          br.skipWhites();
          objNameS = br.readWord();
          br.skipWhites();
          objNumS = Integer.parseInt(br.readWord());

          if (i >= _numOfIterators) {
            StringWriter buffer = new StringWriter();
            PrintWriter  out    = new PrintWriter(buffer);
            out.print(_scfFile + ": Illegal iterator element number: i " + i);
            String errMsg = buffer.toString();
            out.close();
            throw new OhError(errMsg);
          }

          // The receiver.
          //
          String strR;
          br.skipWhites();
          strR = br.readWord();

          int j;

          char r = strR.charAt(0);

          switch (r) {

          case 'c':    // The receiver is a usual component.
            br.skipWhites();
            j = Integer.parseInt(br.readWord());
            br.skipWhites();
            objNameR = br.readWord();
            br.skipWhites();
            objNumR = Integer.parseInt(br.readWord());

            if (j >= _numOfElements) {
              StringWriter buffer = new StringWriter();
              PrintWriter  out    = new PrintWriter(buffer);
              out.print(_scfFile + ": Illegal element number: c " + j);
              String errMsg = buffer.toString();
              out.close();
              throw new OhError(errMsg);
            }

            connect(_iterators[i], objNameS, objNumS, _elements[j], objNameR, objNumR);

            break;

          case 'i':    // The receiver is an iterative component.
            br.skipWhites();
            j = Integer.parseInt(br.readWord());
            br.skipWhites();
            objNameR = br.readWord();
            br.skipWhites();
            objNumR = Integer.parseInt(br.readWord());

            if (j >= _numOfIterators) {
              StringWriter buffer = new StringWriter();
              PrintWriter  out    = new PrintWriter(buffer);
              out.print(_scfFile + ": Illegal iterator element number: i " + j);
              String errMsg = buffer.toString();
              out.close();
              throw new OhError(errMsg);
            }

            connect(_iterators[i], objNameS, objNumS, _iterators[j], objNameR, objNumR);

            break;

          case 's':    // The receiver is a send port.
            br.skipWhites();
            j = Integer.parseInt(br.readWord());
            
            if (j >= _numOfSendPorts) {
              StringWriter buffer = new StringWriter();
              PrintWriter  out    = new PrintWriter(buffer);
              out.print(_scfFile + ": Illegal output port number: s " + j);
              String errMsg = buffer.toString();
              out.close();
              throw new OhError(errMsg);
            }

            connect((OhElement)_iterators[i], objNameS, objNumS, (OhPort)_sendPorts[j]);

            break;

          default:
            throw new OhError("");
          }
        }
        else {
          throw new OhError("");
        }
      }  // -------------------------------------------------------------- ends while(true) loop.
    }
    catch (OhError e) {
      throw e;
    }
    catch (FileNotFoundException e) {
      throw new OhError(OhError.getStackTraceString(e));
    }
    catch (IOException e) {
      throw new OhError(OhError.getStackTraceString(e));
    }
    finally {
      if (br != null) {
        try { br.close(); } catch (Exception e) {}
      }
    }
  }


  /**
   * <japanese>
   * nf̒ڒʐMڑ
   * OhElement.register(Object, int)
   * pĒ`܂B
   *
   * <pre>
   * [ConnectComponents]
   * c 2 c 3 0 # mElmnt[2]->Register((void*)mElmnt[3], 0)
   * i 0 i 1 2 # mIte[0]->Register((void*)mIte[1], 2);
   * z # end mark
   * </pre>
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public void connectComponents() throws OhError {    //; (*6)
    // Open the SCF. Comment strings are removed.
    //
    FileReader       fr = null;
    OhBufferedReader br = null;

    try {
      fr = new FileReader(_scfFile);
      br = new OhBufferedReader(fr, true);

      String  strLine = null;
      boolean foundP  = false;
      
      br.skipWhites();
      while ((strLine = br.readWord()) != null) {
        if (strLine.equals("[ConnectComponents]")) {
          foundP = true;
          break;
        }
      }
      
      if (foundP == false) {
        throw new OhError("No [ConnectComponents] section.");
      }
      
      while (true) {
        String strCom1;
        String strCom2;
        int i1;
        int i2;
        int j;
        
        br.skipWhites();
        strCom1 = br.readWord();
        
        if (strCom1.charAt(0) == 'z') {
          break;
        }
        
        br.skipWhites();
        i1 = Integer.parseInt(br.readWord());
        br.skipWhites();
        strCom2 = br.readWord();
        br.skipWhites();
        i2 = Integer.parseInt(br.readWord());
        br.skipWhites();
        j = Integer.parseInt(br.readWord());
        
        OhElement elm1;
        OhElement elm2;
        
        elm1 = (strCom1.charAt(0) == 'c') ? _elements[i1] : _iterators[i1];
        elm2 = (strCom2.charAt(0) == 'c') ? _elements[i2] : _iterators[i2];
        
        elm1.register(elm2, j);
      }
    }
    catch (FileNotFoundException e) {
      throw new OhError(OhError.getStackTraceString(e));
    }
    catch (IOException e) {
      throw new OhError(OhError.getStackTraceString(e));
    }
    finally {
      if (br != null) {
        try { br.close(); } catch (Exception e) {}
      }
    }
  }


  /**
   * <japanese>
   * ̓Xg[ aBr  _currentTime ǂݍ݂܂B
   * t@C _currentTime ǂݎꍇقƂǂƎv܂B
   * </japanese>
   * <p>
   * Set initial state by reading a file.
   *
   * @param aBr an <code>OhBufferedReader</code> value
   * @exception OhError if an error occurs
   */
  public void setInitialState(OhBufferedReader aBr) throws OhError {    //; (*9)
    int  year = 0;
    long tsec = 0L;
    double frac = 0.0;
    
    try {
      aBr.skipWhites();
      year = Integer.parseInt(aBr.readWord());
      tsec = Long.parseLong(aBr.readWord());
      frac = Double.parseDouble(aBr.readWord());
    }
    catch (IOException e) {
      throw new OhError(
        "Error in OhScfTotalSystem.setInitialState().",
        oPrintString("\t"));
    }
    catch (NumberFormatException e) {
      throw new OhError(
        "Error in OhScfTotalSystem.setInitialState().",
        oPrintString("\t"));
    }

    if (year > 0) {
      _currentTime = new OhTimePoint(year, tsec);
    }
  }

  // @end {methods} --------------------------------------------

} // OhScfSubSystem
