/**
 * @(#) OhSenderPort; class to express a sender port.
 *
 * (Ref "OhObject")
 * <-- (Ref "OhPort")
 * <-- (Ref "OhSenderPort")
 */

package ohymosj;

import java.util.*;
import java.io.*;


/**
 * Abstract class for creating sender ports. It extends
 * (Ref "OhPort").<p>
 *
 * By inheritance from this OhSenderPort class,
 * the classes OhSendPort
 * and OhRelayPort are created.
 *
 * <japanese>
 * <p>
 * {^Mn[qf̃NX <p>
 *
 * ̃NX́AMn[q쐬邽߂̒ۃNXłB
 * Mn[q͑M@\[qłB<p>
 *
 * ̒ۃNX OhSenderPort A
 * ŎĂf[^𑗐M OhSendPort 
 * ̒[q瓾f[^𒆌p OhRelayPort h܂B<p>
 *
 * <ul>
 *  <li> OhSendPort ́AgŃf[^Ă鑗M[qłB
 *  <li> OhRelayPort ́A
 *       ł̓f[^ĂȂAM[q̑Mv
 *       󂯂A̒[qf[^ē][q
 *       B
 * </ul>
 *
 * ɂĂAM炷΁Af[^𑗐MĂ@\
 * Ă̂ŁA̋ʕ̃C^[tF[X 
 * OhSenderPort Œ`Ă܂B<p>
 *
 * OhSenderPort ɂ͕̑M[qif[^M[qj
 * 邱Ƃł܂B
 * </japanese>
 *
 *
 * @author satoh, shiiba
 * @version $Revision: 1.49 $ $Date: 2004/01/09 01:36:27 $
 */
public abstract class OhSenderPort extends OhPort
{
  // @begin {instance variable or static variable} -------------

  /**
   * The list for work.
   *
   * <japanese>
   * <p>
   * ƗpXg
   * </japanese>
   */
  protected OhLinkedList _bufferList;

  
  /**
   * Array for memorizing the ports of sending place.
   *
   * <japanese>
   * <p>
   * MM[qLpz
   * </japanese>
   */
  protected Object[] _receiverArray;
  
  /**
   * The number of elements of "_receiverArray".
   *
   * <japanese>
   * <p>
   * _receiverArray ̗vf
   * </japanese>
   */
  protected int _numberOfReceivers;
  
  /**
   * Comment about data.
   *
   * <japanese>
   * <p>
   * f[^
   * </japanese>
   */
  protected String _commentString;

  // @end {instance variable or static variable} ---------------
  
  
  // @begin {abstract methods} ---------------------------------

  // 1. Connection function.
  
  /**
   * Returns the datapack type string of this port. Because this
   * is an abstract method, it should be implemented in derived
   * classes.
   *
   * <japanese>
   * <p>
   * f[^̌^ԂB<p>
   *
   * NX OhPort ̒ۃ\bhłB
   * ̃NX OhSenderPort łۃ\bhƂĂ܂B
   * </japanese>
   *
   * @return a <code>String</code> value
   */
  public abstract String dataType();
  
  
  /**
   * Abstract method.
   * It should return the data connection vector.
   *
   * <japanese>
   * <p>
   * ̑M[qf[^̃f[^Ԍ^xNgԂ
   * B <p>
   *
   * </japanese>
   *
   * @return a <code>Vector</code> value
   */
  public abstract Vector getDctVec();


  // 2. Function for sending datapacks.
  
  /**
   * Abstract method.
   * Makes data pack array.
   *
   * <japanese>
   * <p>
   * f|^pbNz̍쐬 <p>
   *
   * NX OhPort ̒ۃ\bhłB
   * ̃NX OhSenderPort łۃ\bhƂĂ܂B<p>
   *
   * ۂɂ́A
   * _dataPackArray ̗vf _numberOfDataPacks ɐݒ肷邾
   * łB
   * </japanese>
   *
   */
  public abstract void setDataPackArray();

  
  /**
   * Abstract method.
   * Checks the data time recorded.
   *
   * <japanese>
   * 擾\f[^pbN̊mF <p>
   *
   * NX OhPort ̒ۃ\bhłB
   * ̃NX OhSenderPort łۃ\bhƂĂ܂B<p>
   *
   * canYouGetData(OhTimePoint aTFrom, OhTimePoint aTTo) ,
   * aTFrom ȑOAaTTo ȍ~܂ł̃f[^pbN̒[q
   * 擾\Ȃ΁Atrue ԂAs\Ȃ false Ԃ悤
   * ehNXŒ`܂.
   * </japanese>
   *
   * @param aTFrom an <code>OhTimePoint</code> value
   * @param aTTo an <code>OhTimePoint</code> value
   * @return a <code>boolean</code> value
   */
  public abstract boolean canYouGetData(
    OhTimePointImmutable aTFrom,
    OhTimePointImmutable aTTo);
  
  /**
   * Abstract method.
   * Gets the datapack array which this port can access.
   *
   * <japanese>
   * <p>
   * f[^pbNz̑Mi擾j <p>
   *
   * NX OhPort ̒ۃ\bhłB
   * ̃NX OhSenderPort łۃ\bhƂĂ܂B<p>
   *
   * hNXɂāAf[^pbNz擾郁\bh
   * `܂BaDpArray ɂ̓f[^pbNzi[AaNumOfDp 
   * ɂ́AaDpArray ̗vfi[܂B
   * </japanese>
   *
   * @param aDpArray an <code>OhArrayList</code> value
   * @param aNumOfDp an <code>int[]</code> value
   */
  public abstract void getDataPackArray(
    OhArrayList[] aDpArray, int[] aNumOfDp);
  
  /**
   * Abstract method.
   * Gets the datapack array.
   *
   * @return an <code>OhArrayList</code> value
   */
  public abstract OhArrayList getDataPackArray();
  
  /**
   * Abstract method.
   * Sends the datapacks.
   *
   * <japanese>
   * <p>
   * wɂf[^pbN̑Mi擾j <p>
   *
   * NX OhPort ̒ۃ\bhłB
   * ̃NX OhSenderPort łۃ\bhƂĂ܂B<p>
   *
   * hNXɂāAw肵ăf[^pbN擾
   * `܂B<p>
   *
   * aFormer ɂ́Af[^L^ aDataTime ȑO aDataTime 
   * ɍł߂^f[^pbN擾Ai[܂B<p>
   *
   * aLater ɂ́Af[^L^ aDataTime ȍ~ aDataTime 
   * ł߂^f[^pbN擾Ai[܂B<p>
   *
   * aFlag ͎擾̃[hI邽߂̃IvVŁAL̒
   * łB
   *
   * <pre>
   *   E"b"       aFormer, aLater ̗K擾B
   *               擾łȂꍇ́AƂ null ɐݒ肷B
   *   E"f"       aLater 擾łȂĂAaFormer ̂ݎ擾B
   *               aLater 擾łȂꍇAnull ɐݒ肷B
   *   E"l"       aFormer 擾łȂĂAaLater ̂ݎ擾B
   *               aFormer 擾łȂꍇAnull ɐݒ肷B
   *   Ȇ    aFormer, aLater ̂ȂƂ擾B
   *               擾łȂ̂ null ɐݒ肷B
   * </pre>
   *
   * aFlag ̃ftHgĺA "b" łB<p>
   *
   * ܂A  aLater nȂꍇ́A aFlag  f Ƃ 
   * aFormer ̂ݎ擾A aFormer nȂꍇ́A aFlag 
   *  l Ƃ aLater ̂ݎ擾܂B<p>
   *
   * f[^L^x aDataTime łf[^pbN^
   * Ăꍇ́AaFormer, aLater Ƃi[܂B<p>
   *
   * 𖞂f[^pbN擾łȂꍇAaFormer, aLater 
   *  null ɐݒ肳܂B<p>
   *
   * f[^擾łꍇ́Atrue Ԃ܂B
   * </japanese>
   *
   * @param aFormer an <code>Object[]</code> value
   * @param aDataTime an <code>OhTimePoint</code> value
   * @param aLater an <code>Object[]</code> value
   * @param aMode a <code>char</code> value
   * @return a <code>boolean</code> value
   */
  public abstract boolean getDataPack(
    Object[] aFormer, OhTimePoint aDataTime,
    Object[] aLater, char aMode);
  
  /**
   * Abstract method.
   * Deletion of an unnecessary datapacks.
   *
   * <japanese>
   * svf[^pbN̏ <p>
   *
   * NX OhPort ̒ۃ\bhłB
   * ̃NX OhSenderPort łۃ\bhƂĂ܂B<p>
   *
   * ̒[qŎ擾\ȃf[^pbN̂A
   * getNecessaryTimeFrom() f[^L^f[^pbN
   * 悤AhNXŒ`܂B
   *
   * </japanese>
   *
   */
  public abstract void reduceDataPackStock();

  // @end {abstract methods} -----------------------------------



  // @begin{initialization} ------------------------------------
  
  /**
   * Initialization.
   *
   * @param aPortName a <code>String</code> value
   * @param aPortNum an <code>int</code> value
   */
  public void init(String aPortName, int aPortNum)
    throws OhError {
    super.init(aPortName, aPortNum);
  }

  // @end{initialization} --------------------------------------


  
  // @begin {constructors} -------------------------------------
  
  /**
   * Creates a new default <code>OhSenderPort</code> instance.
   *
   * <japanese>
   * <p>
   * ftHgRXgN^
   * OhSenderPort.init(String, int) s܂B
   * ̌ʁAȉ̐ݒƂȂ܂B
   * <dl>
   *  <li> _objectName        : aPortNameiq[vɃRs[j
   *  <li> _objectNumber      : aPortNum
   * </dl>
   *
   * ̑̃oɑ΂鏈͈RXgN^Ɠ
   * B̏ꍇAOhObject.init(String, int) păo
   * ĉB
   * </jananese>
   */
  public OhSenderPort() {
    super();

    _receiverArray     = null;
    _numberOfReceivers = -1;
    _className         = "OhSenderPort";
    _commentString     = null;

    _bufferList = new OhLinkedList();
  }

  
  /**
   * Creates a new <code>OhSenderPort</code> instance.
   *
   * <japanese>
   * <p>
   * NX OhPort ̃ftHgRXgN^pA
   * OhSenderPort.init(String, int) pāA
   * oϐȉ̒ʂ菉܂B
   *
   * <dl>
   *  <li> _className         : "OhSenderPort"
   *  <li> _objectName        : aPortNameiq[vɃRs[j
   *  <li> _objectNumber      : aPortNum
   *  <li> _owner             : null
   *  <li> _bufferList        : q[vɊmۂBiڑƗpj
   *  <li> _receiverArray     : null
   *  <li> _numberOfReceivers : -1
   * </dl>
   * </japanese>
   *
   * @param aPortName a <code>String</code> value
   * @param aPortNumber an <code>int</code> value
   * @exception OhError if an error occurs
   */
  public OhSenderPort(String aPortName, int aPortNumber) {
    super();
    _receiverArray     = null;
    _numberOfReceivers = -1;
    _className         = "OhSenderPort";
    _commentString     = null;
    _objectName = aPortName;
    _objectNumber = aPortNumber;
    _bufferList = new OhLinkedList();
  }
  
  // @end {constructors} ---------------------------------------

  
  // @begin {methods} ------------------------------------------

  // connection methods
  
  /**
   * Register a receiver.
   * This method is an abstract method declared in the base
   * class <b>OhPort</b>. Implemented in this <b>OhSenerPort</b>
   * class.
   *
   * <japanese>
   * <p>
   * M[q̓o^ <p>
   *
   * NX OhPort ̃\bhI[o[ChĂ܂B<p>
   *
   * ̃\bh́A[qԂ̐ڑ̍ۂɌĂяo܂B<p>
   *
   * OhLinkedList.add(Object) ɂ
   * _bufferList  ł aReceiver o^A
   * true Ԃ܂B<p>
   *
   * Ȁꍇ̓G[bZ[W\ false Ԃ
   * B
   *
   * <ol>
   *  <li> _bufferList  null ̏ꍇB
   *  <li> _receive_port_array 쐬ς݂̏ꍇB
   *  <li> ̒[q aReceiver ƂŁA_dataPackType قȂꍇB
   * <ol>
   *
   * ܂AaReceiver  _bufferList ɓo^ς݂̏ꍇ́Ado^
   *  true Ԃ܂B<p>
   *
   *
   * </japanese>
   *
   * @param aReceiver an <code>Object</code> value
   * @return a <code>boolean</code> value
   */
  public void registerReceiver(OhObject aReceiver)
    throws OhError {
    if (aReceiver == null) {
      throw new IllegalArgumentException(
        "Cannot register 'receiver'" + oPrintString(" > "));
    }

    if (_numberOfReceivers >= 0) {
      throw new OhError(
        "\"_bufferList\" has already been switched.message"
        + NEWLINE + aReceiver.oPrintString(" > ") +
        oPrintString(" > "));
    }

    if (_bufferList == null) {
      throw new OhError(
        "Cannot register 'receiver'" + NEWLINE +
        oPrintString(" > "));
    }      

    // Gets the datapack type which this port sends, and the
    // datapack type which "aReceiver" receives.
    //
    String sType = dataType();   // Abstract method call.
    String rType = ((OhPort)aReceiver).dataType();

    if (!sType.equals(rType)) {
      throw new OhError(
        "'dataType' of this port and \"aReceiver\"(arg) are " +
        "different." + NEWLINE +
        aReceiver.oPrintString("    r > ") +
        oPrintString(" > "));
    }
    
    if (_bufferList.contains(aReceiver) == true)
      return;

    _bufferList.add(aReceiver);
  }
  
  /**
   * Makes "_receiverArray" from the port of the sending place
   * registered into "_bufferList". And, the number of elements
   * of "_receiverArray" is set as "_numberOfReceivers".
   *
   * <japanese>
   * <p>
   * [q̐ڑƂ̏Iуf[^MJn̂߂̏
   *  <p>
   *
   * NX OhPort ̒ۃ\bhĂ܂B<p>
   *
   * _bufferList ɓo^Ă鑗MM[q _receiverArray 
   * ɔz񉻂A_receiverArray ̗vf _numberOfReceivers 
   * ݒ肵܂B<p>
   *
   * _receiverArray 쐬ς݂̏ꍇ́AG[bZ[W\
   *  false Ԃ܂B<p>
   *
   * ڑƂ̏IA_owner ̌vZJnOɕKpĉB
   * Ã\bh̎s͐ڑ֌WύXł܂B<p>
   *
   * _receiverArray ɍ쐬ꂽꍇ true Ԃ܂B
   * </japanese>
   *
   * @exception OhError if an error occurs
   */
  public void doSwitch() throws OhError {    //; (*2)
    if (_numberOfReceivers >= 0) {
      throw new OhError(
        "\"_receiverArray\" has already been set."
        + NEWLINE + oPrintString(" > "));
    }
    _receiverArray     =
      _bufferList.toArray(new Object[_bufferList.size()]);
    _numberOfReceivers = _bufferList.size();
    _bufferList = null;
  }

  
  /**
   * Check whether this is in the state where data packs can be
   * transmitted. If _numberOfReceivers >= 0, _receiverArray has
   * been made, and you can send data packs to a send port.
   *
   * <japanese>
   * <p>
   * vZƑO̊mF <p>
   *
   * NX OhPort ̒ۃ\bh areYouReady() Ă
   * ܂B<p>
   *
   * f[^pbN𑗐M\ȏԂł邱ƂmF܂B_
   * numberOfReceivers  0 ȏȂ _receiverArray ݒς
   * łāAM[qɃf[^pbN𑗐M邱Ƃł܂B
   * ̏ꍇAtrue Ԃ܂BłȂ΁Afalse Ԃ܂B
   * </japanese>
   *
   * @return a <code>boolean</code> value
   */
  public boolean areYouReady() {                    //; (*8)
    return (_numberOfReceivers >= 0) ? true : false;
  }
  
  /**
   * Copy the string "cmt" to "_commentString" which is
   * a instance variable of this class.
   *
   * @param aCmt a <code>String</code> value
   */
  public void setCommentString(String aCmt) {    //; (*6)
    _commentString = aCmt;
  }
  
  /**
   * Returns "_commentString".
   *
   * @return a <code>String</code> value
   */
  public String getCommentString() {    //; (*7)
    return _commentString;
  }

  // Methods for sending datapacks.
  
  /**
   * The minimum thing is returned
   * among OhElement.getNecessaryTimeFrom() of the "_owner"
   * and OhPort.getNecessaryTimeFrom() of
   * the port of a sending place. <p>
   *
   * If "_owner" is "null",
   * Returns OhTimePoint.OhTimePointMin.
   *
   * <japanese>
   * <p>
   * NX OhPort ̒ۃ\bhĂ܂B<p>
   *
   * ̃\bhł́ȂM[qf[^󂯂ƂM[q
   * ԂOhPort.getNecessaryTimeFrom() ̂Aŏ̂̂
   * ܂B<p>
   *
   * ̑M[qf[^󂯎[q΁A
   * OhTimePoint.OhTimePointMax Ԃ܂B
   * </japanese>
   *
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePointImmutable getNecessaryTimeFrom() {
    if (_numberOfReceivers <= 0) {
      return OhTimePoint.OhTimePointMax;
    }
    OhTimePointImmutable t0 = OhTimePoint.OhTimePointMax;
    OhTimePointImmutable t1 = new OhTimePoint();
    for (int k = 0; k < _numberOfReceivers; k++) {
      Object r = _receiverArray[k];
      t1 = ((OhPort)r).getNecessaryTimeFrom();
      if (t1.compareTo(t0) < 0)
        t0 = t1;
    }
    return t0;
  }
  
  /**
   * The maxmum thing is returned
   * among OhElement.getNecessaryTimeTo() of the "_owner"
   * and OhPort.getNecessaryTimeTo() of
   * the port of a sending place. <p>
   *
   * If "_owner" is "null",
   * Returns OhTimePoint.OhTimePointMax.
   *
   * <japanese>
   * <p>
   * NX OhPort ̒ۃ\bhĂ܂B<p>
   *
   * ̃\bhł́A
   * _owner  OhElement.getNecessaryTimeTo()A
   * M[q OhPort.getNecessaryTimeTo() ̂A
   * ő̂̂Ԃ܂B
   * ̎A_owner  null Ȃ
   * OhTimePoint.OhTimePointMax Ԃ܂B
   * </japanese>
   *
   * @return an <code>OhTimePoint</code> value
   */
  public OhTimePointImmutable getNecessaryTimeTo() {        //; (*4)
    if (_owner == null || _numberOfReceivers <= 0) {
      return OhTimePoint.OhTimePointMax;
    }

    OhTimePointImmutable t0 = OhTimePoint.OhTimePointMin;
    OhTimePointImmutable t1 = new OhTimePoint();

    for (int k = 0; k < _numberOfReceivers; k++) {
      Object r = _receiverArray[k];
      t1 = ((OhPort)r).getNecessaryTimeTo();
      if (t1.compareTo(t0) > 0)
        t0 = t1;
    }
    return t0;
  }


  
  // 3. imformation output function.
  
  /**
   * The information about the connection relation of this port
   * is outputted to <b>aPw</b>. <p>
   *
   * <japanese>
   * <p>
   * ڑ̏o <p>
   *
   * ̒[q̐ڑ֌WɊւłXg[ 
   * <b>aPw</b> ɏo͂܂B
   * </japanese>
   *
   * @param aPw a <code>PrintWriter</code> value
   * @param aHead a <code>String</code> value
   */
  public void printConnection(PrintWriter aPw, String aHead) {
    oPrint(aPw, aHead);
    if (_numberOfReceivers <= 0) {
      aPw.print(aHead + "\tNot connected." + NEWLINE);
      return;
    }
    StringBuffer sb = new StringBuffer("");
    if (aHead != null) {
      sb.append(aHead);
    }
    sb.append("Connected > ");
    String str = sb.toString();
    for (int k = 0; k < _numberOfReceivers; k++) {
      ((OhPort)_receiverArray[k]).oPrint(aPw, str);
    }
  }
  
  // @end {methods} --------------------------------------------

} // OhSenderPort
