/* @(#) class OhTime
 *
 * created: 2004/01/04
 * 
 * Development of the OhTime classs<p>
 *
 * <japanese>
 * NX OhTime ̊JB
 * </japanese>
 */
package ohymosj;

import java.io.*;

/**
 * An OhTime object stands for an amount of time in seconds. An
 * OhTime object has two field variables: <b>sec</b> and
 * <b>frac</b>. <b>sec</b> is the integer part of the time, and
 * <b>frac</b> is the fractional part of the time. <b>frac</b> is
 * greater than or equal to -0.5 and less than 0.5.  In other
 * words, <b>sec</b> is the closest long to <b>sec</b> +
 * <b>frac</b>. <p>
 *
 * For example, 10.7 seconds are stood for by <b>sec</b> =
 * 11 and <b>frac</b> = -0.3 and 10.4 seconds are stood for
 * by <b>sec</b> = 10 and <b>frac</b> = 0.4<p>
 *
 * <japanese>
 * <p>
 * OhTime IuWFNǵAԂbPʂŕ\킵܂BOhTime ́A
 * Ԃ\킷߂ <b>sec</b>  <b>frac</b> 2̃tB[
 * hgp܂B<b>sec</b> ́AAfrac ́A
 * \킵܂Bfrac ́A-0.5 ȏA0.5 Ƃ܂B<p>
 *
 * Ⴆ΁A10.7 b́A<b>sec</b> = 11, <b>frac</b> = -0.3 ƕ\
 * ܂B10.3 b́A<b>sec</b> = 10, <b>frac</b> = 0.3 ƕ\
 * 킳܂B<p>
 * </japanese>
 * 
 * @author shiiba
 *
 */

public class OhTime
  implements Cloneable, OhTimeImmutable {
  /**
   * integer part of time.
   *
   * <japanese>
   * <p>
   * ԂbPʂŕ\ƂA̐B
   * </japanese>
   */
  private long _sec;
  
  /**
   * Negligibly small time amount whose default value is 0.0001
   * sec.  When the absolute value of an amount of time is less
   * than this value, the time is regarded as negligibly small.
   *
   * <japanese>
   * <p>
   * ł鎞ԁBΒl̒l菬Ԃ́Ał
   * 炢Ƃ܂B10000 1bɂ܂B
   * </japanese>
   */
  public static double NEGLIGIBLY_SMALL_TIME = 0.0001;
    // on the second time scale

  /**
   * OhTime object with sec = 0 and frac = 0.
   *
   * <japanese>
   * sec  frac  0  OhTime IuWFNgB
   * </japanese>
   *
   */
  public static OhTime ZERO = new OhTime(0, 0.0);

  
  /**
   * fraction part of time
   *
   * <japanese>
   * <p>
   * ԂbPʂŕ\ƂȀB-0.5 ȏA0.5 
   * Ƃ܂B
   * </japanese>
   */
  private double _frac;

  /**
   * Constructs a new OhTime object with <b>sec</b> 0 and
   * <b>frac</b> 0.0.
   *
   * <japanese>
   * <p>
   * OhTime IuWFNg𐶐܂BԂ0ɐݒ肳܂B
   * </japanese>
   */
  public OhTime() {
    super();
    _sec = 0;  _frac = 0.0;
  }

  /**
   * Constructs a new OhTime object by setting <b>sec</b>
   * to <b>aSec</b> and <b>frac</b> to <b>aFrac</b>.<p>
   *
   * <japanese>
   * <p>
   * w肵 <b>sec</b>, <b>frac</b>  OhTime IuWFNg
   * 𐶐܂B
   * </japanese>
   *
   * @param aSec   integer part of time
   * @param aFrac  fraction part of time
   */
  public OhTime(long aSec, double aFrac) {
    super();
    _sec = aSec;
    _frac = aFrac;
    long n = Math.round(_frac);
    _sec += n;
    _frac = _frac - n;
  }

  /**
   * Constructs a new OhTime object which is the same time
   * as <b>aTime</b>
   *
   * @param aTime an <code>OhTimeImmutable</code> value
   */
  public OhTime(OhTimeImmutable aTime) {
    super();
    _sec = aTime.getSec();
    _frac = aTime.getFrac();
  }
  
  /**
   * Constructs a new <code>OhTime</code> instance from a
   * double.
   *
   * <japanese>
   * <p>
   * double ̒l aTime  OhTime IuWFNg𐶐܂B
   * </japanese>
   *
   * @param aTime a <code>double</code> value
   */
  public OhTime(double aTime) {
    _sec = Math.round(aTime);
    _frac = aTime - _sec;
  }

  /**
   * Set <b>sec</b> (integer part of time) to aSec.
   *
   * <japanese>
   * <p>
   *  <b>sec</b>  aSec ɐݒ肵܂B
   * </japanese>
   *
   * @param aSec a <code>long</code> value
   */
  public void setSec(long aSec) {
    _sec = aSec;
  }
  
  /**
   * Set <b>frac</b> (fraction part of time) to aFrac.
   *
   * <japanese>
   * <p>
   * [ <b>frac</b>  aFrac ɐݒ肵܂B
   * </japanese>
   *
   * @param aFrac a <code>double</code> value
   */
  public void setFrac(double aFrac) {
    _frac = aFrac;
  }

  /**
   * Returns <b>sec</b> (integer part of time).
   *
   * <japanese>
   * <p>
   *  <b>sec</b> Ԃ܂B
   * </japanese>
   *
   * @return a <code>long</code> value
   */
  public long getSec() {
    return _sec;
  }

  /**
   * Returns <b>frac</b> (fraction part of time).
   *
   * <japanese>
   * <p>
   * [ <b>frac</b> Ԃ܂B
   * </japanese>
   *
   * @return a <code>double</code> value
   */
  public double getFrac() {
    return _frac;
  }

  /**
   * Adjust <b>sec</b> and <b>frac</b> so that they have compatible signs
   * and the absolute value of <b>frac</b> is less than 1.
   *
   * <japanese>
   * <p>
   * <b>sec</b>  <b>frac</b> ̕܂ĂāA[ <b>frac</b> 
   * -0.5 ȏA0.5 ɂȂ悤ɒ܂B
   * </japanese>
   */
  public void adjust() {
    long n = Math.round(_frac);
    _sec += n;
    _frac = _frac - n;
  }
  
  /**
   * Returns an OhTime object whose time value is equal to
   * aSec + aFrac.<p>
   * 
   * aFrac need not to be greater than or equal to -0.5 and less
   * than 0.5.<p>
   *
   * <japanese>
   * <p>
   * aSec + aFrac ɓԂ OhTime IuWFNgԂ
   * B aFrac ́A-0.5 ȏA0.5 łȂĂ\܂
   * B̃\bh̒Œ܂B<p>
   * </japanese>
   *
   * @param aSec a <code>long</code> value
   * @param aFrac a <code>double</code> value
   * @return an <code>OhTime</code> value
   */
  public static OhTime adjust(long aSec, double aFrac) {
    long n = Math.round(aFrac);
    long sec = aSec + n;
    double frac = aFrac - n;
    return new OhTime(sec, frac);
  }
  
  /**
   * Add the current OhTime object with aTime seconds.
   *
   * <japanese>
   * <p>
   * ݂ OhTime IuWFNgɁAaTime b܂B
   * </japanese>
   *
   * @param aTime a <code>double</code> value
   */
  public void add(double aTime) {
    long sec = Math.round(aTime);
    double frac = aTime - sec;
    _sec += sec;
    _frac += frac;
    this.adjust();
  }

  /**
   * Add the current OhTime object with aT.
   *
   * <japanese>
   * <p>
   * ݂ OhTime IuWFNg OhTime IuWFNg aT 
   * ܂B
   * </japanese>
   *
   * @param aT an <code>OhTime</code> value
   */
  public void add(OhTime aT) {
    _sec += aT.getSec();
    _frac += aT.getFrac();
    this.adjust();
  }
  
  /**
   * Add the current OhTime object with (aSec + aFrac) seconds.
   *
   * <japanese>
   * <p>
   * ݂ OhTime IuWFNgɁA(aSec + aFrac) b܂B
   * </japanese>
   *
   * @param aSec a <code>long</code> value
   * @param aFrac a <code>double</code> value
   */
  public void add(long aSec, double aFrac) {
    _sec += aSec;
    _frac += aFrac;
    this.adjust();
  }

  /**
   * Returns an OhTime object by adding two OhTime objects aTa
   * and aTb.
   *
   * <japanese>
   * <p>
   *  OhTime IuWFNg aTa  aTb 𑫂 OhObject 
   * ĕԂ܂B
   * </japanese>
   *
   * @param aTa an <code>OhTime</code> value
   * @param aTb an <code>OhTime</code> value
   * @return an <code>OhTime</code> value
   */
  public static OhTime add(OhTime aTa, OhTime aTb) {
    long sec = aTa.getSec() + aTb.getSec();
    double frac = aTa.getFrac() + aTb.getFrac();
    return adjust(sec, frac);
  }
  
  /**
   * Subtract aTime seconds fromt the current OhTime object.
   *
   * <japanese>
   * <p>
   * ݂ OhTime IuWFNgAaTime b܂B
   * </japanese>
   *
   * @param aTime a <code>double</code> value
   */
  public void sub(double aTime) {
    long sec = Math.round(aTime);
    double frac = aTime - sec;
    _sec -= sec;
    _frac -= frac;
    this.adjust();
  }

  /**
   * Subtract an OhObject aT from the current OhObject.
   *
   * <japanese>
   * <p>
   * ݂ OhTime IuWFNg OhTime IuWFNg aT 
   * ܂B
   * </japanese>
   *
   * @param aT an <code>OhTime</code> value
   */
  public void sub(OhTime aT) {
    _sec -= aT.getSec();
    _frac -= aT.getFrac();
    this.adjust();
  }
  
  /**
   * Returns an OhTime object by subtracting an OhTime object aTb
   * from an OhTime object aTa.
   *
   * <japanese>
   * <p>
   * OhTime IuWFNg aTa  OhTime IuWFNg aTb ̍ł
   * OhTime IuWFNgԂ܂B
   * </japanese>
   *
   * @param aTa an <code>OhTime</code> value
   * @param aTb an <code>OhTime</code> value
   * @return an <code>OhTime</code> value
   */
  public static OhTime sub(OhTime aTa, OhTime aTb) {
    long sec = aTa.getSec() - aTb.getSec();
    double frac = aTa.getFrac() - aTb.getFrac();
    return adjust(sec, frac);
  }

  /**
   * Multiplies the current OhTime object by aI.
   *
   * <japanese>
   * <p>
   * ݂ OhTime IuWFNg aI {܂B
   * </japanese>
   *
   * @param aI an <code>int</code> value
   */
  public void mul(int aI) {
    _sec = _sec * aI;
    _frac = _frac * aI;
    this.adjust();
  }

  /**
   * Returns the product of an OhTime object aTime and an
   * integer aI.
   *
   * <japanese>
   * <p>
   * OhTime IuWFNg aTime Ɛ aI ̐ς OhTime IuWF
   * NgƂĕԂ܂B
   * </japanese>
   *
   * @param aTime an <code>OhTime</code> value
   * @param aI an <code>int</code> value
   * @return an <code>OhTime</code> value
   */
  public static OhTime mul(OhTime aTime, int aI) {
    long sec = aTime.getSec() * aI;
    double frac = aTime.getFrac() * aI;
    return adjust(sec, frac);
  }

  /**
   * Copies the field varibles of the source OhTime object into
   * those of the current OhTime object.
   *
    * <japanese>
   * <p>
   * Ŏw肳ꂽ OhTimeImmutable IuWFNg̃tB[h
   * ݂ OhTime IuWFNg̃tB[hϐɕʂ܂B
   * </japanese>
   * 
   * @param aTime an <code>OhTimeImmutable</code> value
   */
  public final void copy(OhTimeImmutable aTime) {
    _sec  = aTime.getSec();
    _frac = aTime.getFrac();
  }

  /**
   * Returns the double value which is equal to <b>sec</b> +
   * <b>frac</b>.
   *
   * <japanese>
   * <p>
   * ݂ OhTime IuWFNgɑΉ鎞Ԃ double lԂ
   * ܂B
   * </japanese>
   *
   * @return a <code>double</code> value
   */
  public double toDouble() {
    return _frac + _sec;
  }
  
  /**
   * Returns the string expression of the current object.<p>
   *
   * <japanese>
   * <p>
   * IuWFNg String \Ԃ܂B
   * </japanese>
   *
   * @return a <code>String</code> value
   */
  public String toString() {
    return Long.toString(_sec) + " " + Double.toString(_frac);
  }
  
  /**
   * Returns a deep copy of the current OhTime object
   *
   * @return an <code>Object</code> value
   */
  public Object clone() {
    try {
      OhTime ot = (OhTime)super.clone();
      return ot;
    } catch (CloneNotSupportedException e) {
      throw new Error("This should never happen!");
    }
  }
  
  /**
   * Check whether the current OhTimeImmutable object is equal
   * to <b>anObj</b>.
   * <p>
   *
   * <japanese>
   * Ŏw肳ꂽIuWFNgAOhTimeImmutable ^ (܂
   * ̃TuNX) ̃IuWFNgłāA\킵Ă鎞
   * ỸIuWFNgƓł邩 (true) قȂ邩 (false) 
   * ܂B
   * </japanese>
   *
   * @param anObj an <code>Object</code> value
   * @return a <code>boolean</code> value
   */
  public boolean equals(Object anObj) {
    if ((anObj != null) && (anObj instanceof OhTimeImmutable)) {
      if (((OhTime)anObj)._sec == _sec
        && ((OhTime)anObj)._frac == _frac) return true;
    }
    return false;
  }
    
  /**
   * Check whether the time difference between the current time
   * point object and <b>anObj</b> is less than <b>anEps</b>.
   * <p>
   *
   * <japanese>
   * ݂ OhTimePointImmutable IuWFNg <b>anObj</b> 
   * \킷Ƃ̎ԍ <b>anEps</b> 菬 (true)A
   * łȂ (false) Ԃ܂B
   * <p>
   * </japanese>
   *
   * @param anObj an <code>Object</code> value
   * @param anEps a <code>double</code> value
   * @return a <code>boolean</code> value
   */
  public boolean equals(Object anObj, double anEps) {
    if ((anObj != null) && (anObj instanceof OhTimeImmutable)) {
      if (((OhTime)anObj)._sec == _sec
        && (Math.abs(((OhTime)anObj)._frac - _frac) < anEps)  ) return true;
    }
    return false;
  }

}
